/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 2; tab-width: 2 -*- */
/* GeoIP.h
 *
 * Copyright (C) 2003 MaxMind LLC
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: GeoIP.h 229 2005-05-14 05:02:06Z trystan $
 *
 */

#ifndef GEOIP_H
#define GEOIP_H

typedef struct GeoIPRecordTag {
	char *country_code;
	char *country_code3;
	char *country_name;
	char *region;
	char *city;
	char *postal_code;
	float latitude;
	float longitude;
	int dma_code;
	int area_code;
} GeoIPRecord;

typedef struct GeoIPBitReaderTag {
  FILE *GeoIPBitFH;
  char bits;
  short int position;
} GeoIPBitReader;

typedef enum {
  EOF_FLAG    = 0,
  NOOP_FLAG   = 1,
  VALUE_FLAG  = 2
} GeoIPBitReaderRecordCode;

typedef struct GeoIPTag {
  FILE *GeoIPDatabase;
  char *file_path;
	unsigned char *cache;
	unsigned char *index_cache;
	unsigned int *databaseSegments;
	char databaseType;
	time_t mtime;
	int flags;
	char record_length;
	int record_iter; /* used in GeoIP_next_record */
} GeoIP;

typedef enum {
	GEOIP_STANDARD = 0,
	GEOIP_MEMORY_CACHE = 1,
	GEOIP_CHECK_CACHE = 2,
	GEOIP_INDEX_CACHE = 4
} GeoIPOptions;

typedef enum {
	GEOIP_COUNTRY_EDITION     = 1
} GeoIPDBTypes;

GeoIP* GeoIP_open_type (int type, int flags);
GeoIP* GeoIP_new(int flags);
GeoIP* GeoIP_open(char *filename, int flags);
int GeoIP_db_avail(int type);
void GeoIP_delete(GeoIP* gi);
const char *GeoIP_country_code_by_addr (GeoIP* gi, const char *addr);
const char *GeoIP_country_code_by_name (GeoIP* gi, const char *host);
const char *GeoIP_country_code3_by_addr (GeoIP* gi, const char *addr);
const char *GeoIP_country_code3_by_name (GeoIP* gi, const char *host);
const char *GeoIP_country_name_by_addr (GeoIP* gi, const char *addr);
const char *GeoIP_country_name_by_name (GeoIP* gi, const char *host);
const char *GeoIP_country_code_by_ipnum (GeoIP* gi, unsigned long ipnum);
int GeoIP_id_by_addr (GeoIP* gi, const char *addr);
int GeoIP_id_by_name (GeoIP* gi, const char *host);
int GeoIP_id_by_ipnum (GeoIP* gi, unsigned long ipnum);
char *GeoIP_name_by_addr (GeoIP* gi, const char *addr);
char *GeoIP_name_by_name (GeoIP* gi, const char *host);
char *GeoIP_database_info (GeoIP* gi);
unsigned int _seek_record (GeoIP *gi, unsigned long ipnum);
unsigned long _addr_to_num (const char *addr);

#endif /* GEOIP_H */
