/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 *
 * $Id: config.c 433 2005-07-26 16:58:24Z drstein $
 *
 */

/*************************************************************************/

#include "denora.h"
#define CONF_TABS
#include "confparse.h"

/*************************************************************************/

/* notes on config.c
 * The parser works on two primary depths - blocks and tokens:
 *
 * block { 
 *      token value;
 *      token "string value";
 *      token 123;               # int value
 *      token;                   # nonvar token
 * };
 *
 * It can also parse non-token blocks:
 *
 * block (
 *      "string string string";
 *      "string blah";
 * };
 *
 * Blocks are defined by tconftab (in confparse.h)
 * Tokens are defined by sconftab (^^^^^^^^^^^^^^)
 *
 * Each block must have a function that takes the values collected
 * and checks them against the requirements.  These functions are also
 * handy for getting variables out of the array that they are stored in.
 *
 * The array variables are placed in (an array of cVar structs) contains
 * all values for the the block, and the corrisponding sconftab item.
 */

/* these are our global lists of ACTIVE conf entries */

aAdmin *admins = NULL;
Conf_Modules *modules = NULL;
char *RemoteServer;
int RemotePort;
char *RemotePassword;
char *IRCDModule;

char *LocalHost;
int LocalPort;

char *ExcludeServer;

char *ServerName;
char *ServerDesc;
char *ServiceUser;
char *ServiceHost;
static char *temp_userhost;

char *LogChannel;
char *NetworkName;

char *s_StatServ;
char *desc_StatServ;

char *s_StatServ_alias;
char *desc_StatServ_alias;

char *PIDFilename;
char *MOTDFilename;
char *HTMLFilename;

int ReadTimeout;
int WarningTimeout;
int KeepLogs;
int KeepBackups;
int DumpCore;
int UseTokens;
int StatsLanguage = 0;

int HideStatsO;

int ModulesNumber;
int ModulesDelayedNumber;

char *ChannelDB;
char *ctcpDB;
char *ServerDB;
char *ChannelStatsDB;
char *TLDDB;
char *excludeDB;
char *statsDB;

char *SqlHost;
char *SqlUser;
char *SqlPass;
char *SqlName;
int SqlPort;
char *SqlSock;
int SqlRetries = 0;
int SqlRetryGap = 0;
int sqltype = 0;

char *Numeric;
int LargeNet;

int UserCacheTime;
int UserCleanFreq;
int ServerCacheTime;
int ServerCleanFreq;
int UpdateTimeout;
int HTMLTimeout;
int UptimeTimeout;
int PingFrequency;
int SQLPingFreq;
int ClearInActive;
int ClearChanInActive;

int HidePasswords;
char *Smiley;

char *UserTable;
char *ChanBansTable;
char *ChanExceptTable;
char *IsOnTable;
char *ServerTable;
char *GlineTable;
char *ChanTable;
char *MaxValueTable;
char *ChanInviteTable;
char *TLDTable;
char *CTCPTable;
char *SglineTable;
char *SqlineTable;
char *ChanStatsTable;
char *ServerStatsTable;

char *AliasesTable;
char *CStatsTable;
char *UStatsTable;
char *CurrentTable;
char *StatsTable;
char *SpamTable;

int SP_HTML;

int UseTS6;
int UseZIP;
char *NickChar;
int CTCPUsers;

int CSDefFlag = 0;
int BackupFreq;
int KeepBackupsFor;

int AutoOp;

/* this set of lists is used for loading and rehashing the config file */

aAdmin *new_admins = NULL;
Conf_Modules *new_modules = NULL;

/*************************************************************************/

aAdmin *make_admin(void)
{
    aAdmin *i;
    i = (struct Conf_Admin *) smalloc(sizeof(aAdmin));
    memset((char *) i, '\0', sizeof(aAdmin));
    return i;
}

/*************************************************************************/

void free_admin(aAdmin * ptr)
{
    int i = 0;
    while (ptr->hosts[i]) {
        denora_free(ptr->hosts[i]);
        i++;
    }
    denora_free(ptr->passwd);
    denora_free(ptr->name);
    denora_free(ptr);
    return;
}

/*************************************************************************/

aAdmin *find_admin(char *name, User * u)
{
    aAdmin *aadmin;
    int i;

    for (aadmin = admins; aadmin; aadmin = aadmin->next) {
        if (aadmin->legal == -1)
            continue;
        for (i = 0; aadmin->hosts[i]; i++) {
            if ((!stricmp(name, aadmin->name))
                && match_usermask(aadmin->hosts[i], u)) {
                return aadmin;
            }
        }
    }
    return NULL;
}

/*************************************************************************/

aAdmin *find_admin_byname(char *name)
{
    aAdmin *aadmin;
    for (aadmin = admins; aadmin; aadmin = aadmin->next)
        if (!stricmp(name, aadmin->name))
            break;
    return aadmin;
}

/*************************************************************************/

int confadd_connect(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SCONFF_HOSTNAME)) {
            tmp->type = NULL;
            RemoteServer = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_PORT)) {
            tmp->type = NULL;
            RemotePort = atoi(tmp->value);
            if (RemotePort < 1 || RemotePort > 65535) {
                confparse_error(langstring(CONFIG_PORT_ERROR), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & SCONFF_PASSWD)) {
            tmp->type = NULL;
            RemotePassword = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_PROTOCOL)) {
            tmp->type = NULL;
            IRCDModule = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_BINDHOST)) {
            tmp->type = NULL;
            LocalHost = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_BINDPORT)) {
            tmp->type = NULL;
            LocalPort = atoi(tmp->value);
            if (LocalPort < 1 || LocalPort > 65535) {
                confparse_error(langstring(CONFIG_PORT_ERROR), lnum);
                return -1;
            }
        }
    }
    if (!RemoteServer) {
        confparse_error(langstring(CONFIG_HOSTNAME_ERROR), lnum);
        return -1;
    }
    if (BadPtr(IRCDModule)) {
        confparse_error(langstring(CONFIG_PROTOCOL_ERROR), lnum);
        return -1;
    }
    if (!RemotePort) {
        confparse_error(langstring(CONFIG_PORT_ERROR_NOTDEF), lnum);
        return -1;
    }
    if (BadPtr(RemotePassword)) {
        confparse_error(langstring(CONFIG_PASSWD_ERROR), lnum);
        return -1;
    }
    return lnum;
}

int is_valid_server(char *name, int num)
{
    unsigned char *s;
    int valid = 0;
    for (s = (unsigned char *) name; *s; s++) {
        if (*s < ' ' || *s > '~') {
            valid = 0;
            break;
        }
        if (*s == '.') {
            valid = 1;
        }
    }
    if (!valid) {
        confparse_error("Invalid server name", num);
        return 0;
    }
    return 1;
}

/*************************************************************************/

int confadd_identity(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SCONFF_NAME)) {
            /* validate server name */
            if (!is_valid_server(tmp->value, lnum)) {
                return -1;
            }
            tmp->type = NULL;
            ServerName = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_DESC)) {
            tmp->type = NULL;
            ServerDesc = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_USER)) {
            char *s;
            tmp->type = NULL;
            temp_userhost = sstrdup(tmp->value);

            /* Strip the user and host from user token */
            if (!(s = strchr(temp_userhost, '@'))) {
                confparse_error("Missing `@' for Stats User", lnum);
                return -1;
            }
            *s++ = 0;
            ServiceUser = temp_userhost;
            ServiceHost = s;
        } else if (tmp->type && (tmp->type->flag & SCONFF_LANGUAGE)) {
            tmp->type = NULL;
            StatsLanguage = strtol(tmp->value, NULL, 10);
            if (StatsLanguage < 1 || StatsLanguage > NUM_LANGS) {
                StatsLanguage = 1;
                confparse_error(langstring(CONFIG_INVALID_LANG), lnum);
                return -1;
            }
        }
    }
    if (BadPtr(ServerName)) {
        confparse_error(langstring(CONFIG_ID_NAME_ERROR), lnum);
        return -1;
    }
    if (BadPtr(ServerDesc) || !ServerDesc) {
        confparse_error(langstring(CONFIG_ID_DESC_ERROR), lnum);
        return -1;
    }
    if (BadPtr(ServiceUser)) {
        confparse_error(langstring(CONFIG_ID_USER_ERROR), lnum);
        return -1;
    }
    if (BadPtr(ServiceHost)) {
        confparse_error(langstring(CONFIG_ID_HOST_ERROR), lnum);
        return -1;
    }
    if (!StatsLanguage) {
        confparse_error(langstring(CONFIG_ID_LANG_ERROR), lnum);
        return -1;
    }
    if (StatsLanguage) {
        /* Now we reduce StatsLanguage by 1 */
        StatsLanguage--;
    }
    return lnum;
}

/*************************************************************************/

int confadd_statserv(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SCONFF_NICK)) {
            tmp->type = NULL;
            s_StatServ = sstrdup(tmp->value);
            if (strlen(s_StatServ) > NICKMAX) {
                confparse_error(langstring(CONFIG_SS_TOLONG), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & SCONFF_REAL)) {
            tmp->type = NULL;
            desc_StatServ = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_ALIASNICK)) {
            tmp->type = NULL;
            s_StatServ_alias = sstrdup(tmp->value);
            if (strlen(s_StatServ_alias) > NICKMAX) {
                confparse_error(langstring(CONFIG_SS_TOLONG), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & SCONFF_ALIASREAL)) {
            tmp->type = NULL;
            desc_StatServ_alias = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_AUTOOP)) {
            tmp->type = NULL;
            AutoOp = 1;
        }
    }
    if (BadPtr(s_StatServ)) {
        confparse_error(langstring(CONFIG_SS_NAME_ERROR), lnum);
        return -1;
    }
    if (BadPtr(desc_StatServ)) {
        confparse_error(langstring(CONFIG_SS_DESC_ERROR), lnum);
        return -1;
    }
    if (s_StatServ_alias && BadPtr(desc_StatServ_alias)) {
        confparse_error(langstring(CONFIG_SS_ALIASDESC_ERROR), lnum);
        return -1;
    }
    if (s_StatServ_alias) {
        if (BadPtr(s_StatServ_alias)) {
            confparse_error(langstring(CONFIG_SS_ALIASDESC_ERROR), lnum);
            return -1;
        }
    }
    return lnum;
}

/*************************************************************************/

int confadd_filenames(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SCONFF_PID)) {
            tmp->type = NULL;
            PIDFilename = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_MOTD)) {
            tmp->type = NULL;
            MOTDFilename = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_HTMLFILE)) {
            tmp->type = NULL;
            HTMLFilename = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_CHANDB)) {
            tmp->type = NULL;
            ChannelDB = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_STATSDB)) {
            tmp->type = NULL;
            statsDB = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_CTCPDB)) {
            tmp->type = NULL;
            ctcpDB = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_SERVERDB)) {
            tmp->type = NULL;
            ServerDB = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_CHANSTATSDB)) {
            tmp->type = NULL;
            ChannelStatsDB = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_TLDDB)) {
            tmp->type = NULL;
            TLDDB = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_EXCLUDEDB)) {
            tmp->type = NULL;
            excludeDB = sstrdup(tmp->value);
        }
    }
    if (!PIDFilename) {
        confparse_error(langstring(CONFIG_FILENAME_PID_ERROR), lnum);
        return -1;
    }
    if (!MOTDFilename) {
        confparse_error(langstring(CONFIG_FILENAME_MOTD_ERROR), lnum);
        return -1;
    }
    if (HTMLFilename) {
        denora->do_html = 1;
    }
    if (!ChannelDB) {
        confparse_error(langstring(CONFIG_FILENAME_CHANDB_ERROR), lnum);
        return -1;
    }
    if (!ctcpDB) {
        confparse_error(langstring(CONFIG_FILENAME_CTCP_ERROR), lnum);
        return -1;
    }
    if (!ServerDB) {
        confparse_error(langstring(CONFIG_FILENAME_SERVER_ERROR), lnum);
        return -1;
    }
    if (!ChannelStatsDB) {
        confparse_error(langstring(CONFIG_FILENAME_CHANSTATS_ERROR), lnum);
        return -1;
    }
    if (!TLDDB) {
        confparse_error(langstring(CONFIG_FILENAME_TLD_ERROR), lnum);
        return -1;
    }
    if (!excludeDB) {
        confparse_error(langstring(CONFIG_FILENAME_EXCLUDE_ERROR), lnum);
        return -1;
    }
    if (!statsDB) {
        confparse_error("statsDB is not defined", lnum);
        return -1;
    }
    return lnum;
}

/*************************************************************************/

int confadd_netinfo(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SCONFF_NAME)) {
            tmp->type = NULL;
            NetworkName = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_NUMERIC)) {
            tmp->type = NULL;
            Numeric = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_LARGENET)) {
            tmp->type = NULL;
            LargeNet = 1;
        } else if (tmp->type && (tmp->type->flag & SCONFF_SP_HTML)) {
            tmp->type = NULL;
            SP_HTML = 1;
        } else if (tmp->type && (tmp->type->flag & SCONFF_EXCLUDESERV)) {
            tmp->type = NULL;
            ExcludeServer = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_NICKCHAR)) {
            tmp->type = NULL;
            NickChar = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_CTCPUSERS)) {
            tmp->type = NULL;
            CTCPUsers = 1;
        }

    }
    if (BadPtr(NetworkName)) {
        confparse_error(langstring(CONFIG_NETINFO_NAME_ERROR), lnum);
        return -1;
    }
    if (LargeNet) {
        alog(LOG_NORMAL, langstr(ALOG_LARGENET_WARNING_1));
        alog(LOG_NORMAL, langstr(ALOG_LARGENET_WARNING_2));
        alog(LOG_NORMAL, langstr(ALOG_LARGENET_WARNING_3));
    }
    return lnum;
}

/*************************************************************************/

int confadd_timeout(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & TIMEFF_READ)) {
            tmp->type = NULL;
            ReadTimeout = dotime(tmp->value);
            if (ReadTimeout < 0) {
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_WARNING)) {
            tmp->type = NULL;
            WarningTimeout = dotime(tmp->value);
            if (WarningTimeout < 0) {
                printf("WarningTimeout");
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_UPDATE)) {
            tmp->type = NULL;
            UpdateTimeout = dotime(tmp->value);
            if (UpdateTimeout < 0) {
                printf("UpdateTimeout");
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_HTMLUPDATE)) {
            tmp->type = NULL;
            HTMLTimeout = dotime(tmp->value);
            if (HTMLTimeout < 119) {
                confparse_error(langstring(CONFIG_HTML_UPDATE_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_UPTIMEFREQ)) {
            tmp->type = NULL;
            UptimeTimeout = dotime(tmp->value);
            if (UptimeTimeout < 119) {
                confparse_error(langstring(CONFIG_UPTIME_TO_LOW), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_PINGFREQ)) {
            tmp->type = NULL;
            PingFrequency = dotime(tmp->value);
            if (PingFrequency < 119) {
                confparse_error(langstring(CONFIG_PING_TO_LOW), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_SQLFREQ)) {
            tmp->type = NULL;
            SQLPingFreq = dotime(tmp->value);
            if (SQLPingFreq < 119) {
                confparse_error(langstring(CONFIG_SQLPING_TO_LOW), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_CINACTIVE)) {
            tmp->type = NULL;
            ClearInActive = dotime(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TIMEFF_CHANINACTIVE)) {
            tmp->type = NULL;
            ClearChanInActive = dotime(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TIMEFF_UCACHE)) {
            tmp->type = NULL;
            UserCacheTime = dotime(tmp->value);
            if (UserCacheTime < 0) {
                printf("UserCacheTime");
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_UFREQ)) {
            tmp->type = NULL;
            UserCleanFreq = dotime(tmp->value);
            if (UserCleanFreq < 0) {
                printf("UserCleanFreq");
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_SCACHE)) {
            tmp->type = NULL;
            ServerCacheTime = dotime(tmp->value);
            if (ServerCacheTime < 0) {
                printf("ServerCacheTime");
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & TIMEFF_SFREQ)) {
            tmp->type = NULL;
            ServerCleanFreq = dotime(tmp->value);
            if (ServerCleanFreq < 0) {
                printf("ServerCleanFreq");
                confparse_error(langstring(CONFIG_INVALID_TIME), lnum);
                return -1;
            }
        }
    }

    if (!ReadTimeout) {
        confparse_error(langstring(CONFIG_TIMEOUT_READ_ERROR), lnum);
        return -1;
    }
    if (!WarningTimeout) {
        confparse_error(langstring(CONFIG_TIMEOUT_WARNING_ERROR), lnum);
        return -1;
    }
    if (!UpdateTimeout) {
        confparse_error(langstring(CONFIG_TIMEOUT_UPDATE_ERROR), lnum);
        return -1;
    }
    if (!PingFrequency) {
        confparse_error(langstring(CONFIG_TIMEOUT_PING_ERROR), lnum);
        return -1;
    }
    if (!UptimeTimeout) {
        confparse_error(langstring(CONFIG_TIMEOUT_UPTIME_ERROR), lnum);
        return -1;
    }
    if (!HTMLTimeout) {
        confparse_error(langstring(CONFIG_TIMEOUT_HTML_ERROR), lnum);
        return -1;
    }
    if (!SQLPingFreq) {
        confparse_error(langstring(CONFIG_TIMEOUT_SQL_ERROR), lnum);
        return -1;
    }
    if (!ClearInActive) {
        ClearInActive = dotime("30d");
    }
    if (!ClearChanInActive) {
        ClearChanInActive = dotime("30d");
    }
    return lnum;
}

/*************************************************************************/

int confadd_options(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & OPTF_KEEPLOGS)) {
            tmp->type = NULL;
            KeepLogs = atoi(tmp->value);
            if (KeepLogs < 0) {
                confparse_error(langstring(CONFIG_ERROR_POSTIVE_VALUE),
                                lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & OPTF_LOGCHAN)) {
            tmp->type = NULL;
            LogChannel = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & OPTF_SMILEY)) {
            tmp->type = NULL;
            Smiley = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & OPTF_DUMPCORE)) {
            tmp->type = NULL;
            DumpCore = 1;
        } else if (tmp->type && (tmp->type->flag & OPTF_HIDESTATSO)) {
            tmp->type = NULL;
            HideStatsO = 1;
        } else if (tmp->type && (tmp->type->flag & OPTF_TOKENS)) {
            tmp->type = NULL;
            UseTokens = 1;
        } else if (tmp->type && (tmp->type->flag & OPTF_TS6)) {
            tmp->type = NULL;
            UseTS6 = 1;
        } else if (tmp->type && (tmp->type->flag & OPTF_ZIP)) {
            tmp->type = NULL;
            UseZIP = 1;
        } else if (tmp->type && (tmp->type->flag & OPTF_CSDEF)) {
            tmp->type = NULL;
            if (!stricmp(tmp->value, "on")) {
                CSDefFlag = 1;
            }
            if (!stricmp(tmp->value, "notice")) {
                CSDefFlag = 2;
            }
        }
    }
    if (!Smiley) {
        Smiley = sstrdup(SMILEYS);
    }
    return lnum;
}

/*************************************************************************/

int confadd_admin(cVar * vars[], int lnum)
{
    cVar *tmp;
    aAdmin *x = make_admin();
    int c = 0, hc = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SCONFF_NAME)) {
            if (x->name) {
                confparse_error("Multiple name definitions", lnum);
                free_admin(x);
                return -1;
            }
            tmp->type = NULL;
            x->name = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_PASSWD)) {
            if (x->passwd) {
                confparse_error("Multiple password definitions", lnum);
                free_admin(x);
                return -1;
            }
            tmp->type = NULL;
            x->passwd = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SCONFF_HOSTNAME)) {
            if ((hc + 1) > MAXHOSTS) {
                confparse_error("Excessive host definitions", lnum);
                free_admin(x);
                return -1;
            }
            tmp->type = NULL;
            if (!strchr(tmp->value, '@') && *tmp->value != '/') {
                char *newhost;
                int len = 3;
                len += strlen(tmp->value);
                newhost = (char *) smalloc(len);
                vsnprintf(newhost, sizeof(newhost), "*@%s", tmp->value);
                x->hosts[hc] = newhost;
            } else
                x->hosts[hc] = sstrdup(tmp->value);
            hc++;
        } else if (tmp->type && (tmp->type->flag & SCONFF_LANGUAGE)) {
            tmp->type = NULL;
            x->language = atoi(tmp->value);
            if (x->language < 1 || x->language > NUM_LANGS) {
                confparse_error(langstring(CONFIG_INVALID_LANG), lnum);
                return -1;
            }
            x->language--;
        }
    }
    if (!x->name) {
        confparse_error(langstring(CONFIG_ADMIN_NAME_ERROR), lnum);
        free_admin(x);
        return -1;
    }
    if (!x->hosts[0]) {
        confparse_error(langstring(CONFIG_ADMIN_HOST_ERROR), lnum);
        free_admin(x);
        return -1;
    }
    if (!x->passwd) {
        confparse_error(langstring(CONFIG_ADMIN_PASS_ERROR), lnum);
        free_admin(x);
        return -1;
    }
    x->next = new_admins;
    new_admins = x;
    return lnum;
}

/*************************************************************************/

int confadd_sql(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & SQLFF_TYPE)) {
            tmp->type = NULL;
            if (!stricmp(tmp->value, "mysql")) {
                sqltype = SQL_MYSQL;
            }
            if (!stricmp(tmp->value, "postgre")) {
                sqltype = SQL_POSTGRE;
            }
        } else if (tmp->type && (tmp->type->flag & SQLFF_HOST)) {
            tmp->type = NULL;
            SqlHost = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SQLFF_USER)) {
            tmp->type = NULL;
            SqlUser = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SQLFF_PASSWD)) {
            tmp->type = NULL;
            SqlPass = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SQLFF_NAME)) {
            tmp->type = NULL;
            SqlName = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SQLFF_SOCK)) {
            tmp->type = NULL;
            SqlSock = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & SQLFF_PORT)) {
            tmp->type = NULL;
            SqlPort = atoi(tmp->value);
            if (SqlPort < 1 || SqlPort > 65535) {
                confparse_error(langstring(CONFIG_PORT_ERROR), lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & SQLFF_RETRIES)) {
            tmp->type = NULL;
            SqlRetries = atoi(tmp->value);
            if (SqlRetries < 0) {
                confparse_error(langstring(CONFIG_ERROR_POSTIVE_VALUE),
                                lnum);
                return -1;
            }
        } else if (tmp->type && (tmp->type->flag & SQLFF_RETRYGAP)) {
            tmp->type = NULL;
            SqlRetryGap = atoi(tmp->value);
            if (SqlRetryGap < 0) {
                confparse_error(langstring(CONFIG_ERROR_POSTIVE_VALUE),
                                lnum);
                return -1;
            }
        }
    }
    if (!SqlHost) {
        denora->do_sql = 0;
    }
    if (SqlHost && !SqlUser) {
        confparse_error(langstring(CONFIG_SQL_USER_ERROR), lnum);
        return -1;
    }
    if (SqlHost && !SqlName) {
        confparse_error(langstring(CONFIG_SQL_NAME_ERROR), lnum);
        return -1;
    }
    return lnum;
}

/*************************************************************************/

int confadd_backup(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & BACKUPFF_KEEPBACKUPS)) {
            tmp->type = NULL;
            KeepBackups++;
        } else if (tmp->type && (tmp->type->flag & BACKUPFF_KEEPFOR)) {
            tmp->type = NULL;
            KeepBackupsFor = dotime(tmp->value);
        } else if (tmp->type && (tmp->type->flag & BACKUPFF_FREQ)) {
            tmp->type = NULL;
            BackupFreq = dotime(tmp->value);
        }
    }
    if (KeepBackups && !KeepBackupsFor) {
        confparse_error("KeepBackupsFor is not defined", lnum);
        return -1;
    }
    if (KeepBackups && !BackupFreq) {
        confparse_error("BackupFreq is not defined", lnum);
        return -1;
    }
    return lnum;
}

/*************************************************************************/

int confadd_tables(cVar * vars[], int lnum)
{
    cVar *tmp;
    int c = 0;

    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & TABFF_USERS)) {
            tmp->type = NULL;
            UserTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CHANBANS)) {
            tmp->type = NULL;
            ChanBansTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CHANEXCEPT)) {
            tmp->type = NULL;
            ChanExceptTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_ISON)) {
            tmp->type = NULL;
            IsOnTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_SERVER)) {
            tmp->type = NULL;
            ServerTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_GLINE)) {
            tmp->type = NULL;
            GlineTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CHAN)) {
            tmp->type = NULL;
            ChanTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_MAXVALUE)) {
            tmp->type = NULL;
            MaxValueTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CHANINVITE)) {
            tmp->type = NULL;
            ChanInviteTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_TLD)) {
            tmp->type = NULL;
            TLDTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CTCP)) {
            tmp->type = NULL;
            CTCPTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_SGLINE)) {
            tmp->type = NULL;
            SglineTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_SQLINE)) {
            tmp->type = NULL;
            SqlineTable = sstrdup(tmp->value);

        } else if (tmp->type && (tmp->type->flag & TABFF_CSALIASES)) {
            tmp->type = NULL;
            AliasesTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CSCSTATS)) {
            tmp->type = NULL;
            CStatsTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CSUSTATS)) {
            tmp->type = NULL;
            UStatsTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_STATS)) {
            tmp->type = NULL;
            StatsTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CHANSTATS)) {
            tmp->type = NULL;
            ChanStatsTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_SERVSTATS)) {
            tmp->type = NULL;
            ServerStatsTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_SPAMFILTER)) {
            tmp->type = NULL;
            SpamTable = sstrdup(tmp->value);
        } else if (tmp->type && (tmp->type->flag & TABFF_CURRENT)) {
            tmp->type = NULL;
            CurrentTable = sstrdup(tmp->value);
        }

    }
    if (!UserTable) {
        confparse_error(langstring(CONFIG_TABLE_USER_ERROR), lnum);
        return -1;
    }
    if (!ChanBansTable) {
        confparse_error(langstring(CONFIG_TABLE_CHANBANS_ERROR), lnum);
        return -1;
    }
    if (!ChanExceptTable) {
        confparse_error(langstring(CONFIG_TABLE_CHANEXCEPT_ERROR), lnum);
        return -1;
    }
    if (!IsOnTable) {
        confparse_error(langstring(CONFIG_TABLE_ISON_ERROR), lnum);
        return -1;
    }
    if (!ServerTable) {
        confparse_error(langstring(CONFIG_TABLE_SERVER_ERROR), lnum);
        return -1;
    }
    if (!GlineTable) {
        confparse_error(langstring(CONFIG_TABLE_GLINE_ERROR), lnum);
        return -1;
    }
    if (!ChanTable) {
        confparse_error("Lacking chan definition in tables block", lnum);
        return -1;
    }
    if (!MaxValueTable) {
        confparse_error("Lacking maxvalue definition in tables block",
                        lnum);
        return -1;
    }
    if (!ChanInviteTable) {
        ChanInviteTable = sstrdup("chaninvites");
        return -1;
    }
    if (!TLDTable) {
        confparse_error("Lacking tld definition in tables block", lnum);
        return -1;
    }
    if (!CTCPTable) {
        confparse_error("Lacking ctcp definition in tables block", lnum);
        return -1;
    }
    if (!ChanStatsTable) {
        confparse_error("Lacking chanstats definition in tables block",
                        lnum);
        return -1;
    }
    if (!ServerStatsTable) {
        confparse_error("Lacking servstats definition in tables block",
                        lnum);
        return -1;
    }
    if (!SglineTable) {
        SglineTable = sstrdup("sgline");
    }
    if (!SqlineTable) {
        SqlineTable = sstrdup("sqline");
    }
    if (!AliasesTable) {
        AliasesTable = sstrdup("aliases");
    }
    if (!CStatsTable) {
        CStatsTable = sstrdup("cstats");
    }
    if (!UStatsTable) {
        UStatsTable = sstrdup("ustats");
    }
    if (!StatsTable) {
        StatsTable = sstrdup("stats");
    }
    if (!SpamTable) {
        SpamTable = sstrdup("spamfilters");
    }
    if (!CurrentTable) {
        CurrentTable = sstrdup("current");
    }
    return lnum;
}

/*************************************************************************/

int confadd_modules(cVar * vars[], int lnum)
{
    cVar *tmp;
    Conf_Modules *x = new_modules;
    int c = 0, ac = 0, oc = 0;

    /* we dont free here because we do it if we fail */

    if (!x) {
        x = (Conf_Modules *) smalloc(sizeof(Conf_Modules));
        memset((char *) x, '\0', sizeof(Conf_Modules));
        new_modules = x;
    } else {
        confparse_error("Multiple module blocks in config file", lnum);
        return -1;
    }
    for (tmp = vars[c]; tmp; tmp = vars[++c]) {
        if (tmp->type && (tmp->type->flag & MBTF_AUTOLOAD)) {
            if ((ac + 1) > 128) {
                confparse_error("Excessive autoloading modules (max 128)",
                                lnum);
                return -1;
            }
            tmp->type = NULL;
            x->autoload[ac] = sstrdup(tmp->value);
            ac++;
        } else if (tmp->type && (tmp->type->flag & MBTF_DELAYED)) {
            if ((oc + 1) > 128) {
                confparse_error(langstring(CONFIG_DELAYMOD_T0MANY), lnum);
                return -1;
            }
            tmp->type = NULL;
            x->delayed[oc] = sstrdup(tmp->value);
            oc++;
        }
    }
    ModulesNumber = ac;
    ModulesDelayedNumber = oc;
    return lnum;
}

/*************************************************************************/

/* merge_ functions */

static void merge_admins(void)
{
    aAdmin *aadmin, *old_admin, *ptrn = NULL, *ptr = NULL;

    for (old_admin = admins; old_admin; old_admin = old_admin->next)
        old_admin->legal = -1;

    /* add or merge and del new ones */
    for (aadmin = new_admins; aadmin; aadmin = ptrn) {
        ptrn = aadmin->next;
        if ((old_admin = find_admin_byname(aadmin->name))) {
            int i;

            for (i = 0; old_admin->hosts[i]; i++)
                denora_free(old_admin->hosts[i]);

            denora_free(old_admin->passwd);
            for (i = 0; aadmin->hosts[i]; i++)
                old_admin->hosts[i] = aadmin->hosts[i];
            old_admin->hosts[i] = NULL;
            old_admin->passwd = aadmin->passwd;
            old_admin->legal = 1;

            denora_free(aadmin->name);
            denora_free(aadmin);
        } else {
            aadmin->legal = 1;
            aadmin->next = admins;
            admins = aadmin;
        }
    }
    new_admins = NULL;

    /* del old ones */
    ptr = NULL;
    aadmin = admins;
    while (aadmin) {
        ptrn = aadmin->next;
        if ((aadmin->legal == -1) && (aadmin->admins <= 0)) {
            if (ptr)
                ptr->next = aadmin->next;
            else
                admins = aadmin->next;
            free_admin(aadmin);
        } else
            ptr = aadmin;
        aadmin = ptrn;
    }
    return;
}

/*************************************************************************/

/* Used on RELOAD */
void merge_confs()
{
    int i;

    merge_admins();
    /* dont worry about accually merging module data - its fairly
     * inactive and static data.  Just replace it.
     */
    if (modules) {
        for (i = 0; modules->autoload[i]; i++)
            denora_free(modules->autoload[i]);
        for (i = 0; modules->delayed[i]; i++)
            denora_free(modules->delayed[i]);
        denora_free(modules);
    }
    modules = new_modules;
    new_modules = NULL;
    return;
}

/*************************************************************************/

/* Used if RELOAD fails */
void clear_newconfs()
{
    int i;

    aAdmin *aadmin = new_admins, *aadmin_p;

    while (aadmin) {
        aadmin_p = aadmin->next;
        free_admin(aadmin);
        aadmin = aadmin_p;
    }
    new_admins = NULL;
    if (new_modules) {
        for (i = 0; new_modules->autoload[i]; i++)
            denora_free(new_modules->autoload[i]);
        for (i = 0; new_modules->delayed[i]; i++)
            denora_free(new_modules->delayed[i]);
        denora_free(new_modules);
        new_modules = NULL;
    }
    return;
}

/*************************************************************************/

static void free_vars(cVar * vars[])
{
    int i = 0;

    while (vars[i]) {
        denora_free(vars[i]->value);
        denora_free(vars[i]);
        i++;
    }
}

/*************************************************************************/

static const char *current_file = "unknown";
static int quote = 0;

void confparse_error(const char *problem, int line)
{
    alog(LOG_ERROR, langstr(CONFIG_ERROR_BASE), problem, line,
         current_file);
    if (!denora->nofork && isatty(2)) {
        fprintf(stderr, langstr(CONFIG_ERROR_BASE), problem, line,
                current_file);
        fprintf(stderr, "\n");
    }
    return;
}

/*************************************************************************/

/* check_quote
 * this routine skips over any ignored items inside our file
 */
static char *check_quote(char *cur)
{
    if (quote) {
        while ((cur = strchr(cur, '*')))
            if ((*(++cur) == '/')) {
                cur++;
                quote = 0;
                break;
            }
        if (!cur)
            return cur;
    }
    while ((*cur == ' ') || (*cur == '\t'))
        cur++;
    /* now we've hit something .. check for single line quotes */
    if (!*cur || *cur == '#' || *cur == '\n' ||
        (*cur == '/' && *(cur + 1) == '/'))
        return NULL;
    /* check for multiple line quotes */
    if ((*cur == '/') && (*(cur + 1) == '*')) {
        cur += 2;
        quote = 1;
        while ((cur = strchr(cur, '*')))
            if ((*(++cur) == '/')) {
                cur++;
                quote = 0;
                break;
            }
        if (!cur)
            return cur;
        else
            return check_quote(cur);
    }
    return cur;
}

/*************************************************************************/

static char *parse_block(tConf * block, char *cur, FILE * file, int *lnum)
{
    char *tok, *var, *var2;
    char line[LINE_MAX];
    tConf *b2 = NULL;
    sConf *item = NULL;
    sConf *subconftab = block->subtok;
    cVar *vars[MAX_VALUES] = { 0 };
    int vnum = 0, tlnum = 0, clear = 0, done = 0, skip = 0;

    if ((subconftab) && (subconftab->flag == SCONFF_STRING)) {
        /* this subtype only takes freeform variables
         * dont bother looking for tokens
         */
        int i = 0;
        while (!BadPtr(cur) || ((fgets(line, LINE_MAX, file) != NULL) &&
                                (*lnum)++ && (cur = line))) {
            cur = check_quote(cur);
            if (BadPtr(cur))
                continue;
            if (clear) {
                if (*cur != ';') {
                    confparse_error("Missing semicolon", *lnum);
                    free_vars(vars);
                    return NULL;
                } else
                    cur++;
                clear = 0;
                cur = check_quote(cur);
                if (BadPtr(cur))
                    continue;
            }
            if (done) {
                if (*cur != ';') {
                    confparse_error("Missing block end semicolon", *lnum);
                    free_vars(vars);
                    return NULL;
                } else
                    cur++;
                if (((*block->func) (vars, *lnum)) == -1) {
                    free_vars(vars);
                    return NULL;
                }
                if (BadPtr(cur))
                    *cur = '#'; /* we cant return a bad pointer because
                                 * that will pull us out of the conf read
                                 * so this will just get ignored
                                 * kludgy, but effective */
                free_vars(vars);
                return cur;
            }
            cur = check_quote(cur);
            if (BadPtr(cur))
                continue;
            if (*cur == '}') {
                done = 1;
                cur++;
                cur = check_quote(cur);
                if (BadPtr(cur))
                    continue;
                if (*cur != ';') {
                    confparse_error("Missing block end semicolon", *lnum);
                    free_vars(vars);
                    return NULL;
                } else
                    cur++;
                if (((*block->func) (vars, *lnum)) == -1) {
                    free_vars(vars);
                    return NULL;
                }
                if (BadPtr(cur))
                    *cur = '#'; /* we cant return a bad pointer because
                                 * that will pull us out of the conf read
                                 * so this will just get ignored
                                 * kludgy, but effective */
                free_vars(vars);
                return cur;
            }
            vars[vnum] = (cVar *) malloc(sizeof(cVar));
            memset((char *) vars[vnum], '\0', sizeof(cVar));
            vars[vnum]->loaded = 1;
            vars[vnum]->type = NULL;
            tok = cur;
            if (*cur == '"') {
                i = 1;
                cur++;
            }
            var = cur;
            if (i == 1) {
                while (!BadPtr(cur) && (*cur != '"'))
                    cur++;
                if (BadPtr(cur)) {
                    confparse_error("Cant find closequote", *lnum);
                    free_vars(vars);
                    return NULL;
                }
                *cur = '\0';
                cur++;
                while (!BadPtr(cur) && (*cur != ';'))
                    cur++;
            } else {
                while (!BadPtr(cur) && (*cur != ';')) {
                    if ((*cur == ' ')) {
                        *cur = '\0';
                        if (vars[vnum]->loaded == 1) {
                            vars[vnum]->value = sstrdup(var);
                            vars[vnum]->loaded = 2;
                        }
                    } else if (vars[vnum]->loaded == 2) {
                        confparse_error("Junk after value", *lnum);
                        free_vars(vars);
                        return NULL;
                    }
                    cur++;
                }
            }
            tlnum = *lnum;
            if (BadPtr(cur)) {
                clear = 1;
                continue;
            }
            *cur = '\0';
            cur++;
            if (vars[vnum]->loaded == 1)
                vars[vnum]->value = sstrdup(var);
            vars[vnum]->loaded = 3;
            vnum++;
        }
        confparse_error("Unexpected EOF: Syntax Error", tlnum);
        free_vars(vars);
        return NULL;
    }

    while (!BadPtr(cur)
           || ((fgets(line, LINE_MAX, file) != NULL) && (*lnum)++
               && (cur = line))) {
        cur = check_quote(cur);
        if (BadPtr(cur))
            continue;
        if (clear) {
            /* if we're looking for a closing semicolon, check for it first
             * if we cant find it, ignore it and hope for the best
             */
            if (*cur != ';') {
                confparse_error("Missing semicolon ", *lnum);
                free_vars(vars);
                return NULL;
            } else
                cur++;
            clear = 0;
            if (vars[vnum]) {
                vars[vnum]->loaded = 3;
                vnum++;
            }
            item = NULL;
            cur = check_quote(cur);
            if (BadPtr(cur))
                continue;
        }
        if (done) {
            /* we've found the end of our block, now we're looking for the
             * closing semicolon.  if we cant find it, ignore it and
             * hope for the best
             */
            if (*cur != ';') {
                confparse_error("Missing block end semicolon", *lnum);
                free_vars(vars);
                return NULL;
            } else
                cur++;
            if (((*block->func) (vars, *lnum)) == -1) {
                free_vars(vars);
                return NULL;
            }
            if (BadPtr(cur))
                *cur = '#';     /* we cant return a bad pointer because
                                 * that will pull us out of the conf read
                                 * so this will just get ignored
                                 * kludgy, but effective */
            free_vars(vars);
            return cur;
        }
        if (b2 && b2->tok) {
            /* we've identified a nested block in a previous loop.
             * we didnt get an openquote yet, so look for that.
             * we must find this.  keep looking til we do.
             */
            if (*cur != '{') {
                confparse_error("Junk after nested block token", *lnum);
                free_vars(vars);
                return NULL;
            }
            cur++;
            cur = check_quote(cur);
            cur = parse_block(b2, cur, file, lnum);
            b2 = NULL;
            continue;
        }
        if (!item || !item->tok) {
            /* if we dont already have a specific token we're working on
             * find one here.
             */
            cur = check_quote(cur);
            if (BadPtr(cur))
                continue;
            tok = cur;
            tlnum = *lnum;
            if (*cur == '}') {
                /* if we've got a closebracket, then we've hit the end
                 * of our block.
                 */
                done = 1;
                cur++;
                cur = check_quote(cur);
                if (BadPtr(cur))
                    continue;
                if (*cur != ';') {
                    confparse_error("Missing block end semicolon", *lnum);
                    free_vars(vars);
                    return NULL;
                } else
                    cur++;
                if (((*block->func) (vars, *lnum)) == -1) {
                    free_vars(vars);
                    return NULL;
                }
                if (BadPtr(cur))
                    *cur = '#'; /* we cant return a bad pointer because
                                 * that will pull us out of the conf read
                                 * so this will just get ignored
                                 * kludgy, but effective */
                free_vars(vars);
                return cur;

            }
            /* our token ends where whitespace or a semicolon begins */
            while (!BadPtr(cur) && ((*cur != ' ') && (*cur != ';') &&
                                    (*cur != '\t') && (*cur != '\n')))
                cur++;
            if (BadPtr(cur)) {
                confparse_error("Unterminated token", *lnum);
                free_vars(vars);
                return NULL;
            } else {
                if (*cur == ';')
                    skip = 1;
                *cur = '\0';
            }
            cur++;
            if (block->nest) {
                /* we allow nested stuff inside here, so check for it. */
                for (b2 = tconftab; b2->tok; b2++)
                    if (!stricmp(b2->tok, tok))
                        break;
                if (b2 && b2->tok)
                    if (!(block->nest & b2->flag))
                        b2 = NULL;
                if (b2 && b2->tok) {
                    /* recurse through the block we found */
                    tlnum = *lnum;
                    cur = check_quote(cur);
                    if (BadPtr(cur))
                        continue;
                    if (*cur != '{') {
                        confparse_error("Junk after nested block name",
                                        *lnum);
                        free_vars(vars);
                        return NULL;
                    }
                    cur++;
                    cur = check_quote(cur);
                    cur = parse_block(b2, cur, file, lnum);
                    if (!cur) {
                        free_vars(vars);
                        return NULL;
                    }
                    b2 = NULL;
                    continue;
                }
            }
            /* find our token */
            for (item = subconftab; item && item->tok; item++)
                if (!stricmp(item->tok, tok))
                    break;
            if (!item->tok) {
                confparse_error("Unknown token", *lnum);
                alog(LOG_DEBUG, "debug: error token value = %s", tok);
                free_vars(vars);
                return NULL;
            }
            /* create our variable */
            vars[vnum] = (cVar *) malloc(sizeof(cVar));
            memset((char *) vars[vnum], '\0', sizeof(cVar));
            vars[vnum]->type = item;
            vars[vnum]->loaded = 1;
        }
        if (item->var & VARTYPE_NONE) {
            /* we dont need to grab a variable for this type 
             * just look for the closing semicolon, and move on */
            vars[vnum]->loaded = 2;
            if (!skip) {
                /* we've already gotten our semicolon back
                 * at the end of our token.  dont look for it. */
                cur = check_quote(cur);
                while (!BadPtr(cur) && (*cur != ';'))
                    cur++;
                if (BadPtr(cur)) {
                    clear = 1;
                    continue;
                }
                cur++;
            }
            skip = 0;
            vars[vnum]->loaded = 3;
            vnum++;
            item = NULL;
            continue;
        }
        if (item->var & VARTYPE_STRING) {
            /* we're looking for a string here, so we require
             * quotes around the string...
             */
            cur = check_quote(cur);
            while (!BadPtr(cur) && (*cur != '"'))
                cur++;
            if (BadPtr(cur))
                continue;
            cur++;
            var = cur;
            while (!BadPtr(cur) && (*cur != '"'))
                cur++;
            if (BadPtr(cur)) {
                confparse_error("Unterminated quote", *lnum);
                free_vars(vars);
                return NULL;
            }
            *cur = '\0';
            cur++;
            vars[vnum]->value = sstrdup(var);
            vars[vnum]->loaded = 2;
            while (!BadPtr(cur) && (*cur != ';'))
                cur++;
            if (BadPtr(cur)) {
                clear = 1;
                continue;
            }
            cur++;
            vars[vnum]->loaded = 3;
            vnum++;
            item = NULL;
            continue;
        }
        if (item->var & VARTYPE_INT) {
            cur = check_quote(cur);
            var = cur;
            while (!BadPtr(cur) && ((*cur != ';') && (*cur != '\t') &&
                                    (*cur != '\n') && (*cur != ' ')))
                cur++;
            if (BadPtr(cur)) {
                clear = 1;
                continue;
            }
            if (*cur != ';')
                clear = 1;
            *cur = '\0';
            cur++;
            var2 = var;
            while (*var) {
                if (IsDigit(*var))
                    var++;
                else {
                    confparse_error("Expecting integer value", *lnum);
                    free_vars(vars);
                    return NULL;
                }
            }
            if (!item)
                continue;
            var = var2;
            vars[vnum]->value = sstrdup(var);
            vars[vnum]->loaded = 3;
            vnum++;
            item = NULL;
            continue;
        }
        if (item->var & VARTYPE_NAME) {
            cur = check_quote(cur);
            if (!BadPtr(cur) && (*cur == '"'))
                cur++;
            var = cur;
            while (!BadPtr(cur) && (*cur != ';')) {
                if ((*cur == ' ') || (*cur == '"') || (*cur == '\t')) {
                    *cur = '\0';
                    if (vars[vnum]->loaded == 1) {
                        vars[vnum]->value = sstrdup(var);
                        vars[vnum]->loaded = 2;
                    }
                }
                cur++;
            }
            if (BadPtr(cur)) {
                clear = 1;
                continue;
            }
            *cur = '\0';
            cur++;
            if (vars[vnum]->loaded == 1)
                vars[vnum]->value = sstrdup(var);
            vars[vnum]->loaded = 3;
            vnum++;
            item = NULL;
            continue;
        }
        confparse_error("Unexpected EOF:  Syntax Error", tlnum);
        free_vars(vars);
        return NULL;
    }
    confparse_error("Unexpected EOF:  Syntax Error", tlnum);
    free_vars(vars);
    return NULL;
}

/*************************************************************************/

/* Called on startup, check the config file */

int initconf(const char *filename, int reload)
{
    int lnum = 0, blnum = 0, clear = 0;
    char line[LINE_MAX];
    char *cur = NULL;
    char *tok;
    tConf *block = NULL;
    FILE *file;

    current_file = filename;

    if (!(file = fopen(filename, "r"))) {
        if (!reload) {
            log_perror(langstr(CONFIG_ERROR_OPENING), filename);
        } else {
            alog(LOG_ERROR, langstr(CONFIG_ERROR_OPENING), filename);
        }
        return -1;
    }

    while (!BadPtr(cur)
           || ((fgets(line, LINE_MAX + 1, file) != NULL) && ++lnum
               && (cur = line))) {
        cur = check_quote(cur);
        if (BadPtr(cur))
            continue;
        /* now, we should be ok to get that token.. */
        if (!block) {
            tok = cur;
            while ((*cur != ' ') && (*cur != '\n') && (*cur != '{'))
                cur++;          /* find the whitespace following the token */
            if (*cur == '{')
                clear = 1;
            *cur = '\0';
            cur++;
            if (!stricmp("INCLUDE", tok)) {
                /* this is an include - find pull out the file name
                 * and parse this file now
                 */
                char *var;
                cur = check_quote(cur);
                if ((*cur == '"') || *cur == '<')
                    cur++;
                var = cur;
                while ((*cur != ' ') && (*cur != '"') && (*cur != '>') &&
                       (*cur != '\n') && (*cur != ';') && (*cur != '\t'))
                    cur++;
                if (BadPtr(cur)) {
                    confparse_error("Bad include line", lnum);
                    return -1;
                }
                *cur = '\0';
                cur++;
                if (initconf(var, reload) == -1)
                    return -1;
                current_file = filename;        /* reset */
                continue;
            }
            for (block = tconftab; block->tok; block++)
                if (block->tok && tok) {
                    if (!stricmp(block->tok, tok))
                        break;
                }
            if (!block->tok) {
                alog(LOG_DEBUG, "debug: token is %s", tok);
                confparse_error("Unknown block type", lnum);
                return -1;
            }
            blnum = lnum;
        }
        cur = check_quote(cur);
        if (BadPtr(cur))
            continue;
        if ((*cur == '{') || clear)
            cur++;
        else {
            confparse_error("Junk after block name", lnum);
            return -1;
        }
        if ((cur = parse_block(block, cur, file, &lnum)) == NULL) {
            return -1;
        }
        clear = 0;
        block = NULL;
        continue;
    }
    if (clear) {
        confparse_error("Unexpected EOF:  Syntax error", blnum);
        return -1;
    }
    return 1;
}

/*************************************************************************/
