/* StatServ core functions
 *
 * (C) 2003-2005 Anope Team
 * Contact us at info@anope.org
 *
 * Please read COPYING and README for further details.
 *
 * Based on the original code of Epona by Lara.
 * Based on the original code of Services by Andy Church. 
 * 
 * $Id: ss_export.c 427 2005-07-23 08:40:36Z trystan $
 *
 */
/*************************************************************************/

#include "denora.h"

static int do_export(User * u);
int DenoraInit(int argc, char **argv);
void DenoraFini(void);

/**
 * Create the command, and tell anope about it.
 * @param argc Argument count
 * @param argv Argument list
 * @return MOD_CONT to allow the module, MOD_STOP to stop it
 **/
int DenoraInit(int argc, char **argv)
{
    Command *c;

    if (denora->debug >= 2) {
        protocol_debug(NULL, argc, argv);
    }
    moduleAddAuthor("Denora");
    moduleAddVersion("$Id: ss_export.c 427 2005-07-23 08:40:36Z trystan $");
    moduleSetType(CORE);

    c = createCommand("EXPORT",    do_export,     is_stats_admin,      -1, -1,-1, STAT_HELP_EXPORT); 
    moduleAddCommand(STATSERV, c, MOD_UNIQUE);

    return MOD_CONT;
}

/**
 * Unload the module
 **/
void DenoraFini(void)
{

}

/*************************************************************************/

static int do_export(User * u)
{
    char *temp;
    char *temp2 = NULL;
    char *cmd = NULL;
    char *filename = NULL;

    temp = moduleGetLastBuffer();

    if (temp) {
        temp2 = sstrdup(temp);
        cmd = myStrGetTokenRemainder(temp, ' ', 0);
    }

    if (!cmd) {
        syntax_error(s_StatServ, u, "EXPORT", STATS_EXPORT_SYNTAX);
        denora_free(temp2);
        return MOD_CONT;
    }
    if (!stricmp("channels", cmd)) {
        filename = sstrdup("channels.xml");
        xml_export_channels(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_CHANNELS, "channels.xml");
    } else if (!stricmp("users", cmd)) {
        filename = sstrdup("users.xml");
        xml_export_users(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_USERS, "users.xml");
    } else if (!stricmp("tld", cmd)) {
        filename = sstrdup("tld.xml");
        xml_export_tld(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_TLD, "tld.xml");
    } else if (!stricmp("servers", cmd)) {
        filename = sstrdup("servers.xml");
        xml_export_servers(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_SERVERS, "servers.xml");
    } else if (!stricmp("stats", cmd)) {
        filename = sstrdup("stats.xml");
        xml_export_stats(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_STATS, "stats.xml");
    } else if (!stricmp("all", cmd)) {
        filename = sstrdup("denora.xml");
        xml_export_all(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_ALL, "denora.xml");
    } else if (!stricmp("ctcp", cmd)) {
        filename = sstrdup("ctcp.xml");
        xml_export_ctcp(filename);
        notice_lang(s_StatServ, u, STATS_EXPORT_ALL, "ctcp.xml");
    } else {
        syntax_error(s_StatServ, u, "EXPORT", STATS_EXPORT_SYNTAX);
    }
    denora_free(temp);
    denora_free(temp2);
    denora_free(cmd);
    denora_free(filename);
    return MOD_CONT;
}
