/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: events.c 444 2005-08-02 02:28:27Z trystan $
 *
 */

#include "denora.h"

#if defined(USE_MODULES) && !defined(_WIN32)
#include <dlfcn.h>
/* Define these for systems without them */
#ifndef RTLD_NOW
#define RTLD_NOW 0
#endif
#ifndef RTLD_LAZY
#define RTLD_LAZY RTLD_NOW
#endif
#ifndef RTLD_GLOBAL
#define RTLD_GLOBAL 0
#endif
#ifndef RTLD_LOCAL
#define RTLD_LOCAL 0
#endif
#endif

#ifdef _WIN32
const char *deno_moderr(void);
#endif

/*************************************************************************/

/**
 * Declare all the list's we want to use here
 **/
CommandHash *STATSERV[MAX_CMD_HASH];
MessageHash *IRCD[MAX_CMD_HASH];
ModuleHash *MODULE_HASH[MAX_CMD_HASH];
EvtMessageHash *EVENT[MAX_CMD_HASH];
EvtHookHash *EVENTHOOKS[MAX_CMD_HASH];

Module *mod_current_module;
char *mod_current_module_name = NULL;
char *mod_current_buffer = NULL;
int mod_current_op;
User *mod_current_user;
ModuleCallBack *moduleCallBackHead = NULL;
char *mod_current_evtbuffer = NULL;

/*************************************************************************/

/**
 * Automaticaly load modules at startup.
 * This will load modules at startup before the IRCD link is attempted, this
 * allows admins to have a module relating to ircd support load
 */
void modules_init(void)
{
#ifdef USE_MODULES
    int idx;
    Module *m;
    Conf_Modules *amodules = modules;

    SET_SEGV_LOCATION();

    for (idx = 0; idx < ModulesNumber; idx++) {
        m = findModule(amodules->autoload[idx]);
        if (!m) {
            m = createModule(amodules->autoload[idx]);
            mod_current_module = m;
            mod_current_user = NULL;
            alog(LOG_NORMAL, langstr(ALOG_MOD_TRY_LOAD),
                 mod_current_module->name);
            alog(LOG_NORMAL, langstr(ALOG_MOD_STATUS),
                 loadModule(mod_current_module, NULL));
            mod_current_module = NULL;
            mod_current_user = NULL;
        }
    }
#endif
}

/*************************************************************************/

#ifdef _WIN32
void Win32LoadCoreDir(char *dirname)
{
#ifdef USE_MODULES
    BOOL fFinished;
    HANDLE hList;
    TCHAR szDir[MAX_PATH + 1];
    TCHAR szSubDir[MAX_PATH + 1];
    WIN32_FIND_DATA FileData;
    char *mname = NULL;
    Module *m;
    int status;
    char buffer[_MAX_PATH];

    /* Get the current working directory: */
    if (_getcwd(buffer, _MAX_PATH) == NULL) {
        alog(LOG_DEBUG, "debug: Unable to set Current working directory");
    }
    sprintf(szDir, "%s\\%s\\*", buffer, dirname);

    hList = FindFirstFile(szDir, &FileData);
    if (hList != INVALID_HANDLE_VALUE) {
        fFinished = FALSE;
        while (!fFinished) {
            if (!(FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
                mname = sstrdup(FileData.cFileName);
                mname[strlen(FileData.cFileName) - 4] = '\0';
                m = findModule(mname);
                if (!m) {
                    m = createCoreModule(mname);
                    mod_current_module = m;
                    mod_current_user = NULL;
                    alog(LOG_DEBUG, "trying to load [%s]",
                         mod_current_module->name);
                    status = loadCoreModule(mod_current_module, NULL);
                    alog(LOG_DEBUG, "status: [%d]", status);
                    mod_current_module = NULL;
                    mod_current_user = NULL;
                }
                free(mname);
            }
            if (!FindNextFile(hList, &FileData)) {
                if (GetLastError() == ERROR_NO_MORE_FILES) {
                    fFinished = TRUE;
                }
            }
        }
    } else {
        printf("Invalid File Handle. GetLastError reports %d\n",
               GetLastError());
    }
    FindClose(hList);
#endif
}
#endif

/*************************************************************************/

/**
 * Automaticaly load modules at startup.
 * This will load modules at startup before the IRCD link is attempted, this
 * allows admins to have a module relating to ircd support load
 */
void core_modules_init(void)
{
#ifdef USE_MODULES
    Module *m;
    char *mname = NULL;
#ifndef _WIN32
    DIR *dp;
    struct direct *dir;
#endif
    char buf[BUFSIZE];
    int status;

#ifndef _WIN32
    snprintf(buf, BUFSIZE, "%s/core", denora->dir);
#else
    snprintf(buf, BUFSIZE, "\\%s", "core");
#endif

#ifndef _WIN32
    if ((dp = opendir(buf)) == NULL) {
        alog(LOG_ERROR, "ERROR: cannot open directory (%s)", buf);
        return;
    }

    while ((dir = readdir(dp)) != NULL) {
        if (dir->d_ino == 0) {
            continue;
        }
        if (!stricmp(dir->d_name, ".") || !stricmp(dir->d_name, "..")
            || !stricmp(dir->d_name, "runtime")) {
            continue;
        }
        mname = sstrdup(dir->d_name);
        mname[strlen(dir->d_name) - 3] = '\0';
        m = findModule(mname);
        if (!m) {
            m = createCoreModule(mname);
            mod_current_module = m;
            mod_current_user = NULL;
            alog(LOG_DEBUG, "trying to load [%s]",
                 mod_current_module->name);
            status = loadCoreModule(mod_current_module, NULL);
            alog(LOG_DEBUG, "status: [%d]", status);
            mod_current_module = NULL;
            mod_current_user = NULL;
        }
        free(mname);
    }
    closedir(dp);
#else
    Win32LoadCoreDir(buf);
#endif
#endif
}

/*************************************************************************/

/**
 * Loads a given module.
 * @param m the module to load
 * @param u the user who loaded it, NULL for auto-load
 * @return MOD_ERR_OK on success, anything else on fail
 */
int loadCoreModule(Module * m, User * u)
{
#ifdef USE_MODULES
    char buf[4096];
    int len;
    const char *err;
    int (*func) (int, char **);
    int ret = 0;
    char *argv[1];
    int argc = 0;

    Module *m2;
    if (!m || !m->name) {
        return MOD_ERR_PARAMS;
    }
    if (m->handle) {
        return MOD_ERR_EXISTS;
    }
    if ((m2 = findModule(m->name)) != NULL) {
        return MOD_ERR_EXISTS;
    }
#ifndef _WIN32
    snprintf(buf, BUFSIZE, "%s/core/", denora->dir);    /* Get full path with module extension */
#else
    snprintf(buf, BUFSIZE, "%s\\core\\", denora->dir);  /* Get full path with module extension */
#endif

    len = strlen(buf);
#ifndef _WIN32
    strncat(buf, "runtime/", 4095 - len);
#else
    strncat(buf, "runtime\\", 4095 - len);
#endif
    len = strlen(buf);
    strncat(buf, m->name, 4095 - len);
    len = strlen(buf);
    strncat(buf, MODULE_EXT, 4095 - len);
    buf[4095] = '\0';

    m->filename = sstrdup(buf);
    moduleCopyCoreFile(m->name);        /* We have to use the full filename */

    deno_modclearerr();
    m->handle = deno_modopen(m->filename);
    if ((err = deno_moderr()) != NULL) {
        alog(LOG_NORMAL, langstr(ALOG_MOD_ERROR), err);
        if (u) {
            notice_lang(s_StatServ, u, STAT_MODULE_LOAD_FAIL, m->name);
        }
        return MOD_ERR_NOLOAD;
    }
    deno_modclearerr();
    func = deno_modsym(m->handle, "DenoraInit");
    if ((err = deno_moderr()) != NULL) {
        deno_modclose(m->handle);       /* If no DenoraInit - it isnt an Denora Module, close it */
        return MOD_ERR_NOLOAD;
    }
    if (func) {
        mod_current_module_name = m->name;
        /* argv[0] is the user if there was one, or NULL if not */
        if (u) {
            argv[0] = sstrdup(u->nick);
        } else {
            argv[0] = NULL;
        }
        argc++;

        ret = func(argc, argv); /* exec AhearnInit */
        if (u) {
            free(argv[0]);
        }
        if (ret == MOD_STOP) {
            alog(LOG_DEBUG, "%s requested unload...", m->name);
            unloadModule(m, NULL);
            mod_current_module_name = NULL;
            return MOD_ERR_NOLOAD;
        }

        mod_current_module_name = NULL;
    }

    addModule(m);
    return MOD_ERR_OK;

#else
    return MOD_ERR_NOLOAD;
#endif
}

/*************************************************************************/

int moduleCopyCoreFile(char *name)
{
#ifdef USE_MODULES
    int ch;
    FILE *source, *target;
    char output[4096];
    char input[4096];
    int len;
#ifdef _WIN32
    char buffer[_MAX_PATH];
    char win32filename[MAXPATHLEN];

    /* Get the current working directory: */
    if (_getcwd(buffer, _MAX_PATH) == NULL) {
        alog(LOG_DEBUG, "debug: Unable to set Current working directory");
    }
#endif
#ifndef _WIN32
    snprintf(output, BUFSIZE, "%s/core/", denora->dir); /* Get full path with module extension */
#else
    snprintf(output, BUFSIZE, "%s\\core\\", buffer);    /* Get full path with module extension */
#endif
#ifndef _WIN32
    snprintf(input, BUFSIZE, "%s/core/", denora->dir);  /* Get full path with module extension */
#else
    snprintf(input, BUFSIZE, "%s\\core\\", buffer);     /* Get full path with module extension */
#endif

    len = strlen(output);
#ifdef _WIN32
    strncat(output, "runtime\\", 4095 - len);
#else
    strncat(output, "runtime/", 4095 - len);
#endif
    len += strlen(output);
    strncat(output, name, 4095 - len);
    strncat(input, name, 4095 - len);
    len += strlen(output);
    strncat(output, MODULE_EXT, 4095 - len);
    strncat(input, MODULE_EXT, 4095 - len);

#ifndef _WIN32
    if ((source = fopen(input, "r")) == NULL) {
#else
    if ((source = fopen(input, "rb")) == NULL) {
#endif
        alog(LOG_ERROR, "ERROR: the file (%s) does not exist", input);
        return MOD_ERR_NOEXIST;
    }
#ifndef _WIN32
    if ((target = fopen(output, "w")) == NULL) {
#else
    if ((target = fopen(output, "wb")) == NULL) {
#endif
        alog(LOG_ERROR, "ERROR: the file (%s) can not be open for writing",
             output);
        return MOD_ERR_FILE_IO;
    }
    while ((ch = fgetc(source)) != EOF) {
        fputc(ch, target);
    }
    fclose(source);
    if (fclose(target) != 0) {
        return MOD_ERR_FILE_IO;
    }
#endif
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Automaticaly load modules at startup, delayed.
 * This function waits until the IRCD link has been made, and then attempts
 * to load the specified modules.
 */
void modules_delayed_init(void)
{
#ifdef USE_MODULES
    int idx;
    Module *m;
    Conf_Modules *amodules = modules;

    SET_SEGV_LOCATION();

    for (idx = 0; idx < ModulesDelayedNumber; idx++) {
        m = findModule(amodules->delayed[idx]);
        if (!m) {
            m = createModule(amodules->delayed[idx]);
            mod_current_module = m;
            mod_current_user = NULL;
            alog(LOG_NORMAL, langstr(ALOG_MOD_TRY_LOAD),
                 mod_current_module->name);
            alog(LOG_NORMAL, langstr(ALOG_MOD_STATUS),
                 loadModule(mod_current_module, NULL));
            mod_current_module = NULL;
            mod_current_user = NULL;
        }
    }
#endif
}

/*************************************************************************/

/**
 * Create a new module, setting up the default values as needed.
 * @param filename the filename of the new module
 * @return a newly created module struct
 */
Module *createModule(char *filename)
{
    Module *m;
    if (!filename) {
        return NULL;
    }

    SET_SEGV_LOCATION();

    if ((m = malloc(sizeof(Module))) == NULL) {
        fatal("Out of memory!");
    }
    m->name = sstrdup(filename);        /* Our Name */
    m->handle = NULL;           /* Handle */
    m->version = NULL;
    m->author = NULL;
    m->operHelp = NULL;
    m->type = THIRD;

    return m;                   /* return a nice new module */
}

/*************************************************************************/

/**
 * Create a new module, setting up the default values as needed.
 * @param filename the filename of the new module
 * @return a newly created module struct
 */
Module *createCoreModule(char *filename)
{
    Module *m;
    if (!filename) {
        return NULL;
    }

    SET_SEGV_LOCATION();

    if ((m = malloc(sizeof(Module))) == NULL) {
        fatal("Out of memory!");
    }
    m->name = sstrdup(filename);        /* Our Name */
    m->handle = NULL;           /* Handle */
    m->version = NULL;
    m->author = NULL;
    m->operHelp = NULL;
    m->type = CORE;

    return m;                   /* return a nice new module */
}

/*************************************************************************/

/**
 * Destroy the module.
 * free up all memory used by our module struct.
 * @param m the module to free
 * @return MOD_ERR_OK on success, anything else on fail
 */
int destroyModule(Module * m)
{
    SET_SEGV_LOCATION();

    if (!m) {
        return MOD_ERR_PARAMS;
    }
    if (m->name) {
        denora_free(m->name);
    }
    if (m->filename) {
        remove(m->filename);
        denora_free(m->filename);
    }
    m->handle = NULL;
    if (m->author) {
        denora_free(m->author);
    }
    if (m->version) {
        denora_free(m->version);
    }
    /* No need to free our cmd/msg list, as they will always be empty by the module is destroyed */
    denora_free(m);
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Add the module to the list of currently loaded modules.
 * @param m the currently loaded module
 * @return MOD_ERR_OK on success, anything else on fail
 */
int addModule(Module * m)
{
    int modindex = 0;
    ModuleHash *current = NULL;
    ModuleHash *newHash = NULL;
    ModuleHash *lastHash = NULL;

    SET_SEGV_LOCATION();

    modindex = CMD_HASH(m->name);

    for (current = MODULE_HASH[modindex]; current; current = current->next) {
        if (stricmp(m->name, current->name) == 0)
            return MOD_ERR_EXISTS;
        lastHash = current;
    }

    if ((newHash = malloc(sizeof(ModuleHash))) == NULL) {
        fatal("Out of memory");
    }
    m->time = time(NULL);
    newHash->next = NULL;
    newHash->name = sstrdup(m->name);
    newHash->m = m;

    if (lastHash == NULL)
        MODULE_HASH[modindex] = newHash;
    else
        lastHash->next = newHash;
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Remove the module from the list of loaded modules.
 * @param m module to remove
 * @return MOD_ERR_OK on success anything else on fail
 */
int delModule(Module * m)
{
    int modindex = 0;
    ModuleHash *current = NULL;
    ModuleHash *lastHash = NULL;

    SET_SEGV_LOCATION();

    if (!m) {
        return MOD_ERR_PARAMS;
    }

    modindex = CMD_HASH(m->name);

    for (current = MODULE_HASH[modindex]; current; current = current->next) {
        if (stricmp(m->name, current->name) == 0) {
            if (!lastHash) {
                MODULE_HASH[modindex] = current->next;
            } else {
                lastHash->next = current->next;
            }
            destroyModule(current->m);
            denora_free(current->name);
            denora_free(current);
            return MOD_ERR_OK;
        }
        lastHash = current;
    }
    return MOD_ERR_NOEXIST;
}

/*************************************************************************/

/**
 * Search the list of loaded modules for the given name.
 * @param name the name of the module to find
 * @return a pointer to the module found, or NULL
 */
Module *findModule(char *name)
{
    int idx;
    ModuleHash *current = NULL;
    if (!name) {
        return NULL;
    }
    idx = CMD_HASH(name);

    SET_SEGV_LOCATION();

    for (current = MODULE_HASH[idx]; current; current = current->next) {
        if (stricmp(name, current->name) == 0) {
            return current->m;
        }
    }
    return NULL;

}

/*************************************************************************/

/** 
 * Copy the module from the modules folder to the runtime folder.
 * This will prevent module updates while the modules is loaded from
 * triggering a segfault, as the actaul file in use will be in the 
 * runtime folder.
 * @param name the name of the module to copy
 * @return MOD_ERR_OK on success
 */
int moduleCopyFile(char *name)
{
#ifdef USE_MODULES
    int ch;
    FILE *source, *target;
    char output[4096];
    char input[4096];
    int len;

    SET_SEGV_LOCATION();

    strncpy(output, MODULE_PATH, 4095); /* Get full path with .so extension */
    strncpy(input, MODULE_PATH, 4095);  /* Get full path with .so extension */
    len = strlen(output);
#ifdef _WIN32
    strncat(output, "runtime\\", 4095 - len);
#else
    strncat(output, "runtime/", 4095 - len);
#endif
    len += strlen(output);
    strncat(output, name, 4095 - len);
    strncat(input, name, 4095 - len);
    len += strlen(output);
    strncat(output, MODULE_EXT, 4095 - len);
    strncat(input, MODULE_EXT, 4095 - len);

#ifndef _WIN32
    if ((source = fopen(input, "r")) == NULL) {
#else
    if ((source = fopen(input, "rb")) == NULL) {
#endif
        return MOD_ERR_NOEXIST;
    }
#ifndef _WIN32
    if ((target = fopen(output, "w")) == NULL) {
#else
    if ((target = fopen(output, "wb")) == NULL) {
#endif
        return MOD_ERR_FILE_IO;
    }
    while ((ch = fgetc(source)) != EOF) {
        fputc(ch, target);
    }
    fclose(source);
    if (fclose(target) != 0) {
        return MOD_ERR_FILE_IO;
    }
#endif
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Loads a given module.
 * @param m the module to load
 * @param u the user who loaded it, NULL for auto-load
 * @return MOD_ERR_OK on success, anything else on fail
 */
int loadModule(Module * m, User * u)
{
#ifdef USE_MODULES
    char buf[4096];
    int len;
    const char *err;
    int (*func) (int, char **);
    int ret = 0;
    char *argv[1];
    int argc = 0;

    Module *m2;
    if (!m || !m->name) {
        return MOD_ERR_PARAMS;
    }
    if (m->handle) {
        return MOD_ERR_EXISTS;
    }
    if ((m2 = findModule(m->name)) != NULL) {
        return MOD_ERR_EXISTS;
    }

    moduleCopyFile(m->name);

    strncpy(buf, MODULE_PATH, 4095);    /* Get full path with module extension */
    len = strlen(buf);
#ifndef _WIN32
    strncat(buf, "runtime/", 4095 - len);
#else
    strncat(buf, "runtime\\", 4095 - len);
#endif
    len = strlen(buf);
    strncat(buf, m->name, 4095 - len);
    len = strlen(buf);
    strncat(buf, MODULE_EXT, 4095 - len);
    buf[4095] = '\0';

    m->filename = sstrdup(buf);
    deno_modclearerr();
    m->handle = deno_modopen(m->filename);
    if ((err = deno_moderr()) != NULL) {
        alog(LOG_NORMAL, langstr(ALOG_MOD_ERROR), err);
        if (u) {
            notice_lang(s_StatServ, u, STAT_MODULE_LOAD_FAIL, m->name);
        }
        return MOD_ERR_NOLOAD;
    }
    deno_modclearerr();
    func = deno_modsym(m->handle, "DenoraInit");
    if ((err = deno_moderr()) != NULL) {
        deno_modclose(m->handle);       /* If no DenoraInit - it isnt an Denora Module, close it */
        return MOD_ERR_NOLOAD;
    }
    if (func) {
        mod_current_module_name = m->name;
        /* argv[0] is the user if there was one, or NULL if not */
        if (u) {
            argv[0] = sstrdup(u->nick);
        } else {
            argv[0] = NULL;
        }
        argc++;

        ret = func(argc, argv); /* exec DenoraInit */
        if (u) {
            denora_free(argv[0]);
        }
        if (m->type == PROTOCOL && protocolModuleLoaded()) {
            alog(LOG_NORMAL, langstr(ALOG_MOD_BE_ONLY_ONE));
            ret = MOD_STOP;
        }
        if (ret == MOD_STOP) {
            alog(LOG_DEBUG, langstr(ALOG_MOD_UNLOAD_SELF), m->name);
            unloadModule(m, NULL);
            mod_current_module_name = NULL;
            return MOD_ERR_NOLOAD;
        }

        mod_current_module_name = NULL;
    }

    if (u) {
        denora_cmd_global(s_StatServ, langstring(GLOBAL_LOAD_MOD), u->nick,
                          m->name);
        notice_lang(s_StatServ, u, STAT_MODULE_LOADED, m->name);
    }
    addModule(m);
    return MOD_ERR_OK;

#else
    return MOD_ERR_NOLOAD;
#endif
}

/*************************************************************************/

/**
 * Unload the given module.
 * @param m the module to unload
 * @param u the user who unloaded it
 * @return MOD_ERR_OK on success, anything else on fail
 */
int unloadModule(Module * m, User * u)
{
#ifdef USE_MODULES
    void (*func) (void);

    if (!m || !m->handle) {
        if (u) {
            notice_lang(s_StatServ, u, STAT_MODULE_REMOVE_FAIL, m->name);
        }
        return MOD_ERR_PARAMS;
    }

    if (m->type == PROTOCOL) {
        if (u) {
            notice_lang(s_StatServ, u, STAT_MODULE_NO_UNLOAD);
        }
        return MOD_ERR_NOUNLOAD;
    }

    if (prepForUnload(mod_current_module) != MOD_ERR_OK) {
        return MOD_ERR_UNKNOWN;
    }

    func = deno_modsym(m->handle, "DenoraFini");
    if (func) {
        func();                 /* exec DenoraFini */
    }

    if ((deno_modclose(m->handle)) != 0) {
        alog(LOG_ERROR, langstr(ALOG_MOD_ERROR), deno_moderr());
        if (u) {
            notice_lang(s_StatServ, u, STAT_MODULE_REMOVE_FAIL, m->name);
        }
        return MOD_ERR_NOUNLOAD;
    } else {
        if (u) {
            denora_cmd_global(s_StatServ, langstring(GLOBAL_UNLOAD_MOD),
                              u->nick, m->name);
            notice_lang(s_StatServ, u, STAT_MODULE_UNLOADED, m->name);
        }
        delModule(m);
        return MOD_ERR_OK;
    }
#else
    return MOD_ERR_NOUNLOAD;
#endif
}

/*************************************************************************/

/**
 * Prepare a module to be unloaded.
 * Remove all commands and messages this module is providing, and delete 
 * any callbacks which are still pending.
 * @param m the module to prepare for unload
 * @return MOD_ERR_OK on success
 */
int prepForUnload(Module * m)
{
    int idx;
    CommandHash *current = NULL;
    MessageHash *mcurrent = NULL;
    EvtMessageHash *ecurrent = NULL;
    EvtHookHash *ehcurrent = NULL;
    Command *c;
    EvtMessage *eMsg;
    EvtHook *eHook;
    Message *msg;
    int status = 0;

    SET_SEGV_LOCATION();

    if (!m) {
        return MOD_ERR_PARAMS;
    }

    /* Kill any active callbacks this module has */
    moduleCallBackPrepForUnload(m->name);

    /* Remove any stored data this module has */
    moduleDelAllDataMod(m);

    /**
     * ok, im going to walk every hash looking for commands we own, now, not exactly elegant or efficiant :)
     **/

    for (idx = 0; idx < MAX_CMD_HASH; idx++) {
        for (current = OS_cmdTable[idx]; current; current = current->next) {
            for (c = current->c; c; c = c->next) {
                if ((c->mod_name) && (stricmp(c->mod_name, m->name) == 0)) {
                    moduleDelCommand(STATSERV, c->name);
                }
            }
        }

        for (mcurrent = IRCD[idx]; mcurrent; mcurrent = mcurrent->next) {
            for (msg = mcurrent->m; msg; msg = msg->next) {
                if ((msg->mod_name)
                    && (stricmp(msg->mod_name, m->name) == 0)) {
                    moduleDelMessage(msg->name);
                }
            }
        }


        for (ecurrent = EVENT[idx]; ecurrent; ecurrent = ecurrent->next) {
            for (eMsg = ecurrent->evm; eMsg; eMsg = eMsg->next) {
                if ((eMsg->mod_name)
                    && (stricmp(eMsg->mod_name, m->name) == 0)) {
                    status = delEventHandler(EVENT, eMsg, m->name);
                }
            }
        }
        for (ehcurrent = EVENTHOOKS[idx]; ehcurrent;
             ehcurrent = ehcurrent->next) {
            for (eHook = ehcurrent->evh; eHook; eHook = eHook->next) {
                if ((eHook->mod_name)
                    && (stricmp(eHook->mod_name, m->name) == 0)) {
                    status = delEventHook(EVENTHOOKS, eHook, m->name);
                }
            }
        }
    }
    return MOD_ERR_OK;
}

/*******************************************************************************
 * Command Functions
 *******************************************************************************/
/**
 * Create a Command struct ready for use in denora.
 * @param name the name of the command
 * @param func pointer to the function to execute when command is given
 * @param has_priv pointer to function to check user priv's
 * @param help_all help file index for all users
 * @param help_reg help file index for all regustered users
 * @param help_oper help file index for all opers
 * @param help_admin help file indenx for all stats admins
 * @return a "ready to use" Command struct will be returned
 */
Command *createCommand(const char *name, int (*func) (User * u),
                       int (*has_priv) (User * u), int help_all,
                       int help_reg, int help_oper, int help_admin)
{
    Command *c;
    if (!name || !*name) {
        return NULL;
    }

    SET_SEGV_LOCATION();

    if ((c = malloc(sizeof(Command))) == NULL) {
        fatal("Out of memory!");
    }
    c->name = sstrdup(name);
    c->routine = func;
    c->has_priv = has_priv;
    c->helpmsg_all = help_all;
    c->helpmsg_reg = help_reg;
    c->helpmsg_admin = help_admin;
    c->helpmsg_oper = help_oper;
    c->help_param1 = NULL;
    c->help_param2 = NULL;
    c->help_param3 = NULL;
    c->help_param4 = NULL;
    c->next = NULL;
    c->mod_name = NULL;
    c->service = NULL;
    c->all_help = NULL;
    c->regular_help = NULL;
    c->admin_help = NULL;
    return c;
}

/*************************************************************************/

/**
 * Destroy a command struct freeing any memory.
 * @param c Command to destroy
 * @return MOD_ERR_OK on success, anything else on fail
 */
int destroyCommand(Command * c)
{
    SET_SEGV_LOCATION();

    if (!c) {
        return MOD_ERR_PARAMS;
    }
    if (c->core == 1 && !denora->delayed_quit) {
        return MOD_ERR_UNKNOWN;
    }
    if (c->name) {
        denora_free(c->name);
    }
    c->routine = NULL;
    c->has_priv = NULL;
    c->helpmsg_all = -1;
    c->helpmsg_reg = -1;
    c->helpmsg_admin = -1;
    if (c->help_param1) {
        denora_free(c->help_param1);
    }
    if (c->help_param2) {
        denora_free(c->help_param2);
    }
    if (c->help_param3) {
        denora_free(c->help_param3);
    }
    if (c->help_param4) {
        denora_free(c->help_param4);
    }
    if (c->mod_name) {
        denora_free(c->mod_name);
    }
    if (c->service) {
        denora_free(c->service);
    }
    c->next = NULL;
    denora_free(c);
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Add a CORE command ot the given command hash
 * @param cmdTable the command table to add the command to
 * @param c the command to add
 * @return MOD_ERR_OK on success
 */
int addCoreCommand(CommandHash * cmdTable[], Command * c)
{
    SET_SEGV_LOCATION();

    if (!cmdTable || !c) {
        return MOD_ERR_PARAMS;
    }
    c->core = 1;
    c->next = NULL;
    return addCommand(cmdTable, c, 0);
}

/*************************************************************************/

/**
 * Add a module provided command to the given service.
 * e.g. moduleAddCommand(NICKSERV,c,MOD_HEAD);
 * @param cmdTable the services to add the command to
 * @param c the command to add
 * @param pos the position to add to, MOD_HEAD, MOD_TAIL, MOD_UNIQUE
 * @see createCommand
 * @return MOD_ERR_OK on successfully adding the command
 */
int moduleAddCommand(CommandHash * cmdTable[], Command * c, int pos)
{
    int status;

    SET_SEGV_LOCATION();

    if (!cmdTable || !c) {
        return MOD_ERR_PARAMS;
    }
    /* ok, this appears to be a module adding a command from outside
     * of DenoraInit, try to look up its module struct for it */
    if ((mod_current_module_name) && (!mod_current_module)) {
        mod_current_module = findModule(mod_current_module_name);
    }

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }                           /* shouldnt happen */
    c->core = 0;
    if (!c->mod_name) {
        c->mod_name = sstrdup(mod_current_module->name);
    }


    if (cmdTable == STATSERV) {
        if (s_StatServ) {
            c->service = sstrdup(s_StatServ);
        } else {
            return MOD_ERR_NOSERVICE;
        }
    } else
        c->service = sstrdup("Unknown");

    if (denora->debug) {
        displayCommandFromHash(cmdTable, c->name);
    }
    status = addCommand(cmdTable, c, pos);
    if (denora->debug) {
        displayCommandFromHash(cmdTable, c->name);
    }
    if (status != MOD_ERR_OK) {
        alog(LOG_NORMAL, "ERROR! [%d]", status);
    }
    return status;
}

/*************************************************************************/

/**
 * Delete a command from the service given.
 * @param cmdTable the cmdTable for the services to remove the command from
 * @param name the name of the command to delete from the service
 * @return returns MOD_ERR_OK on success
 */
int moduleDelCommand(CommandHash * cmdTable[], char *name)
{
    Command *c = NULL;
    Command *cmd = NULL;
    int status = 0;

    SET_SEGV_LOCATION();

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }

    c = findCommand(cmdTable, name);
    if (!c) {
        return MOD_ERR_NOEXIST;
    }


    for (cmd = c; cmd; cmd = cmd->next) {
        if (cmd->mod_name
            && stricmp(cmd->mod_name, mod_current_module->name) == 0) {
            if (denora->debug) {
                displayCommandFromHash(cmdTable, name);
            }
            status = delCommand(cmdTable, cmd, mod_current_module->name);
            if (denora->debug) {
                displayCommandFromHash(cmdTable, name);
            }
        }
    }
    return status;
}

/*************************************************************************/

/**
 * Output the command stack into the log files.
 * This will print the call-stack for a given command into the log files, very useful for debugging.
 * @param cmdTable the command table to read from
 * @param name the name of the command to print
 * @return 0 is returned, it has no relevence yet :)
 */
int displayCommandFromHash(CommandHash * cmdTable[], char *name)
{
    CommandHash *current = NULL;
    int modindex = 0;
    modindex = CMD_HASH(name);

    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, langstr(ALOG_CMD_DISPLAY_START), name);

    for (current = cmdTable[modindex]; current; current = current->next) {
        if (stricmp(name, current->name) == 0) {
            displayCommand(current->c);
        }
    }
    alog(LOG_EXTRADEBUG, langstr(ALOG_CMD_DISPLAY_END), name);
    return 0;
}

/*************************************************************************/

/**
 * Output the command stack into the log files.
 * This will print the call-stack for a given command into the log files, very useful for debugging.
 * @param c the command struct to print
 * @return 0 is returned, it has no relevence yet :)
 */

int displayCommand(Command * c)
{
    Command *cmd = NULL;
    int i = 0;

    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_CMD_DISPLAY_LIST), c->name);
    for (cmd = c; cmd; cmd = cmd->next) {
        alog(LOG_DEBUG, "%d:  0x%p", ++i, (void *) cmd);
    }
    alog(LOG_DEBUG, "debug: end");
    return 0;
}

/*************************************************************************/

/**
 * Display the message call stak.
 * Prints the call stack for a message based on the message name, again useful for debugging and little lese :)
 * @param name the name of the message to print info for
 * @return the return int has no relevence atm :)
 */
int displayMessageFromHash(char *name)
{
    MessageHash *current = NULL;
    int modindex = 0;
    modindex = CMD_HASH(name);

    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, langstr(ALOG_MSG_DISPLAY_START), name);
    for (current = IRCD[modindex]; current; current = current->next) {
        if (stricmp(name, current->name) == 0) {
            displayMessage(current->m);
        }
    }
    alog(LOG_EXTRADEBUG, langstr(ALOG_MSG_DISPLAY_END), name);
    return 0;
}

/*************************************************************************/

/**
 * Displays a message list for a given message.
 * Again this is of little use other than debugging.
 * @param m the message to display
 * @return 0 is returned and has no meaning 
 */
int displayMessage(Message * m)
{
    Message *msg = NULL;
    int i = 0;

    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, langstr(ALOG_MSG_DISPLAY_LIST), m->name);
    for (msg = m; msg; msg = msg->next) {
        alog(LOG_DEBUG, "%d: 0x%p", ++i, (void *) msg);
    }
    alog(LOG_DEBUG, "debug: end");
    return 0;
}

/*************************************************************************/

/**
 * Add a command to a command table.
 * only add if were unique, pos = 0;
 * if we want it at the "head" of that command, pos = 1
 * at the tail, pos = 2
 * @param cmdTable the table to add the command to
 * @param c the command to add
 * @param pos the position in the cmd call stack to add the command
 * @return MOD_ERR_OK will be returned on success.
 */
int addCommand(CommandHash * cmdTable[], Command * c, int pos)
{
    /* We can assume both param's have been checked by this point.. */
    int modindex = 0;
    CommandHash *current = NULL;
    CommandHash *newHash = NULL;
    CommandHash *lastHash = NULL;
    Command *tail = NULL;

    SET_SEGV_LOCATION();

    if (!cmdTable || !c || (pos < 0 || pos > 2)) {
        return MOD_ERR_PARAMS;
    }

    if (mod_current_module_name && !c->mod_name) {
        return MOD_ERR_NO_MOD_NAME;
    }

    modindex = CMD_HASH(c->name);

    for (current = cmdTable[modindex]; current; current = current->next) {
        if ((c->service) && (current->c) && (current->c->service)
            && (!strcmp(c->service, current->c->service) == 0)) {
            continue;
        }
        if ((stricmp(c->name, current->name) == 0)) {   /* the cmd exist's we are a addHead */
            if (pos == 1) {
                c->next = current->c;
                current->c = c;
                alog(LOG_DEBUG,
                     "debug: existing cmd: (0x%p), new cmd (0x%p)",
                     (void *) c->next, (void *) c);
                return MOD_ERR_OK;
            } else if (pos == 2) {

                tail = current->c;
                while (tail->next)
                    tail = tail->next;
                alog(LOG_DEBUG,
                     "debug: existing cmd: (0x%p), new cmd (0x%p)",
                     (void *) tail, (void *) c);
                tail->next = c;
                c->next = NULL;

                return MOD_ERR_OK;
            } else
                return MOD_ERR_EXISTS;
        }
        lastHash = current;
    }

    if ((newHash = malloc(sizeof(CommandHash))) == NULL) {
        fatal("Out of memory");
    }
    newHash->next = NULL;
    newHash->name = sstrdup(c->name);
    newHash->c = c;

    if (lastHash == NULL)
        cmdTable[modindex] = newHash;
    else
        lastHash->next = newHash;

    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Remove a command from the command hash.
 * @param cmdTable the command table to remove the command from
 * @param c the command to remove
 * @param mod_name the name of the module who owns the command
 * @return MOD_ERR_OK will be returned on success
 */
int delCommand(CommandHash * cmdTable[], Command * c, char *mod_name)
{
    int modindex = 0;
    CommandHash *current = NULL;
    CommandHash *lastHash = NULL;
    Command *tail = NULL, *last = NULL;

    SET_SEGV_LOCATION();

    if (!c || !cmdTable) {
        return MOD_ERR_PARAMS;
    }

    modindex = CMD_HASH(c->name);
    for (current = cmdTable[modindex]; current; current = current->next) {
        if (stricmp(c->name, current->name) == 0) {
            if (!lastHash) {
                tail = current->c;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                current->c = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    cmdTable[modindex] = current->next;
                    denora_free(current->name);
                    return MOD_ERR_OK;
                }
            } else {
                tail = current->c;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                current->c = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    lastHash->next = current->next;
                    denora_free(current->name);
                    return MOD_ERR_OK;
                }
            }
        }
        lastHash = current;
    }
    return MOD_ERR_NOEXIST;
}

/*************************************************************************/

/**
 * Search the command table gieven for a command.
 * @param cmdTable the name of the command table to search
 * @param name the name of the command to look for
 * @return returns a pointer to the found command struct, or NULL
 */
Command *findCommand(CommandHash * cmdTable[], const char *name)
{
    int idx;
    CommandHash *current = NULL;

    SET_SEGV_LOCATION();

    if (!cmdTable || !name) {
        return NULL;
    }

    idx = CMD_HASH(name);

    for (current = cmdTable[idx]; current; current = current->next) {
        if (stricmp(name, current->name) == 0) {
            return current->c;
        }
    }
    return NULL;
}

/*******************************************************************************
 * Message Functions
 *******************************************************************************/

 /**
  * Create a new Message struct.
  * @param name the name of the message
  * @param func a pointer to the function to call when we recive this message
  * @return a new Message object
  **/
Message *createMessage(const char *name,
                       int (*func) (char *source, int ac, char **av))
{
    Message *m = NULL;

    SET_SEGV_LOCATION();

    if (!name || !func) {
        return NULL;
    }
    if ((m = malloc(sizeof(Message))) == NULL) {
        fatal("Out of memory!");
    }
    m->name = sstrdup(name);
    m->func = func;
    m->mod_name = NULL;
    m->next = NULL;
    return m;
}

/*************************************************************************/

/** 
 * find a message in the given table. 
 * Looks up the message <name> in the MessageHash given
 * @param MessageHash the message table to search for this command, will almost always be IRCD
 * @param name the name of the command were looking for
 * @return NULL if we cant find it, or a pointer to the Message if we can
 **/
Message *findMessage(MessageHash * msgTable[], const char *name)
{
    int idx;
    MessageHash *current = NULL;

    SET_SEGV_LOCATION();

    if (!msgTable || !name) {
        return NULL;
    }
    idx = CMD_HASH(name);

    for (current = msgTable[idx]; current; current = current->next) {
        if (UseTokens) {
            if (ircd->tokencaseless) {
                if (stricmp(name, current->name) == 0) {
                    return current->m;
                }
            } else {
                if (strcmp(name, current->name) == 0) {
                    return current->m;
                }
            }
        } else {
            if (stricmp(name, current->name) == 0) {
                return current->m;
            }
        }
    }
    return NULL;
}

/*************************************************************************/

/* Iterate over all users in the user list.  Return NULL at end of list. */

static CommandHash *currentcmd;
static MessageHash *current;
static int next_index;

Command *first_command(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && currentcmd == NULL)
        currentcmd = STATSERV[next_index++];
    return currentcmd->c;
}

/*************************************************************************/

Command *next_command(void)
{
    SET_SEGV_LOCATION();

    if (currentcmd)
        currentcmd = currentcmd->next;
    if (!currentcmd && next_index < 1024) {
        while (next_index < 1024 && currentcmd == NULL)
            currentcmd = STATSERV[next_index++];
    }
    if (currentcmd) {
        return currentcmd->c;
    } else {
        return NULL;
    }
}

/*************************************************************************/

CommandHash *first_commandhash(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && currentcmd == NULL)
        currentcmd = STATSERV[next_index++];
    return currentcmd;
}

/*************************************************************************/

CommandHash *next_commandhash(void)
{
    SET_SEGV_LOCATION();

    if (currentcmd)
        currentcmd = currentcmd->next;
    if (!currentcmd && next_index < 1024) {
        while (next_index < 1024 && currentcmd == NULL)
            currentcmd = STATSERV[next_index++];
    }
    if (currentcmd) {
        return currentcmd;
    } else {
        return NULL;
    }
}

/*************************************************************************/

Message *first_message(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && current == NULL)
        current = IRCD[next_index++];
    return current->m;
}

/*************************************************************************/

Message *next_message(void)
{
    SET_SEGV_LOCATION();

    if (current)
        current = current->next;
    if (!current && next_index < 1024) {
        while (next_index < 1024 && current == NULL)
            current = IRCD[next_index++];
    }
    if (current) {
        return current->m;
    } else {
        return NULL;
    }
}

/*************************************************************************/

MessageHash *first_messagehash(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && current == NULL)
        current = IRCD[next_index++];
    return current;
}

/*************************************************************************/

MessageHash *next_messagehash(void)
{
    SET_SEGV_LOCATION();

    if (current)
        current = current->next;
    if (!current && next_index < 1024) {
        while (next_index < 1024 && current == NULL)
            current = IRCD[next_index++];
    }
    if (current) {
        return current;
    } else {
        return NULL;
    }
}

/*************************************************************************/

/**
 * Add a message to the MessageHash.
 * @param msgTable the MessageHash we want to add a message to
 * @param m the Message we want to add
 * @param pos the position we want to add the message to, E.G. MOD_HEAD, MOD_TAIL, MOD_UNIQUE 
 * @return MOD_ERR_OK on a successful add.
 **/

int addMessage(MessageHash * msgTable[], Message * m, int pos)
{
    /* We can assume both param's have been checked by this point.. */
    int modindex = 0;
    MessageHash *msgcurrent = NULL;
    MessageHash *newHash = NULL;
    MessageHash *lastHash = NULL;
    Message *tail = NULL;
    int match = 0;

    SET_SEGV_LOCATION();

    if (!msgTable || !m || (pos < 0 || pos > 2)) {
        return MOD_ERR_PARAMS;
    }

    modindex = CMD_HASH(m->name);

    for (msgcurrent = msgTable[modindex]; msgcurrent;
         msgcurrent = msgcurrent->next) {
        if ((UseTokens) && (!ircd->tokencaseless)) {
            match = strcmp(m->name, msgcurrent->name);
        } else {
            match = stricmp(m->name, msgcurrent->name);
        }
        if (match == 0) {       /* the msg exist's we are a addHead */
            if (pos == 1) {
                m->next = msgcurrent->m;
                msgcurrent->m = m;
                alog(LOG_DEBUG,
                     "debug: existing msg: (0x%p) (%s) , new msg (0x%p) (%s)",
                     (void *) m->next, m->next->name, (void *) m, m->name);
                return MOD_ERR_OK;
            } else if (pos == 2) {
                tail = msgcurrent->m;
                while (tail->next)
                    tail = tail->next;
                alog(LOG_DEBUG,
                     "debug: existing msg: (0x%p), new msg (0x%p) (%s)",
                     (void *) tail, (void *) m, m->name);
                tail->next = m;
                m->next = NULL;
                return MOD_ERR_OK;
            } else
                return MOD_ERR_EXISTS;
        }
        lastHash = msgcurrent;
    }

    if ((newHash = malloc(sizeof(MessageHash))) == NULL) {
        fatal("Out of memory");
    }
    newHash->next = NULL;
    newHash->name = sstrdup(m->name);
    newHash->m = m;

    if (lastHash == NULL)
        msgTable[modindex] = newHash;
    else
        lastHash->next = newHash;
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Add the given message (m) to the MessageHash marking it as a core command
 * @param msgTable the MessageHash we want to add to
 * @param m the Message we are adding
 * @return MOD_ERR_OK on a successful add.
 **/
int addCoreMessage(MessageHash * msgTable[], Message * m)
{
    SET_SEGV_LOCATION();

    if (!msgTable || !m) {
        return MOD_ERR_PARAMS;
    }
    m->core = 1;
    return addMessage(msgTable, m, 0);
}

/*************************************************************************/

/**
 * Add a module message to the IRCD message hash
 * @param m the Message to add
 * @param pos the Position to add the message to, e.g. MOD_HEAD, MOD_TAIL, MOD_UNIQUE
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int moduleAddMessage(Message * m, int pos)
{
    int status;

    SET_SEGV_LOCATION();

    if (!m) {
        return MOD_ERR_PARAMS;
    }

    /* ok, this appears to be a module adding a message from outside
     * of DenoraInit, try to look up its module struct for it */
    if ((mod_current_module_name) && (!mod_current_module)) {
        mod_current_module = findModule(mod_current_module_name);
    }

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }                           /* shouldnt happen */
    m->core = 0;
    if (!m->mod_name) {
        m->mod_name = sstrdup(mod_current_module->name);
    }

    status = addMessage(IRCD, m, pos);
    if (denora->debug) {
        displayMessageFromHash(m->name);
    }
    return status;
}

/*************************************************************************/

/**
 * remove the given message from the IRCD message hash
 * @param name the name of the message to remove
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int moduleDelMessage(char *name)
{
    Message *m;
    int status;

    SET_SEGV_LOCATION();

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }
    m = findMessage(IRCD, name);
    if (!m) {
        return MOD_ERR_NOEXIST;
    }

    status = delMessage(IRCD, m, mod_current_module->name);
    if (denora->debug) {
        displayMessageFromHash(m->name);
    }
    return status;
}

/*************************************************************************/

/**
 * remove the given message from the given message hash, for the given module
 * @param msgTable which MessageHash we are removing from
 * @param m the Message we want to remove
 * @mod_name the name of the module we are removing
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int delMessage(MessageHash * msgTable[], Message * m, char *mod_name)
{
    int msgindex = 0;
    MessageHash *msgcurrent = NULL;
    MessageHash *lastHash = NULL;
    Message *tail = NULL, *last = NULL;

    SET_SEGV_LOCATION();

    if (!m || !msgTable) {
        return MOD_ERR_PARAMS;
    }

    msgindex = CMD_HASH(m->name);

    for (msgcurrent = msgTable[msgindex]; msgcurrent;
         msgcurrent = msgcurrent->next) {
        if (stricmp(m->name, msgcurrent->name) == 0) {
            if (!lastHash) {
                tail = msgcurrent->m;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                msgcurrent->m = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    msgTable[msgindex] = msgcurrent->next;
                    denora_free(msgcurrent->name);
                    return MOD_ERR_OK;
                }
            } else {
                tail = msgcurrent->m;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                msgcurrent->m = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    lastHash->next = msgcurrent->next;
                    denora_free(msgcurrent->name);
                    return MOD_ERR_OK;
                }
            }
        }
        lastHash = msgcurrent;
    }
    return MOD_ERR_NOEXIST;
}

/*************************************************************************/

/**
 * Destory a message, freeing its memory.
 * @param m the message to be destroyed
 * @return MOD_ERR_SUCCESS on success
 **/
int destroyMessage(Message * m)
{
    SET_SEGV_LOCATION();

    if (!m) {
        return MOD_ERR_PARAMS;
    }
    if (m->name) {
        denora_free(m->name);
    }
    m->func = NULL;
    if (m->mod_name) {
        denora_free(m->mod_name);
    }
    m->next = NULL;
    denora_free(m);
    return MOD_ERR_OK;
}

/*************************************************************************/

int destroyMessageHash(MessageHash * mh)
{
    SET_SEGV_LOCATION();

    if (!mh) {
        return MOD_ERR_PARAMS;
    }
    if (mh->name) {
        denora_free(mh->name);
    }
    mh->m = NULL;
    mh->next = NULL;
    denora_free(mh);
    return MOD_ERR_OK;
}

/*************************************************************************/

int destroyCommandHash(CommandHash * ch)
{
    SET_SEGV_LOCATION();

    if (!ch) {
        return MOD_ERR_PARAMS;
    }
    if (ch->name) {
        denora_free(ch->name);
    }
    ch->c = NULL;
    ch->next = NULL;
    denora_free(ch);
    return MOD_ERR_OK;
}

/*************************************************************************/

void destroy_all(void)
{
    User *u, *next;
    Message *m, *next2;
    MessageHash *mh, *next3;
    Command *c, *next4;
    CommandHash *ch, *next5;

    if (mod_current_buffer) {
        denora_free(mod_current_buffer);
    }
    if (ircd->chanmodes) {
        denora_free(ircd->chanmodes);
    }
    if (ircd->nickchars) {
        denora_free(ircd->nickchars);
    }
    u = firstuser();
    while (u) {
        next = nextuser();
        delete_user(u);
        u = next;
    }
    do_squit(ServerName);
    alog(LOG_DEBUG, "debug: Clearing Event Messages");
    m = first_message();
    while (m) {
        next2 = next_message();
        if (m) {
            destroyMessage(m);
        }
        m = next2;
    }
    mh = first_messagehash();
    while (mh) {
        next3 = next_messagehash();
        if (mh) {
            destroyMessageHash(mh);
        }
        mh = next3;
    }

    alog(LOG_DEBUG, "debug: Clearing Commands");
    c = first_command();
    while (c) {
        next4 = next_command();
        if (c) {
            if (destroyCommand(c) == MOD_ERR_UNKNOWN) {
                alog(LOG_DEBUG, "debug: could not free command %s",
                     c->name);
            }
        }
        c = next4;
    }
    ch = first_commandhash();
    while (ch) {
        next5 = next_commandhash();
        if (ch) {
            destroyCommandHash(ch);
        }
        ch = next5;
    }
    if (denora->qmsg) {
        denora_free(denora->qmsg);
    }
    denora_free(denora->uplink);
    denora_free(UserTable);
    denora_free(ChanBansTable);
    denora_free(ChanExceptTable);
    denora_free(IsOnTable);
    denora_free(ServerTable);
    denora_free(GlineTable);
    denora_free(ChanTable);
    denora_free(MaxValueTable);
    denora_free(ChanInviteTable);
    denora_free(TLDTable);
    denora_free(NetworkName);
    denora_free(NickChar);
    denora_free(SqlHost);
    denora_free(SqlUser);
    denora_free(SqlPass);
    denora_free(SqlName);
    denora_free(SqlSock);
    denora_free(s_StatServ);
    denora_free(desc_StatServ);
    denora_free(s_StatServ_alias);
    denora_free(desc_StatServ_alias);
    denora_free(RemoteServer);
    denora_free(ExcludeServer);
    denora_free(RemotePassword);
    denora_free(LocalHost);
    denora_free(MOTDFilename);
    denora_free(ServerName);
    denora_free(ServerDesc);
    denora_free(ServiceUser);
    denora_free(NickChar);
    denora_free(HTMLFilename);
    free(GeoIPDBFileName);
    denora_free(excludeDB);
    remove_pidfile();
    denora_free(PIDFilename);
    denora_free(rdb_errmsg);
}

/*************************************************************************/

/**
 * Add the modules version info.
 * @param version the version of the current module
 **/
void moduleAddVersion(const char *version)
{
    SET_SEGV_LOCATION();

    if (mod_current_module && version) {
        mod_current_module->version = sstrdup(version);
    }
}

/*************************************************************************/

/**
 * Add the modules author info
 * @param author the author of the module
 **/
void moduleAddAuthor(const char *author)
{
    SET_SEGV_LOCATION();

    if (mod_current_module && author) {
        mod_current_module->author = sstrdup(author);
    }
}

/*******************************************************************************
 * Module Callback Functions
 *******************************************************************************/
 /**
  * Adds a timed callback for the current module.
  * This allows modules to request that denora executes one of there functions at a time in the future, without an event to trigger it
  * @param name the name of the callback, this is used for refrence mostly, but is needed it you want to delete this particular callback later on
  * @param when when should the function be executed, this is a time in the future, seconds since 00:00:00 1970-01-01 UTC
  * @param func the function to be executed when the callback is ran, its format MUST be int func(int argc, char **argv); 
  * @param argc the argument count for the argv paramter
  * @param atgv a argument list to be passed to the called function.
  * @return MOD_ERR_OK on success, anything else on fail.
  * @see moduleDelCallBack
  **/
int moduleAddCallback(char *name, time_t when,
                      int (*func) (int argc, char *argv[]), int argc,
                      char **argv)
{
    ModuleCallBack *new, *tmp, *prev;
    int i;

    SET_SEGV_LOCATION();

    new = malloc(sizeof(ModuleCallBack));
    if (!new)
        return MOD_ERR_MEMORY;

    if (name)
        new->name = sstrdup(name);
    else
        new->name = NULL;
    new->when = when;
    if (mod_current_module_name) {
        new->owner_name = sstrdup(mod_current_module_name);
    } else {
        new->owner_name = NULL;
    }
    new->func = func;
    new->argc = argc;
    new->argv = malloc(sizeof(char *) * argc);
    for (i = 0; i < argc; i++) {
        new->argv[i] = sstrdup(argv[i]);
    }
    new->next = NULL;

    if (moduleCallBackHead == NULL) {
        moduleCallBackHead = new;
    } else {                    /* find place in list */
        tmp = moduleCallBackHead;
        prev = tmp;
        if (new->when < tmp->when) {
            new->next = tmp;
            moduleCallBackHead = new;
        } else {
            while (tmp && new->when >= tmp->when) {
                prev = tmp;
                tmp = tmp->next;
            }
            prev->next = new;
            new->next = tmp;
        }
    }
    alog(LOG_DEBUG,
         langstr(ALOG_MOD_CALLBACK_DISPLAY),
         new->name ? new->name : "?", (long int) new->when);
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Execute a stored call back
 **/
void moduleCallBackRun(void)
{
    ModuleCallBack *tmp;

    SET_SEGV_LOCATION();

    if (!moduleCallBackHead) {
        return;
    }
    tmp = moduleCallBackHead;
    if (tmp->when <= time(NULL)) {
        alog(LOG_DEBUG, "debug: Executing callback: %s",
             tmp->name ? tmp->name : "?");
        if (tmp->func) {
            mod_current_module_name = tmp->owner_name;
            tmp->func(tmp->argc, tmp->argv);
            mod_current_module = NULL;
            moduleCallBackDeleteEntry(NULL);    /* delete the head */
        }
    }
    return;
}

/*************************************************************************/

/**
 * Removes a entry from the modules callback list
 * @param prev a pointer to the previous entry in the list, NULL for the head
 **/
void moduleCallBackDeleteEntry(ModuleCallBack * prev)
{
    ModuleCallBack *tmp = NULL;
    int i;

    SET_SEGV_LOCATION();

    if (prev == NULL) {
        tmp = moduleCallBackHead;
        moduleCallBackHead = tmp->next;
    } else {
        tmp = prev->next;
        prev->next = tmp->next;
    }
    if (tmp->name)
        denora_free(tmp->name);
    if (tmp->owner_name)
        denora_free(tmp->owner_name);
    tmp->func = NULL;
    for (i = 0; i < tmp->argc; i++) {
        denora_free(tmp->argv[i]);
    }
    tmp->argc = 0;
    tmp->next = NULL;
    denora_free(tmp);
}

/*************************************************************************/

/**
 * Search the module callback list for a given module
 * @param mod_name the name of the module were looking for
 * @param found have we found it?
 * @return a pointer to the ModuleCallBack struct or NULL - dont forget to check the found paramter!
 **/
ModuleCallBack *moduleCallBackFindEntry(char *mod_name, boolean * found)
{
    ModuleCallBack *prev = NULL, *modcurrent = NULL;
    *found = false;
    modcurrent = moduleCallBackHead;

    SET_SEGV_LOCATION();

    while (modcurrent != NULL) {
        if (modcurrent->owner_name
            && (strcmp(mod_name, modcurrent->owner_name) == 0)) {
            *found = true;
            break;
        } else {
            prev = modcurrent;
            modcurrent = modcurrent->next;
        }
    }
    if (modcurrent == moduleCallBackHead) {
        return NULL;
    } else {
        return prev;
    }
}

/*************************************************************************/

/**
 * Allow module coders to delete a callback by name.
 * @param name the name of the callback they wish to delete
 **/
void moduleDelCallback(char *name)
{
    ModuleCallBack *modcurrent = NULL;
    ModuleCallBack *prev = NULL, *tmp = NULL;
    int del = 0;

    SET_SEGV_LOCATION();

    if (!mod_current_module_name) {
        return;
    }
    if (!name) {
        return;
    }
    modcurrent = moduleCallBackHead;
    while (modcurrent) {
        if ((modcurrent->owner_name) && (modcurrent->name)) {
            if ((strcmp(mod_current_module_name, modcurrent->owner_name) ==
                 0)
                && (strcmp(modcurrent->name, name) == 0)) {
                alog(LOG_DEBUG,
                     "debug: Removing CallBack %s for module %s", name,
                     mod_current_module_name);
                tmp = modcurrent->next; /* get a pointer to the next record, as once we delete this record, we'll lose it :) */
                moduleCallBackDeleteEntry(prev);        /* delete this record */
                del = 1;        /* set the record deleted flag */
            }
        }
        if (del == 1) {         /* if a record was deleted */
            modcurrent = tmp;   /* use the value we stored in temp */
            tmp = NULL;         /* clear it for next time */
            del = 0;            /* reset the flag */
        } else {
            prev = modcurrent;  /* just carry on as normal */
            modcurrent = modcurrent->next;
        }
    }
}

/*************************************************************************/

/**
 * Remove all outstanding module callbacks for the given module.
 * When a module is unloaded, any callbacks it had outstanding must be removed, else when they attempt to execute the func pointer will no longer be valid, and we'll seg.
 * @param mod_name the name of the module we are preping for unload 
 **/
void moduleCallBackPrepForUnload(char *mod_name)
{
    boolean found = false;
    ModuleCallBack *tmp = NULL;

    tmp = moduleCallBackFindEntry(mod_name, &found);
    SET_SEGV_LOCATION();

    while (found) {
        alog(LOG_DEBUG, "debug: Removing CallBack for module %s",
             mod_name);
        moduleCallBackDeleteEntry(tmp);
        tmp = moduleCallBackFindEntry(mod_name, &found);
    }
}

/*************************************************************************/

/**
 * Return a copy of the complete last buffer.
 * This is needed for modules who cant trust the strtok() buffer, as we dont know who will have already messed about with it.
 * @reutrn a pointer to a copy of the last buffer - DONT mess with this, copy if first if you must do things to it.
 **/
char *moduleGetLastBuffer(void)
{
    char *tmp = NULL;
    SET_SEGV_LOCATION();

    if (mod_current_buffer) {
        tmp = strchr(mod_current_buffer, ' ');
        if (tmp) {
            tmp++;
        }
    }
    if (tmp) {
        return sstrdup(tmp);
    } else {
        return sstrdup("");
    }
}

/*******************************************************************************
 * Module HELP Functions
 *******************************************************************************/
 /**
  * Add help for Stats Admins.
  * @param c the Command to add help for
  * @param func the function to run when this help is asked for
  **/
int moduleAddAdminHelp(Command * c, int (*func) (User * u))
{
    SET_SEGV_LOCATION();
    if (c) {
        c->admin_help = func;
        return MOD_STOP;
    }
    return MOD_CONT;
}

/*************************************************************************/

/**
  * Add help for registered users
  * @param c the Command to add help for
  * @param func the function to run when this help is asked for
  **/
int moduleAddRegHelp(Command * c, int (*func) (User * u))
{
    SET_SEGV_LOCATION();
    if (c) {
        c->regular_help = func;
        return MOD_STOP;
    }
    return MOD_CONT;
}

/*************************************************************************/

/**
  * Add help for all users
  * @param c the Command to add help for
  * @param func the function to run when this help is asked for
  **/
int moduleAddHelp(Command * c, int (*func) (User * u))
{
    SET_SEGV_LOCATION();
    if (c) {
        c->all_help = func;
        return MOD_STOP;
    }
    return MOD_CONT;
}

/*************************************************************************/

/**
  * Add output to statserv help.
  * when doing a /msg statserv help, your function will be called to allow it to send out a notice() 
  * with the code you wish to dispaly
  * @param func a pointer to the function which will display the code
  **/
void moduleSetOperHelp(void (*func) (User * u))
{
    SET_SEGV_LOCATION();
    if (mod_current_module) {
        mod_current_module->operHelp = func;
    }
}

/*************************************************************************/

/**
 * Display any extra module help for the given service.
 * @param services which services is help being dispalyed for?
 * @param u which user is requesting the help
 **/
void moduleDisplayHelp(int service, User * u)
{
#ifdef USE_MODULES
    int idx;
    int header_shown = 0;
    ModuleHash *modcurrent = NULL;
    SET_SEGV_LOCATION();

    for (idx = 0; idx != MAX_CMD_HASH; idx++) {
        for (modcurrent = MODULE_HASH[idx]; modcurrent;
             modcurrent = modcurrent->next) {
            if ((service == 1) && modcurrent->m->operHelp) {
                if (header_shown == 0) {
                    notice_lang(s_StatServ, u, MODULE_HELP_HEADER);
                    header_shown = 1;
                }
                modcurrent->m->operHelp(u);
            }
        }
    }
#endif
}

/*************************************************************************/

/**
 * Output module data information into the log file.
 * This is a vwey "debug only" function to dump the whole contents
 * of a moduleData struct into the log files.
 * @param md The module data for the struct to be used
 * @return 0 is always returned;
 **/
int moduleDataDebug(ModuleData ** md)
{
    ModuleData *modcurrent = NULL;
    alog(LOG_DEBUG, "debug: Dumping module data....");

    SET_SEGV_LOCATION();

    for (modcurrent = *md; modcurrent; modcurrent = modcurrent->next) {
        alog(LOG_DEBUG, "debug: Module: [%s]", modcurrent->moduleName);
        alog(LOG_DEBUG, "debug: Key [%s]\tValue [%s]", modcurrent->key,
             modcurrent->value);
    }
    alog(LOG_DEBUG, "debug: End of module data dump");
    return 0;
}

/*************************************************************************/

/**
 * Add module data to a struct.
 * This allows module coders to add data to an existing struct
 * @param md The module data for the struct to be used
 * @param key The Key for the key/value pair
 * @param value The value for the key/value pair, this is what will be stored for you
 * @return MOD_ERR_OK will be returned on success
 **/
int moduleAddData(ModuleData ** md, char *key, char *value)
{
    char *mod_name = sstrdup(mod_current_module_name);
    ModuleData *newData = NULL;
    ModuleData *tmp = *md;

    SET_SEGV_LOCATION();

    if (!key || !value) {
        alog(LOG_DEBUG,
             "debug: A module tried to use ModuleAddData() with one ore more NULL arguments... returning");
        do_backtrace(0);
        denora_free(mod_name);
        return MOD_ERR_PARAMS;
    }

    if (mod_current_module_name == NULL) {
        alog(LOG_DEBUG,
             "debug: moduleAddData() called with mod_current_module_name being NULL");
        if (denora->debug) {
            do_backtrace(0);
        }
    }

    moduleDelData(md, key);     /* Remove any existing module data for this module with the same key */

    newData = malloc(sizeof(ModuleData));
    if (!newData) {
        denora_free(mod_name);
        return MOD_ERR_MEMORY;
    }

    newData->moduleName = sstrdup(mod_name);
    newData->key = sstrdup(key);
    newData->value = sstrdup(value);
    if (tmp) {
        newData->next = tmp;
    } else {
        newData->next = NULL;
    }
    *md = newData;

    denora_free(mod_name);

    if (denora->debug) {
        moduleDataDebug(md);
    }
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Returns the value from a key/value pair set.
 * This allows module coders to retrive any data they have previuosly stored in any given struct
 * @param md The module data for the struct to be used
 * @param key The key to find the data for
 * @return the value paired to the given key will be returned, or NULL 
 **/
char *moduleGetData(ModuleData ** md, char *key)
{

    char *mod_name = sstrdup(mod_current_module_name);
    ModuleData *modcurrent = *md;

    SET_SEGV_LOCATION();

    alog(LOG_DEBUG, "debug: moduleGetData %p : key %s", (void *) md, key);
    alog(LOG_DEBUG, "debug: Current Module %s", mod_name);

    if (mod_current_module_name == NULL) {
        alog(LOG_DEBUG,
             "moduleGetData() called with mod_current_module_name being NULL");
        if (denora->debug) {
            do_backtrace(0);
        }
    }

    while (modcurrent) {
        if ((stricmp(modcurrent->moduleName, mod_name) == 0)
            && (stricmp(modcurrent->key, key) == 0)) {
            denora_free(mod_name);
            return sstrdup(modcurrent->value);
        }
        modcurrent = modcurrent->next;
    }
    denora_free(mod_name);
    return NULL;
}

/*************************************************************************/

/**
 * Delete the key/value pair indicated by "key" for the current module.
 * This allows module coders to remove a previously stored key/value pair.
 * @param md The module data for the struct to be used
 * @param key The key to delete the key/value pair for
 **/
void moduleDelData(ModuleData ** md, char *key)
{
    char *mod_name = sstrdup(mod_current_module_name);
    ModuleData *modcurrent = *md;
    ModuleData *prev = NULL;
    ModuleData *next = NULL;

    SET_SEGV_LOCATION();

    if (mod_current_module_name == NULL) {
        alog(LOG_DEBUG,
             "debug: moduleDelData() called with mod_current_module_name being NULL");
        if (denora->debug) {
            do_backtrace(0);
        }
    }

    if (key) {
        while (modcurrent) {
            next = modcurrent->next;
            if ((stricmp(modcurrent->moduleName, mod_name) == 0)
                && (stricmp(modcurrent->key, key) == 0)) {
                if (prev) {
                    prev->next = modcurrent->next;
                } else {
                    *md = modcurrent->next;
                }
                denora_free(modcurrent->moduleName);
                denora_free(modcurrent->key);
                denora_free(modcurrent->value);
                modcurrent->next = NULL;
                denora_free(modcurrent);
            } else {
                prev = modcurrent;
            }
            prev = modcurrent;
            modcurrent = next;
        }
    }
    denora_free(mod_name);
}

/*************************************************************************/

/**
 * This will remove all data for a particular module from existing structs.
 * Its primary use is modulePrepForUnload() however, based on past expericance with module coders wanting to 
 * do just about anything and everything, its safe to use from inside the module.
 * @param md The module data for the struct to be used
 **/
void moduleDelAllData(ModuleData ** md)
{
    char *mod_name = sstrdup(mod_current_module_name);
    ModuleData *modcurrent = *md;
    ModuleData *prev = NULL;
    ModuleData *next = NULL;

    SET_SEGV_LOCATION();

    if (mod_current_module_name == NULL) {
        alog(LOG_DEBUG,
             "debug: moduleDelAllData() called with mod_current_module_name being NULL");
        if (denora->debug) {
            do_backtrace(0);
        }
    }

    while (modcurrent) {
        next = modcurrent->next;
        if ((stricmp(modcurrent->moduleName, mod_name) == 0)) {
            if (prev) {
                prev->next = modcurrent->next;
            } else {
                *md = modcurrent->next;
            }
            denora_free(modcurrent->moduleName);
            denora_free(modcurrent->key);
            denora_free(modcurrent->value);
            modcurrent->next = NULL;
            denora_free(modcurrent);
        } else {
            prev = modcurrent;
        }
        modcurrent = next;
    }
    denora_free(mod_name);
}

/*************************************************************************/

/**
 * This will delete all module data used in any struct by module m.
 * @param m The module to clear all data for
 **/
void moduleDelAllDataMod(Module * m)
{
    boolean freeme = false;
    int i;
    User *user;

    SET_SEGV_LOCATION();

    if (!mod_current_module_name) {
        mod_current_module_name = sstrdup(m->name);
        freeme = true;
    }

    for (i = 0; i < 1024; i++) {
        /* Remove the users */
        for (user = userlist[i]; user; user = user->next) {
            moduleDelAllData(&user->moduleData);
        }
    }

    if (freeme) {
        denora_free(mod_current_module_name);
        mod_current_module_name = NULL;
    }
}

/*************************************************************************/

/**
 * Remove any data from any module used in the given struct.
 * Useful for cleaning up when a User leave's the net, a NickCore is deleted, etc...
 * @param moduleData the moduleData struct to "clean"
 **/
void moduleCleanStruct(ModuleData ** moduleData)
{
    ModuleData *modcurrent = *moduleData;
    ModuleData *next = NULL;

    SET_SEGV_LOCATION();

    while (modcurrent) {
        next = modcurrent->next;
        denora_free(modcurrent->moduleName);
        denora_free(modcurrent->key);
        denora_free(modcurrent->value);
        modcurrent->next = NULL;
        denora_free(modcurrent);
        modcurrent = next;
    }
    *moduleData = NULL;
}

/*************************************************************************/

/**
 * Check the current version of denora against a given version number
 * Specifiying -1 for minor,patch or build 
 * @param major The major version of denora, the first part of the verison number
 * @param minor The minor version of denora, the second part of the version number
 * @param patch The patch version of denora, the third part of the version number
 * @param build The build revision of denora from SVN
 * @return True if the version newer than the version specified.
 **/
boolean moduleMinVersion(int major, int minor, int patch, int build)
{
    boolean ret = false;

    SET_SEGV_LOCATION();

    if (VERSION_MAJOR > major) {        /* Def. new */
        ret = true;
    } else if (VERSION_MAJOR == major) {        /* Might be newer */
        if (minor == -1) {
            return true;
        }                       /* They dont care about minor */
        if (VERSION_MINOR > minor) {    /* Def. newer */
            ret = true;
        } else if (VERSION_MINOR == minor) {    /* Might be newer */
            if (patch == -1) {
                return true;
            }                   /* They dont care about patch */
            if (VERSION_PATCH > patch) {
                ret = true;
            } else if (VERSION_PATCH == patch) {
                if (build == -1) {
                    return true;
                }               /* They dont care about build */
                if (VERSION_BUILD >= build) {
                    ret = true;
                }
            }
        }
    }
    return ret;
}

/*************************************************************************/

/**
 * Return the Command corresponding to the given name, or NULL if no such
 * command exists.
 * @param list Command struct
 * @param cmd Command to look up
 * @return Command Struct for the given cmd
 */
Command *lookup_cmd(Command * list, char *cmd)
{
    Command *c;

    SET_SEGV_LOCATION();

    for (c = list; c->name; c++) {
        if (stricmp(c->name, cmd) == 0) {
            return c;
        }
    }
    return NULL;
}

/*************************************************************************/

/**
 * Run the routine for the given command, if it exists and the user has
 * privilege to do so; if not, print an appropriate error message.
 * @param services Services Client
 * @param u User Struct
 * @param list Command struct
 * @param cmd Command
 * @return void
 */
void run_cmd(char *service, User * u, Command * list, char *cmd)
{
    Command *c;
    SET_SEGV_LOCATION();
    c = lookup_cmd(list, cmd);
    do_run_cmd(service, u, c, cmd);
}

/*************************************************************************/

/**
 * Run the routine for the given command, if it exists and the user has
 * privilege to do so; if not, print an appropriate error message.
 * @param services Services Client
 * @param u User Struct
 * @param Command Hash Table
 * @param cmd Command
 * @return void
 */
void mod_run_cmd(char *service, User * u, CommandHash * cmdTable[],
                 const char *cmd)
{
    Command *c;
    SET_SEGV_LOCATION();
    c = findCommand(cmdTable, cmd);
    do_run_cmd(service, u, c, cmd);
}


/*************************************************************************/

/**
 * Run the given command
 * @param services Services Client
 * @param u User Struct
 * @param c Command Struct
 * @param cmd Command
 * @return void
 */
void do_run_cmd(char *service, User * u, Command * c, const char *cmd)
{
    int retVal = 0;
    Command *cmdcurrent;
    SET_SEGV_LOCATION();

    if (c && c->routine) {
        if ((c->has_priv == NULL) || c->has_priv(u)) {
            mod_current_module_name = c->mod_name;
            retVal = c->routine(u);
            mod_current_module_name = NULL;
            if (retVal == MOD_CONT) {
                cmdcurrent = c->next;
                while (cmdcurrent && retVal == MOD_CONT) {
                    mod_current_module_name = cmdcurrent->mod_name;
                    retVal = cmdcurrent->routine(u);
                    mod_current_module_name = NULL;
                    cmdcurrent = cmdcurrent->next;
                }
            }
        }

        else {
            notice_lang(service, u, ACCESS_DENIED);
            alog(LOG_NORMAL,
                 "Access denied for %s with service %s and command %s",
                 u->nick, service, cmd);
        }
    } else {
        notice_lang(service, u, UNKNOWN_COMMAND_HELP, cmd, service);
    }
}

/*************************************************************************/

/**
 * Print a help message for the given command.
 * @param services Services Client
 * @param u User Struct
 * @param c Command Struct
 * @param cmd Command
 * @return void
 */
void do_help_cmd(char *service, User * u, Command * c, const char *cmd)
{
    Command *cmdcurrent;
    int has_had_help = 0;
    int cont = MOD_CONT;
    const char *p1 = NULL, *p2 = NULL, *p3 = NULL, *p4 = NULL;

    SET_SEGV_LOCATION();

    for (cmdcurrent = c; (cmdcurrent) && (cont == MOD_CONT);
         cmdcurrent = cmdcurrent->next) {
        p1 = cmdcurrent->help_param1;
        p2 = cmdcurrent->help_param2;
        p3 = cmdcurrent->help_param3;
        p4 = cmdcurrent->help_param4;
        if (cmdcurrent->helpmsg_all >= 0) {
            notice_help(service, u, cmdcurrent->helpmsg_all, p1, p2, p3,
                        p4);
            has_had_help = 1;
        } else if (cmdcurrent->all_help) {
            cont = cmdcurrent->all_help(u);
            has_had_help = 1;
        }
        if (is_stats_admin(u)) {
            if (cmdcurrent->helpmsg_admin >= 0) {
                notice_help(service, u, cmdcurrent->helpmsg_admin, p1, p2,
                            p3, p4);
                has_had_help = 1;
            } else if (cmdcurrent->admin_help) {
                cont = cmdcurrent->admin_help(u);
                has_had_help = 1;
            }
        } else {
            if (cmdcurrent->helpmsg_reg >= 0) {
                notice_help(service, u, cmdcurrent->helpmsg_reg, p1, p2,
                            p3, p4);
                has_had_help = 1;
            } else if (cmdcurrent->regular_help) {
                cont = cmdcurrent->regular_help(u);
                has_had_help = 1;
            }
        }
    }
    if (has_had_help == 0) {
        notice_lang(service, u, NO_HELP_AVAILABLE, cmd);
    }
}

/*************************************************************************/

/**
 * Find the Help Command
 * @param services Services Client
 * @param u User Struct
 * @param c Command Struct
 * @param cmd Command
 * @return void
 */
void help_cmd(char *service, User * u, Command * list, char *cmd)
{
    Command *c;
    SET_SEGV_LOCATION();
    c = lookup_cmd(list, cmd);
    do_help_cmd(service, u, c, cmd);
}

/*************************************************************************/

/**
 * Find the Help Command
 * @param services Services Client
 * @param u User Struct
 * @param Command Hash Table
 * @param cmd Command
 * @return void
 */
void mod_help_cmd(char *service, User * u, CommandHash * cmdTable[],
                  const char *cmd)
{
    Command *c;
    SET_SEGV_LOCATION();
    c = findCommand(cmdTable, cmd);
    do_help_cmd(service, u, c, cmd);
}

/*************************************************************************/

EvtMessage *find_event(const char *name)
{
    EvtMessage *m;
    m = findEventHandler(EVENT, name);
    return m;
}

/*************************************************************************/

EvtHook *find_eventhook(const char *name)
{
    EvtHook *m;
    m = findEventHook(EVENTHOOKS, name);
    return m;
}

/*************************************************************************/

void send_event(const char *name, int argc, ...)
{
    va_list va;
    char *a;
    int idx = 0;
    char **argv;

    argv = (char **) malloc(sizeof(char *) * argc);
    va_start(va, argc);
    for (idx = 0; idx < argc; idx++) {
        a = va_arg(va, char *);
        argv[idx] = sstrdup(a);
    }
    va_end(va);

    alog(LOG_EXTRADEBUG, "debug: Emitting event \"%s\" (%d args)", name,
         argc);

    event_process_hook(name, argc, argv);

    /**
     * Now that the events have seen the message, free it up
     **/
    for (idx = 0; idx < argc; idx++) {
        denora_free(argv[idx]);
    }
    denora_free(argv);
}

/*************************************************************************/

void eventprintf(const char *fmt, ...)
{
    va_list args;
    char buf[16384];            /* Really huge, to try and avoid truncation */
    char *event;

    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);
    event = sstrdup(buf);
    event_message_process(event);
    va_end(args);
    if (event) {
        denora_free(event);
    }
    return;
}

/*************************************************************************/

void event_message_process(char *eventbuf)
{
    int retVal = 0;
    EvtMessage *eventcurrent = NULL;
    char source[64];
    char cmd[64];
    char buf[512];              /* Longest legal IRC command line */
    char *s;
    int ac;                     /* Parameters for the command */
    char **av;
    EvtMessage *evm;

    /* zero out the buffers before we do much else */
    *buf = '\0';
    *source = '\0';
    *cmd = '\0';

    strscpy(buf, eventbuf, sizeof(buf));

    if (!BadPtr(buf)) {
        doCleanBuffer((char *) buf);
    } else {
        return;
    }

    /* Split the buffer into pieces. */
    if (*buf == ':') {
        s = strpbrk(buf, " ");
        if (!s)
            return;
        *s = 0;
        while (isspace(*++s));
        strscpy(source, buf + 1, sizeof(source));
        memmove(buf, s, strlen(s) + 1);
    } else {
        *source = 0;
    }
    if (!*buf)
        return;
    s = strpbrk(buf, " ");
    if (s) {
        *s = 0;
        while (isspace(*++s));
    } else
        s = buf + strlen(buf);
    strscpy(cmd, buf, sizeof(cmd));
    ac = split_buf(s, &av);

    /* Do something with the message. */
    evm = find_event(cmd);
    if (evm) {
        if (evm->func) {
            mod_current_module_name = evm->mod_name;
            retVal = evm->func(source, ac, av);
            mod_current_module_name = NULL;
            if (retVal == MOD_CONT) {
                eventcurrent = evm->next;
                while (eventcurrent && eventcurrent->func
                       && retVal == MOD_CONT) {
                    mod_current_module_name = eventcurrent->mod_name;
                    retVal = eventcurrent->func(source, ac, av);
                    mod_current_module_name = NULL;
                    eventcurrent = eventcurrent->next;
                }
            }
        }
    }
    /* Free argument list we created */
    denora_free(av);
}

/*************************************************************************/

void event_process_hook(const char *name, int argc, char **argv)
{
    int retVal = 0;
    EvtHook *eventcurrent = NULL;
    EvtHook *evh;

    /* Do something with the message. */
    evh = find_eventhook(name);
    if (evh) {
        if (evh->func) {
            mod_current_module_name = evh->mod_name;
            retVal = evh->func(argc, argv);
            mod_current_module_name = NULL;
            if (retVal == MOD_CONT) {
                eventcurrent = evh->next;
                while (eventcurrent && eventcurrent->func
                       && retVal == MOD_CONT) {
                    mod_current_module_name = eventcurrent->mod_name;
                    retVal = eventcurrent->func(argc, argv);
                    mod_current_module_name = NULL;
                    eventcurrent = eventcurrent->next;
                }
            }
        }
    }
}

/*************************************************************************/

/**
 * Displays a message list for a given message.
 * Again this is of little use other than debugging.
 * @param m the message to display
 * @return 0 is returned and has no meaning 
 */
int displayEventMessage(EvtMessage * evm)
{
    EvtMessage *msg = NULL;
    int i = 0;
    alog(LOG_DEBUG, "debug: Displaying message list for %s", evm->name);
    for (msg = evm; msg; msg = msg->next) {
        alog(LOG_DEBUG, "%d: 0x%p", ++i, (void *) msg);
    }
    alog(LOG_DEBUG, "debug: end");
    return 0;
}

/*************************************************************************/

/**
 * Displays a message list for a given message.
 * Again this is of little use other than debugging.
 * @param m the message to display
 * @return 0 is returned and has no meaning 
 */
int displayEventHook(EvtHook * evh)
{
    EvtHook *msg = NULL;
    int i = 0;
    alog(LOG_DEBUG, "debug: Displaying message list for %s", evh->name);
    for (msg = evh; msg; msg = msg->next) {
        alog(LOG_DEBUG, "debug: %d: 0x%p", ++i, (void *) msg);
    }
    alog(LOG_DEBUG, "debug: end");
    return 0;
}

/*************************************************************************/

/**
 * Display the message call stak.
 * Prints the call stack for a message based on the message name, again useful for debugging and little lese :)
 * @param name the name of the message to print info for
 * @return the return int has no relevence atm :)
 */
int displayHookFromHash(char *name)
{
    EvtHookHash *eventcurrent = NULL;
    int eventindex = 0;
    eventindex = CMD_HASH(name);

    alog(LOG_EXTRADEBUG, "debug: trying to display message %s", name);

    for (eventcurrent = EVENTHOOKS[eventindex]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(name, eventcurrent->name) == 0) {
            displayEventHook(eventcurrent->evh);
        }
    }
    alog(LOG_EXTRADEBUG, "debug: done displaying message %s", name);
    return 0;
}

/*************************************************************************/

/**
 * Display the message call stak.
 * Prints the call stack for a message based on the message name, again useful for debugging and little lese :)
 * @param name the name of the message to print info for
 * @return the return int has no relevence atm :)
 */
int displayEvtMessageFromHash(char *name)
{
    EvtMessageHash *eventcurrent = NULL;
    int eventindex = 0;
    eventindex = CMD_HASH(name);

    alog(LOG_EXTRADEBUG, "debug: trying to display message %s", name);

    for (eventcurrent = EVENT[eventindex]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(name, eventcurrent->name) == 0) {
            displayEventMessage(eventcurrent->evm);
        }
    }
    alog(LOG_EXTRADEBUG, "debug: done displaying message %s", name);
    return 0;
}

/*******************************************************************************
 * Message Functions
 *******************************************************************************/

 /**
  * Create a new Message struct.
  * @param name the name of the message
  * @param func a pointer to the function to call when we recive this message
  * @return a new Message object
  **/
EvtMessage *createEventHandler(char *name,
                               int (*func) (char *source, int ac,
                                            char **av))
{
    EvtMessage *evm = NULL;
    if (!func) {
        return NULL;
    }
    if ((evm = malloc(sizeof(EvtMessage))) == NULL) {
        fatal("Out of memory!");
    }
    evm->name = sstrdup(name);
    evm->func = func;
    evm->mod_name = NULL;
    evm->next = NULL;
    return evm;
}

/*************************************************************************/

 /**
  * Create a new Message struct.
  * @param name the name of the message
  * @param func a pointer to the function to call when we recive this message
  * @return a new Message object
  **/
EvtHook *createEventHook(char *name, int (*func) (int argc, char **argv))
{
    EvtHook *evh = NULL;
    if (!func) {
        return NULL;
    }
    if ((evh = malloc(sizeof(EvtHook))) == NULL) {
        fatal("Out of memory!");
    }
    evh->name = sstrdup(name);
    evh->func = func;
    evh->mod_name = NULL;
    evh->next = NULL;
    return evh;
}

/*************************************************************************/

/** 
 * find a message in the given table. 
 * Looks up the message <name> in the MessageHash given
 * @param MessageHash the message table to search for this command, will almost always be IRCD
 * @param name the name of the command were looking for
 * @return NULL if we cant find it, or a pointer to the Message if we can
 **/
EvtMessage *findEventHandler(EvtMessageHash * msgEvtTable[],
                             const char *name)
{
    int idx;
    EvtMessageHash *eventcurrent = NULL;
    if (!msgEvtTable || !name) {
        return NULL;
    }
    idx = CMD_HASH(name);

    for (eventcurrent = msgEvtTable[idx]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(name, eventcurrent->name) == 0) {
            return eventcurrent->evm;
        }
    }
    return NULL;
}

/*************************************************************************/

/** 
 * find a message in the given table. 
 * Looks up the message <name> in the MessageHash given
 * @param MessageHash the message table to search for this command, will almost always be IRCD
 * @param name the name of the command were looking for
 * @return NULL if we cant find it, or a pointer to the Message if we can
 **/
EvtHook *findEventHook(EvtHookHash * hookEvtTable[], const char *name)
{
    int idx;
    EvtHookHash *eventcurrent = NULL;
    if (!hookEvtTable || !name) {
        return NULL;
    }
    idx = CMD_HASH(name);

    for (eventcurrent = hookEvtTable[idx]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(name, eventcurrent->name) == 0) {
            return eventcurrent->evh;
        }
    }
    return NULL;
}

/*************************************************************************/

/**
 * Add the given message (m) to the MessageHash marking it as a core command
 * @param msgTable the MessageHash we want to add to
 * @param m the Message we are adding
 * @return MOD_ERR_OK on a successful add.
 **/
int addCoreEventHandler(EvtMessageHash * msgEvtTable[], EvtMessage * evm)
{
    if (!msgEvtTable || !evm) {
        return MOD_ERR_PARAMS;
    }
    evm->core = 1;
    return addEventHandler(msgEvtTable, evm);
}

/*************************************************************************/

/**
 * Add a message to the MessageHash.
 * @param msgTable the MessageHash we want to add a message to
 * @param m the Message we want to add
 * @return MOD_ERR_OK on a successful add.
 **/
int addEventHandler(EvtMessageHash * msgEvtTable[], EvtMessage * evm)
{
    /* We can assume both param's have been checked by this point.. */
    int eventindex = 0;
    EvtMessageHash *eventcurrent = NULL;
    EvtMessageHash *newHash = NULL;
    EvtMessageHash *lastHash = NULL;

    if (!msgEvtTable || !evm) {
        return MOD_ERR_PARAMS;
    }

    eventindex = CMD_HASH(evm->name);

    for (eventcurrent = msgEvtTable[eventindex]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(evm->name, eventcurrent->name) == 0) {      /* the msg exist's we are a addHead */
            evm->next = eventcurrent->evm;
            eventcurrent->evm = evm;
            alog(LOG_DEBUG,
                 "debug: existing msg: (0x%p), new msg (0x%p) (%s)",
                 (void *) evm->next, (void *) evm, evm->name);
            return MOD_ERR_OK;
        }
        lastHash = eventcurrent;
    }

    if ((newHash = malloc(sizeof(EvtMessageHash))) == NULL) {
        fatal("Out of memory");
    }
    newHash->next = NULL;
    newHash->name = sstrdup(evm->name);
    newHash->evm = evm;

    if (lastHash == NULL)
        msgEvtTable[eventindex] = newHash;
    else
        lastHash->next = newHash;
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Add a message to the MessageHash.
 * @param msgTable the MessageHash we want to add a message to
 * @param m the Message we want to add
 * @return MOD_ERR_OK on a successful add.
 **/
int addEventHook(EvtHookHash * hookEvtTable[], EvtHook * evh)
{
    /* We can assume both param's have been checked by this point.. */
    int eventindex = 0;
    EvtHookHash *eventcurrent = NULL;
    EvtHookHash *newHash = NULL;
    EvtHookHash *lastHash = NULL;

    if (!hookEvtTable || !evh) {
        return MOD_ERR_PARAMS;
    }

    eventindex = CMD_HASH(evh->name);

    for (eventcurrent = hookEvtTable[eventindex]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(evh->name, eventcurrent->name) == 0) {      /* the msg exist's we are a addHead */
            evh->next = eventcurrent->evh;
            eventcurrent->evh = evh;
            alog(LOG_DEBUG,
                 "debug: existing msg: (0x%p), new msg (0x%p) (%s)",
                 (void *) evh->next, (void *) evh, evh->name);
            return MOD_ERR_OK;
        }
        lastHash = eventcurrent;
    }

    if ((newHash = malloc(sizeof(EvtHookHash))) == NULL) {
        fatal("Out of memory");
    }
    newHash->next = NULL;
    newHash->name = sstrdup(evh->name);
    newHash->evh = evh;

    if (lastHash == NULL)
        hookEvtTable[eventindex] = newHash;
    else
        lastHash->next = newHash;
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Add the given message (m) to the MessageHash marking it as a core command
 * @param msgTable the MessageHash we want to add to
 * @param m the Message we are adding
 * @return MOD_ERR_OK on a successful add.
 **/
int addCoreEventHook(EvtHookHash * hookEvtTable[], EvtHook * evh)
{
    if (!hookEvtTable || !evh) {
        return MOD_ERR_PARAMS;
    }
    evh->core = 1;
    return addEventHook(hookEvtTable, evh);
}

/*************************************************************************/

/**
 * Add a module message to the IRCD message hash
 * @param m the Message to add
 * @param pos the Position to add the message to, e.g. MOD_HEAD, MOD_TAIL, MOD_UNIQUE
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int moduleAddEventHandler(EvtMessage * evm)
{
    int status;

    if (!evm) {
        return MOD_ERR_PARAMS;
    }

    /* ok, this appears to be a module adding a message from outside
     * of DenoraInit, try to look up its module struct for it */
    if ((mod_current_module_name) && (!mod_current_module)) {
        mod_current_module = findModule(mod_current_module_name);
    }

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }                           /* shouldnt happen */
    evm->core = 0;
    if (!evm->mod_name) {
        evm->mod_name = sstrdup(mod_current_module->name);
    }

    status = addEventHandler(EVENT, evm);
    if (denora->debug) {
        displayEvtMessageFromHash(evm->name);
    }
    return status;
}

/*************************************************************************/

/**
 * Add a module message to the IRCD message hash
 * @param m the Message to add
 * @param pos the Position to add the message to, e.g. MOD_HEAD, MOD_TAIL, MOD_UNIQUE
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int moduleAddEventHook(EvtHook * evh)
{
    int status;

    if (!evh) {
        return MOD_ERR_PARAMS;
    }

    if ((mod_current_module_name) && (!mod_current_module)) {
        mod_current_module = findModule(mod_current_module_name);
    }

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }                           /* shouldnt happen */
    evh->core = 0;
    if (!evh->mod_name) {
        evh->mod_name = sstrdup(mod_current_module->name);
    }

    status = addEventHook(EVENTHOOKS, evh);
    if (denora->debug) {
        displayHookFromHash(evh->name);
    }
    return status;
}

/*************************************************************************/

/**
 * remove the given message from the IRCD message hash
 * @param name the name of the message to remove
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int moduleEventDelHandler(char *name)
{
    EvtMessage *evm;
    int status;

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }
    evm = findEventHandler(EVENT, name);
    if (!evm) {
        return MOD_ERR_NOEXIST;
    }

    status = delEventHandler(EVENT, evm, mod_current_module->name);
    if (denora->debug) {
        displayEvtMessageFromHash(evm->name);
    }
    return status;
}

/*************************************************************************/

/**
 * remove the given message from the IRCD message hash
 * @param name the name of the message to remove
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int moduleEventDelHook(const char *name)
{
    EvtHook *evh;
    int status;

    if (!mod_current_module) {
        return MOD_ERR_UNKNOWN;
    }
    evh = findEventHook(EVENTHOOKS, name);
    if (!evh) {
        return MOD_ERR_NOEXIST;
    }

    status = delEventHook(EVENTHOOKS, evh, mod_current_module->name);
    if (denora->debug) {
        displayHookFromHash(evh->name);
    }
    return status;
}

/*************************************************************************/

/**
 * remove the given message from the given message hash, for the given module
 * @param msgTable which MessageHash we are removing from
 * @param m the Message we want to remove
 * @mod_name the name of the module we are removing
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int delEventHandler(EvtMessageHash * msgEvtTable[], EvtMessage * evm,
                    char *mod_name)
{
    int eventindex = 0;
    EvtMessageHash *eventcurrent = NULL;
    EvtMessageHash *lastHash = NULL;
    EvtMessage *tail = NULL, *last = NULL;

    if (!evm || !msgEvtTable) {
        return MOD_ERR_PARAMS;
    }

    eventindex = CMD_HASH(evm->name);

    for (eventcurrent = msgEvtTable[eventindex]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(evm->name, eventcurrent->name) == 0) {
            if (!lastHash) {
                tail = eventcurrent->evm;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                eventcurrent->evm = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    msgEvtTable[eventindex] = eventcurrent->next;
                    denora_free(eventcurrent->name);
                    return MOD_ERR_OK;
                }
            } else {
                tail = eventcurrent->evm;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                eventcurrent->evm = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    lastHash->next = eventcurrent->next;
                    denora_free(eventcurrent->name);
                    return MOD_ERR_OK;
                }
            }
        }
        lastHash = eventcurrent;
    }
    return MOD_ERR_NOEXIST;
}

/*************************************************************************/

/**
 * remove the given message from the given message hash, for the given module
 * @param msgTable which MessageHash we are removing from
 * @param m the Message we want to remove
 * @mod_name the name of the module we are removing
 * @return MOD_ERR_OK on success, althing else on fail.
 **/
int delEventHook(EvtHookHash * hookEvtTable[], EvtHook * evh,
                 char *mod_name)
{
    int eventindex = 0;
    EvtHookHash *eventcurrent = NULL;
    EvtHookHash *lastHash = NULL;
    EvtHook *tail = NULL, *last = NULL;

    if (!evh || !hookEvtTable) {
        return MOD_ERR_PARAMS;
    }

    eventindex = CMD_HASH(evh->name);

    for (eventcurrent = hookEvtTable[eventindex]; eventcurrent;
         eventcurrent = eventcurrent->next) {
        if (stricmp(evh->name, eventcurrent->name) == 0) {
            if (!lastHash) {
                tail = eventcurrent->evh;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                eventcurrent->evh = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    hookEvtTable[eventindex] = eventcurrent->next;
                    denora_free(eventcurrent->name);
                    return MOD_ERR_OK;
                }
            } else {
                tail = eventcurrent->evh;
                if (tail->next) {
                    while (tail) {
                        if (mod_name && tail->mod_name
                            && (stricmp(mod_name, tail->mod_name) == 0)) {
                            if (last) {
                                last->next = tail->next;
                            } else {
                                eventcurrent->evh = tail->next;
                            }
                            return MOD_ERR_OK;
                        }
                        last = tail;
                        tail = tail->next;
                    }
                } else {
                    lastHash->next = eventcurrent->next;
                    denora_free(eventcurrent->name);
                    return MOD_ERR_OK;
                }
            }
        }
        lastHash = eventcurrent;
    }
    return MOD_ERR_NOEXIST;
}

/*************************************************************************/

/**
 * Destory a message, freeing its memory.
 * @param m the message to be destroyed
 * @return MOD_ERR_SUCCESS on success
 **/
int destroyEventHandler(EvtMessage * evm)
{
    if (!evm) {
        return MOD_ERR_PARAMS;
    }
    if (evm->name) {
        denora_free(evm->name);
    }
    evm->func = NULL;
    if (evm->mod_name) {
        denora_free(evm->mod_name);
    }
    evm->next = NULL;
    return MOD_ERR_OK;
}

/*************************************************************************/

/**
 * Destory a message, freeing its memory.
 * @param m the message to be destroyed
 * @return MOD_ERR_SUCCESS on success
 **/
int destroyEventHook(EvtHook * evh)
{
    if (!evh) {
        return MOD_ERR_PARAMS;
    }
    if (evh->name) {
        denora_free(evh->name);
    }
    evh->func = NULL;
    if (evh->mod_name) {
        denora_free(evh->mod_name);
    }
    evh->next = NULL;
    return MOD_ERR_OK;
}

/*************************************************************************/

#ifdef _WIN32
const char *deno_moderr(void)
{
    static char errbuf[513];
    DWORD err = GetLastError();
    if (err == 0)
        return NULL;
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS, NULL, err, 0, errbuf, 512,
                  NULL);
    return errbuf;
}
#endif

/*************************************************************************/

/** 
    * Search all loaded modules looking for a protocol module. 
    * @return 1 if one is found. 
    **/
int protocolModuleLoaded()
{
    int idx = 0;
    ModuleHash *modcurrent = NULL;

    for (idx = 0; idx != MAX_CMD_HASH; idx++) {
        for (modcurrent = MODULE_HASH[idx]; modcurrent;
             modcurrent = modcurrent->next) {
            if (modcurrent->m->type == PROTOCOL) {
                return 1;
            }
        }
    }
    return 0;
}

/*************************************************************************/

/** 
 * Module setType() 
 * Lets the module set a type, CORE,PROTOCOL,3RD etc.. 
 **/
void moduleSetType(MODType type)
{
    if ((mod_current_module_name) && (!mod_current_module)) {
        mod_current_module = findModule(mod_current_module_name);
    }
    mod_current_module->type = type;
}

/*************************************************************************/

/** 
 * Load the ircd protocol module up 
 **/
int protocol_module_init(void)
{
    int ret = 0;
    Module *m;

    if (BadPtr(IRCDModule)) {
        alog(LOG_NORMAL, "Error: IRCDModule is incorrectly defined");
        return ret;
    }

    m = createModule(IRCDModule);
    mod_current_module = m;
    mod_current_user = NULL;
    alog(LOG_NORMAL, "Loading IRCD Protocol Module: [%s]",
         mod_current_module->name);
    ret = loadModule(mod_current_module, NULL);
    moduleSetType(PROTOCOL);
    alog(LOG_EXTRADEBUG, "debug: status: [%d]", ret);
    mod_current_module = NULL;
    return ret;
}

/*************************************************************************/

/**
 * Count how many modules are loaded up
 **/
int moduleCount(void)
{
    int idx = 0;
    int counter = 0;
    ModuleHash *modcurrent = NULL;

    for (idx = 0; idx != MAX_CMD_HASH; idx++)
        for (modcurrent = MODULE_HASH[idx]; modcurrent;
             modcurrent = modcurrent->next)
            counter++;

    return counter;
}

/*************************************************************************/

/**
 * Allow a module to add a set of language strings to anope
 * @param langNumber the language number for the strings
 * @param ac The language count for the strings
 * @param av The language sring list.
 **/
void moduleInsertLanguage(int langNumber, int ac, char **av)
{
    int i;

    if ((mod_current_module_name)
        && (!mod_current_module
            || strcmp(mod_current_module_name,
                      mod_current_module->name))) {
        mod_current_module = findModule(mod_current_module_name);
    }

    alog(LOG_DEBUG, "debug: %s Adding %d texts for language %d",
         mod_current_module->name, ac, langNumber);

    if (mod_current_module->lang[langNumber].argc > 0) {
        moduleDeleteLanguage(langNumber);
    }

    mod_current_module->lang[langNumber].argc = ac;
    mod_current_module->lang[langNumber].argv =
        malloc(sizeof(char *) * ac);
    for (i = 0; i < ac; i++) {
        mod_current_module->lang[langNumber].argv[i] = sstrdup(av[i]);
    }
}

/**
 * Send a notice to the user in the correct language, or english.
 * @param source Who sends the notice
 * @param u The user to send the message to
 * @param number The message number
 * @param ... The argument list
 **/
void moduleNoticeLang(char *source, User * u, int number, ...)
{
    va_list va;
    char buffer[4096], outbuf[4096];
    char *fmt = NULL;
    int lang = StatsLanguage;
    char *s, *t, *buf;

    if ((mod_current_module_name)
        && (!mod_current_module
            || strcmp(mod_current_module_name,
                      mod_current_module->name))) {
        mod_current_module = findModule(mod_current_module_name);
    }
    /* Find the users lang, and use it if we cant */
    if (u->language) {
        lang = u->language;
    }

    /* If the users lang isnt supported, drop back to Enlgish */
    if (mod_current_module->lang[lang].argc == 0) {
        lang = LANG_EN_US;
    }

    /* If the requested lang string exists for the language */
    if (mod_current_module->lang[lang].argc > number) {
        fmt = mod_current_module->lang[lang].argv[number];

        buf = sstrdup(fmt);
        s = buf;
        while (*s) {
            t = s;
            s += strcspn(s, "\n");
            if (*s)
                *s++ = '\0';
            strscpy(outbuf, t, sizeof(outbuf));

            va_start(va, number);
            vsnprintf(buffer, 4095, outbuf, va);
            va_end(va);
            notice(source, u->nick, buffer);
        }
        free(buf);
    } else {
        alog(LOG_ERROR,
             "%s: INVALID language string call, language: [%d], String [%d]",
             mod_current_module->name, lang, number);
    }
}

/**
 * Delete a language from a module
 * @param langNumber the language Number to delete
 **/
void moduleDeleteLanguage(int langNumber)
{
    int idx = 0;
    if ((mod_current_module_name)
        && (!mod_current_module
            || strcmp(mod_current_module_name,
                      mod_current_module->name))) {
        mod_current_module = findModule(mod_current_module_name);
    }
    for (idx = 0; idx > mod_current_module->lang[langNumber].argc; idx++) {
        free(mod_current_module->lang[langNumber].argv[idx]);
    }
    mod_current_module->lang[langNumber].argc = 0;
}

/* EOF */
