/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: ircd.c 433 2005-07-26 16:58:24Z drstein $
 *
 */

#include "denora.h"

IRCDProto ircdproto;
IRCDModes ircd_modes;

/**
 * Globals we want from the protocol file 
 **/
IRCDVar *ircd;
IRCDCAPAB *ircdcap;
CBModeInfo *cbmodeinfos;
CUMode cumodes[128];
CBMode cbmodes[128];
CMMode cmmodes[128];
char csmodes[128];
unsigned long umodes[128];
int UseTSMODE;
uint32 irc_var;

/*************************************************************************/

/**
 * Initiate a protocol struct ready for use
 **/
void initIrcdProto(void)
{
    ircdproto.ircd_set_mod_current_buffer = NULL;
    ircdproto.ircd_cmd_nick = NULL;
    ircdproto.ircd_cmd_mode = NULL;
    ircdproto.ircd_cmd_bot_nick = NULL;
    ircdproto.ircd_cmd_notice = NULL;
    ircdproto.ircd_cmd_privmsg = NULL;
    ircdproto.ircd_cmd_serv_notice = NULL;
    ircdproto.ircd_cmd_serv_privmsg = NULL;
    ircdproto.ircd_cmd_351 = NULL;
    ircdproto.ircd_cmd_quit = NULL;
    ircdproto.ircd_cmd_pong = NULL;
    ircdproto.ircd_cmd_join = NULL;
    ircdproto.ircd_cmd_part = NULL;
    ircdproto.ircd_cmd_global = NULL;
    ircdproto.ircd_cmd_squit = NULL;
    ircdproto.ircd_cmd_connect = NULL;
    ircdproto.ircd_cmd_eob = NULL;
    ircdproto.ircd_cmd_ctcp = NULL;
    ircdproto.ircd_cmd_version = NULL;
    ircdproto.ircd_cmd_stats = NULL;
    ircdproto.ircd_cmd_motd = NULL;
    ircdproto.ircd_cmd_ping = NULL;
}

/*************************************************************************/

/* Special function, returns 1 if executed, 0 if not */
int denora_set_mod_current_buffer(int ac, char **av)
{
    if (ircdproto.ircd_set_mod_current_buffer) {
        ircdproto.ircd_set_mod_current_buffer(ac, av);
        return 1;
    }
    return 0;
}

/*************************************************************************/

void denora_set_umode(User * user, int ac, char **av)
{
    int add = 1;                /* 1 if adding modes, 0 if deleting */
    char *modes = av[0];

    ac--;

    if (!user || !modes) {
        /* Prevent NULLs from doing bad things */
        return;
    }

    alog(LOG_DEBUG, "debug: Changing mode for %s to %s", user->nick,
         modes);

    while (*modes) {

        add ? (user->mode |= umodes[(int) *modes]) : (user->mode &=
                                                      ~umodes[(int)
                                                              *modes]);

        switch (*modes++) {
        case '+':
            add = 1;
            break;
        case '-':
            add = 0;
            break;
        case 'd':
            if (ircd->has_svid) {
                if (ac <= 0) {
                    break;
                }
                ac--;
                av++;
                if (av) {
                    user->svid = strtoul(*av, NULL, 0);
                }
                break;
            }
        case 'o':
            if (add) {
                add_oper_count(user);
            } else {
                del_oper_count(user);
            }
            break;
        case 'H':
            if (ircd->p10 && ircd->vhost) {
                ac--;
                av++;
                change_user_host(user->nick, *av);
                break;
            }
        }
    }
}

/*************************************************************************/

void denora_motd(char *sender, char *server)
{
    ircdproto.ircd_cmd_motd(sender, server);
}

/*************************************************************************/

void denora_cmd_ping(char *server)
{
    ircdproto.ircd_cmd_ping(server);
}

/*************************************************************************/

void denora_cmd_nick(char *nick, char *name, const char *modes)
{
    char buf[BUFSIZE];
    Uid *ud = NULL;
    User *u;
    char *v[32];
    char *usermodes;

    usermodes = sstrdup(modes);
    *buf = '\0';

    ircdproto.ircd_cmd_nick(nick, name, modes);

    if (ircd->p10) {
        ud = find_uid(nick);
        u = do_nick(buf, nick, ServiceUser, ServiceHost,
                    ServerName, name, time(NULL), 0, 0, ServiceHost,
                    (ud ? ud->uid : NULL), 1, usermodes);
    } else if (ircd->ts6 && UseTS6 && Numeric) {
        ud = find_uid(nick);
        u = do_nick(buf, nick, ServiceUser, ServiceHost,
                    ServerName, name, time(NULL), 0, 0, ServiceHost,
                    (ud ? ud->uid : NULL), 1, usermodes);
    } else {
        u = do_nick(buf, nick, ServiceUser, ServiceHost,
                    ServerName, name, time(NULL), 0, 0, ServiceHost,
                    NULL, 1, usermodes);
    }
    v[0] = sstrdup(modes);
    if (u) {
        denora_set_umode(u, 1, v);
        u->isservice++;
    }
    denora_free(v[0]);
    denora_free(usermodes);
}

/*************************************************************************/

void denora_cmd_mode(char *source, char *dest, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    }
    if (!buf) {
        return;
    }
    ircdproto.ircd_cmd_mode(source, dest, buf);
}

/*************************************************************************/

void denora_cmd_bot_nick(char *nick, char *user, char *host, char *real,
                         char *modes)
{
    ircdproto.ircd_cmd_bot_nick(nick, user, host, real, modes);
}

/*************************************************************************/

void denora_cmd_notice(char *source, char *dest, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    } else {
        return;
    }
    if (!buf) {
        return;
    }
    ircdproto.ircd_cmd_notice(source, dest, buf);
}

/*************************************************************************/

void denora_cmd_notice2(char *source, char *dest, char *msg)
{
    ircdproto.ircd_cmd_notice(source, dest, msg);
}

/*************************************************************************/

void denora_cmd_privmsg(char *source, char *dest, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    }
    if (!buf) {
        return;
    }
    ircdproto.ircd_cmd_privmsg(source, dest, buf);
}

/*************************************************************************/

void denora_cmd_privmsg2(char *source, char *dest, char *msg)
{
    ircdproto.ircd_cmd_privmsg(source, dest, msg);
}

/*************************************************************************/

void denora_cmd_serv_notice(char *source, char *dest, char *msg)
{
    ircdproto.ircd_cmd_serv_notice(source, dest, msg);
}

/*************************************************************************/

void denora_cmd_serv_privmsg(char *source, char *dest, char *msg)
{
    ircdproto.ircd_cmd_serv_privmsg(source, dest, msg);
}

/*************************************************************************/

void denora_cmd_351(char *source)
{
    ircdproto.ircd_cmd_351(source);
}

/*************************************************************************/

void denora_cmd_quit(char *source, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    }
    ircdproto.ircd_cmd_quit(source, buf);
}

/*************************************************************************/

void denora_cmd_pong(char *servname, char *who)
{
    ircdproto.ircd_cmd_pong(servname, who);
}

/*************************************************************************/

void denora_cmd_join(char *user, char *channel, time_t chantime)
{
    char *v[32];
    ircdproto.ircd_cmd_join(user, channel, chantime);
    v[0] = channel;             /* channel */
    if (!LargeNet) {
        do_join(user, 1, v);
    }
    if (AutoOp) {
        v[1] = sstrdup("+o");
        v[2] = user;
        do_cmode(user, 3, v);
        denora_free(v[1]);
    }
}

/*************************************************************************/

void denora_cmd_part(char *nick, char *chan, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    char *v[32];

    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    }
    if (!nick || !chan) {
        return;
    }
    v[0] = chan;
    if (buf) {
        v[1] = buf;
    }
    ircdproto.ircd_cmd_part(nick, chan, buf);
    do_part(nick, (buf ? 2 : 1), v);
}

/*************************************************************************/

void denora_cmd_global(char *source, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    }
    if (!buf) {
        return;
    }
    ircdproto.ircd_cmd_global(source, buf);
}

/*************************************************************************/

void denora_cmd_squit(char *servname, char *message)
{
    if (!servname || !message) {
        return;
    }
    ircdproto.ircd_cmd_squit(servname, message);
}

/*************************************************************************/

void denora_cmd_connect()
{
    ircdproto.ircd_cmd_connect();
}

/*************************************************************************/

void denora_cmd_stats(char *sender, const char *letter, char *server)
{
    ircdproto.ircd_cmd_stats(sender, letter, server);
}

/*************************************************************************/

void denora_cmd_eob()
{
    ircdproto.ircd_cmd_eob();
}

/*************************************************************************/

void denora_cmd_version(char *server)
{
    ircdproto.ircd_cmd_version(server);
}

/*************************************************************************/

void denora_cmd_ctcp(char *source, char *dest, const char *fmt, ...)
{
    va_list args;
    char buf[BUFSIZE];
    char *s;

    *buf = '\0';
    if (fmt) {
        va_start(args, fmt);
        vsnprintf(buf, BUFSIZE - 1, fmt, args);
        va_end(args);
    }

    if (!buf) {
        return;
    } else {
        s = normalizeBuffer(buf);
    }
    ircdproto.ircd_cmd_ctcp(source, dest, s);
    free(s);
}

/*************************************************************************/

void
pmodule_cmd_nick(void (*func) (char *nick, char *name, const char *modes))
{
    ircdproto.ircd_cmd_nick = func;
}

/*************************************************************************/

void pmodule_cmd_version(void (*func) (char *server))
{
    ircdproto.ircd_cmd_version = func;
}

/*************************************************************************/

void pmodule_cmd_mode(void (*func) (char *source, char *dest, char *buf))
{
    ircdproto.ircd_cmd_mode = func;
}

/*************************************************************************/

void pmodule_cmd_bot_nick(void (*func)
                           (char *nick, char *user, char *host, char *real,
                            char *modes))
{
    ircdproto.ircd_cmd_bot_nick = func;
}

/*************************************************************************/

void pmodule_cmd_notice(void (*func) (char *source, char *dest, char *buf))
{
    ircdproto.ircd_cmd_notice = func;
}

/*************************************************************************/

void pmodule_set_mod_current_buffer(void (*func) (int ac, char **av))
{
    ircdproto.ircd_set_mod_current_buffer = func;
}

/*************************************************************************/

void
pmodule_cmd_privmsg(void (*func) (char *source, char *dest, char *buf))
{
    ircdproto.ircd_cmd_privmsg = func;
}

/*************************************************************************/

void
pmodule_cmd_serv_notice(void (*func) (char *source, char *dest, char *msg))
{
    ircdproto.ircd_cmd_serv_notice = func;
}

/*************************************************************************/

void pmodule_cmd_serv_privmsg(void (*func)
                               (char *source, char *dest, char *msg))
{
    ircdproto.ircd_cmd_serv_privmsg = func;
}

/*************************************************************************/

void pmodule_cmd_351(void (*func) (char *source))
{
    ircdproto.ircd_cmd_351 = func;
}

/*************************************************************************/

void pmodule_cmd_ping(void (*func) (char *server))
{
    ircdproto.ircd_cmd_ping = func;
}

/*************************************************************************/

void pmodule_cmd_motd(void (*func) (char *sender, char *server))
{
    ircdproto.ircd_cmd_motd = func;
}

/*************************************************************************/

void pmodule_cmd_quit(void (*func) (char *source, char *buf))
{
    ircdproto.ircd_cmd_quit = func;
}

/*************************************************************************/

void pmodule_cmd_pong(void (*func) (char *servname, char *who))
{
    ircdproto.ircd_cmd_pong = func;
}

/*************************************************************************/

void
pmodule_cmd_join(void (*func) (char *user, char *channel, time_t chantime))
{
    ircdproto.ircd_cmd_join = func;
}

/*************************************************************************/

void pmodule_cmd_part(void (*func) (char *nick, char *chan, char *buf))
{
    ircdproto.ircd_cmd_part = func;
}

/*************************************************************************/

void pmodule_cmd_global(void (*func) (char *source, char *buf))
{
    ircdproto.ircd_cmd_global = func;
}

/*************************************************************************/

void pmodule_cmd_squit(void (*func) (char *servname, char *message))
{
    ircdproto.ircd_cmd_squit = func;
}

/*************************************************************************/

void pmodule_cmd_stats(void (*func)
                        (char *sender, const char *letter, char *server))
{
    ircdproto.ircd_cmd_stats = func;
}

/*************************************************************************/

void pmodule_cmd_connect(void (*func) (void))
{
    ircdproto.ircd_cmd_connect = func;
}

/*************************************************************************/

void pmodule_cmd_ctcp(void (*func) (char *source, char *dest, char *buf))
{
    ircdproto.ircd_cmd_ctcp = func;
}

/*************************************************************************/

void pmodule_cmd_eob(void (*func) (void))
{
    ircdproto.ircd_cmd_eob = func;
}

/*************************************************************************/

void pmodule_ircd_var(IRCDVar * ircdvar)
{
    ircd = ircdvar;
}

/*************************************************************************/

void pmodule_ircd_cap(IRCDCAPAB * cap)
{
    ircdcap = cap;
}

/*************************************************************************/

void pmodule_ircd_version(const char *version)
{
    denora->version_protocol = sstrdup(version);
}

/*************************************************************************/

void pmodule_ircd_cbmodeinfos(CBModeInfo * modeinfos)
{
    cbmodeinfos = modeinfos;
}

/*************************************************************************/

void pmodule_ircd_cumodes(CUMode modes[128])
{
    int i = 0;
    for (i = 0; i < 128; i++) {
        cumodes[i] = modes[i];
    }
}

/*************************************************************************/

void pmodule_ircd_umodes(unsigned long modes[128])
{
    int i = 0;
    for (i = 0; i < 128; i++) {
        umodes[i] = modes[i];
    }
}

/*************************************************************************/

int denora_umode(int m)
{
    return (umodes[m] ? 1 : 0);
}

/*************************************************************************/

int denora_cmode(int m)
{
    CBMode *cbm;

    if ((cbm = &cbmodes[(int) m])->flag != 0) {
        return 1;
    } else {
        return 0;
    }
}

/*************************************************************************/

int has_cmode(char *chan, int m)
{
    CBMode *cbm;
    Channel *c;

    c = findchan(chan);

    if (!c) {
        return 0;
    }

    if ((cbm = &cbmodes[(int) m])->flag != 0) {
        if (cbm->flag & c->mode) {
            return 1;
        } else {
            return 0;
        }
    } else {
        return 0;
    }
}

/*************************************************************************/

int has_umode(char *user, int m)
{
    User *u;

    u = finduser(user);
    if (!u) {
        return 0;
    }
    if (umodes[m] & u->mode) {
        return 1;
    } else {
        return 0;
    }
}

/*************************************************************************/

void pmodule_ircd_cbmodes(CBMode modes[128])
{
    int i = 0;
    for (i = 0; i < 128; i++) {
        cbmodes[i] = modes[i];
    }
}

/*************************************************************************/

void pmodule_ircd_cmmodes(CMMode modes[128])
{
    int i = 0;
    for (i = 0; i < 128; i++) {
        cmmodes[i] = modes[i];
    }
}

/*************************************************************************/

void pmodule_ircd_csmodes(char mode[128])
{
    int i = 0;
    for (i = 0; i < 128; i++) {
        csmodes[i] = mode[i];
    }
}

/*************************************************************************/

void pmodule_ircd_useTSMode(int use)
{
    UseTSMODE = use;
}

/*************************************************************************/

void pmodule_oper_umode(int mode)
{
    ircd_modes.user_oper = mode;
}

/*************************************************************************/

int denora_get_oper_mode()
{
    return ircd_modes.user_oper;
}

/*************************************************************************/

void pmodule_irc_var(uint32 mode)
{
    ircd_modes.ircd_var = mode;
}

/*************************************************************************/

uint32 denora_get_ircd()
{
    return ircd_modes.ircd_var;
}

/*************************************************************************/
