/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 *
 * $Id: messages.c 425 2005-07-23 04:45:20Z trystan $
 *
 */

#include "denora.h"
#include "modules.h"

/*************************************************************************/

int m_nickcoll(char *user)
{
    if (nickIsServices(user)) {
        introduce_user(user);
    }
    return MOD_CONT;
}

/*************************************************************************/

int m_away(char *source, char *msg)
{
    if (!source) {
        return MOD_CONT;
    }
    if (denora->do_sql) {
        source = rdb_escape(source);
    }
    if (msg) {
        stats->away++;
        if (denora->do_sql) {
            msg = rdb_escape(msg);
            rdb_query
                ("UPDATE %s SET away=\'Y\', awaymsg=\'%s\' WHERE nickid=%d",
                 UserTable, msg, db_getnick(source));
            denora_free(msg);
        }
    } else {
        stats->away--;
        if (denora->do_sql) {
            rdb_query
                ("UPDATE %s SET away=\'N\', awaymsg=\'\' WHERE nickid=%d",
                 UserTable, db_getnick(source));
        }
    }
    if (denora->do_sql) {
        denora_free(source);
    }
    return MOD_CONT;
}

/*************************************************************************/

void m_kill(char *source, char *nick, char *msg)
{
    User *u;
    ServStats *s;
    int id;

    if (nickIsServices(nick)) {
        introduce_user(nick);
    } else {
        do_kill(nick, msg);
        if (denora->do_sql) {
            nick = rdb_escape(nick);
            db_removenick(nick);
            if (UserCacheTime) {
                db_cleanuser();
            }
            denora_free(nick);
        }
    }
    if ((u = finduser(source))) {
        if ((s = findserverstats(u->server->name))) {
            s->ircopskills++;
        }
    } else if ((s = findserverstats(source))) {
        s->serverkills++;
    }

    if (s && denora->do_sql) {
        id = db_getserver(s->name);
        if (id) {
            rdb_query
                ("UPDATE %s SET ircopskills=%d, serverkills=%ld WHERE servid=%d",
                 ServerTable, s->ircopskills, s->serverkills, id);
        }
    }
}

/*************************************************************************/

int m_time(char *source, int ac, char **av)
{
    time_t t;
    struct tm *tm;
    char buf[64];

    *buf = '\0';

    if (!source) {
        return MOD_CONT;
    }

    time(&t);
    tm = localtime(&t);
    strftime(buf, 63, "%a %b %d %H:%M:%S %Y %Z", tm);
    denora_cmd_391(source, buf);
    /* avoid compiler warnings */
    if (!ac || !av) {
        return MOD_CONT;
    }
    return MOD_CONT;
}

/*************************************************************************/

int m_motd(char *source)
{
    FILE *f;
    char buf[BUFSIZE];

    *buf = '\0';

    if (!source) {
        return MOD_CONT;
    }

    f = fopen(MOTDFilename, "r");
    denora_cmd_375(source);
    if (f) {
        while (fgets(buf, BUFSIZE - 1, f)) {
            buf[strlen(buf) - 1] = 0;
            denora_cmd_372(source, buf);
        }
        fclose(f);
    } else {
        denora_cmd_422(source);
    }
    denora_cmd_376(source);
    return MOD_CONT;
}

/*************************************************************************/

int m_notice(char *source, char *receiver, char *msg)
{
    char *temp = NULL;
    char *version;
    char *clean;

    if (!stricmp(receiver, s_StatServ)) {
        clean = normalizeBuffer(msg);
        doCleanBuffer((char *) clean);
        temp = myStrGetToken(clean, ' ', 0);
        if (!temp) {
            return MOD_CONT;
        }
        if (!stricmp(temp, "VERSION")) {
            version = myStrGetTokenRemainder(clean, ' ', 1);
            handle_ctcp_version(source, version);
            denora_free(version);
        }
        denora_free(clean);
        denora_free(temp);
    }
    return MOD_CONT;
}

/*************************************************************************/

int m_privmsg(char *source, char *receiver, char *msg)
{
    char *s;
    User *u;
    Exclude *e;

    if (!source || !*source) {
        return MOD_CONT;
    }

    u = finduser(source);

    if (!u) {
        alog(LOG_NONEXISTANT, "%s: user record for %s not found", msg,
             source);
        denora_cmd_notice(msg, source,
                          getstring(NULL, USER_RECORD_NOT_FOUND));
        return MOD_CONT;
    }

    if (!*receiver || !receiver || !msg) {
        return MOD_CONT;
    }

    if (*receiver == '#') {
        if (denora->do_sql && !LargeNet) {
            e = find_exclude(u->nick, u->server->name);
            if (!e) {
                if (stricmp(s_StatServ, u->nick)) {
                    do_cstats(u, receiver, msg);
                }
            }
        }
    } else {

        /* If a server is specified (nick@server format), make sure it matches
         * us, and strip it off. */
        s = strchr(receiver, '@');
        if (s) {
            *s++ = 0;
            if (stricmp(s, ServerName) != 0)
                return MOD_CONT;
        }

        if ((stricmp(receiver, s_StatServ) == 0)) {
            SET_SEGV_LOCATION();
            statserv(u, msg);
        }
    }
    return MOD_CONT;
}

/*************************************************************************/

int m_stats(char *source, int ac, char **av)
{
    User *u;
    aAdmin *aadmin;
    if (ac < 1)
        return MOD_CONT;

    switch (*av[0]) {
    case 'l':
        u = finduser(source);

        if (u && is_oper(u)) {

            denora_cmd_numeric
                (211, "%s Server SendBuf SentBytes SentMsgs RecvBuf "
                 "RecvBytes RecvMsgs ConnTime", source);
            denora_cmd_numeric(211, "%s %s %d %d %d %d %d %d %ld",
                               source, RemoteServer,
                               write_buffer_len(), total_written,
                               total_recmsg, read_buffer_len(), total_read,
                               total_sendmsg,
                               time(NULL) - denora->start_time);
        }

        denora_cmd_219(source, NULL);
        break;
    case 'o':
    case 'O':
/* Check whether the user is an operator */
        u = finduser(source);
        if (u && !is_oper(u) && HideStatsO) {
            denora_cmd_219(source, av[0]);
        } else {
            for (aadmin = admins; aadmin; aadmin = aadmin->next)
                denora_cmd_numeric(243, "%s O * * %s Admin 0", source,
                                   aadmin->name);
            denora_cmd_219(source, av[0]);
        }

        break;

    case 'u':{
            int uptime = time(NULL) - denora->start_time;
            denora_cmd_numeric(242,
                               "%s :Stats up %d day%s, %02d:%02d:%02d",
                               source, uptime / 86400,
                               (uptime / 86400 == 1) ? "" : "s",
                               (uptime / 3600) % 24, (uptime / 60) % 60,
                               uptime % 60);
            denora_cmd_numeric(250,
                               "%s :Current users: %d (%d ops); maximum %d",
                               source, stats->users, stats->opers,
                               stats->users_max);
            denora_cmd_219(source, NULL);
            break;
        }                       /* case 'u' */

    default:
        denora_cmd_219(source, av[0]);
        break;
    }
    return MOD_CONT;
}

/*************************************************************************/

int m_version(char *source, int ac, char **av)
{
    if (source) {
        denora_cmd_351(source);
    }
    /* avoid compiler warnings */
    if (!ac || !av) {
        return MOD_CONT;
    }
    return MOD_CONT;
}


/*************************************************************************/

int m_whois(char *source, char *who)
{
    const char *clientdesc;

    if (source && who) {
        if (stricmp(who, s_StatServ) == 0) {
            clientdesc = desc_StatServ;
        } else if (s_StatServ_alias && stricmp(who, s_StatServ_alias) == 0) {
            clientdesc = desc_StatServ_alias;
        } else {
            denora_cmd_401(source, who);
            return MOD_CONT;
        }
        denora_cmd_numeric(311, "%s %s %s %s * :%s", source, who,
                           ServiceUser, ServiceHost, clientdesc);
        /* denora_cmd_219(source, who); */
        denora_cmd_numeric(317, "%s %s %ld %ld :seconds idle, signon time",
                           source, who, time(NULL) - denora->start_time,
                           denora->start_time);
        denora_cmd_318(source, who);
    }
    return MOD_CONT;
}

/*************************************************************************/

/* NULL route messages */
int denora_event_null(char *source, int ac, char **av)
{
    if (denora->protocoldebug) {
        protocol_debug(source, ac, av);
    }
    return MOD_CONT;
}

/*************************************************************************/

/* *INDENT-OFF* */
void moduleAddMsgs(void) 
{
    Message *m;
    m = createMessage("STATS",     m_stats); addCoreMessage(IRCD,m);
    m = createMessage("TIME",      m_time); addCoreMessage(IRCD,m);
    m = createMessage("VERSION",   m_version); addCoreMessage(IRCD,m);

    m = createMessage("020",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("005",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("105",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("217",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("219",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("242",       denora_event_242); addCoreMessage(IRCD,m);
    m = createMessage("244",       denora_event_null); addCoreMessage(IRCD,m);

    /* 246 is SolidIRC's U:line */
    m = createMessage("246",       denora_event_246); addCoreMessage(IRCD,m);
    /* Most everyone else uses 248 */
    m = createMessage("248",       denora_event_248); addCoreMessage(IRCD,m);

    m = createMessage("249",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("250",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("351",       denora_event_351); addCoreMessage(IRCD,m);
    m = createMessage("372",       denora_event_372); addCoreMessage(IRCD,m);
    m = createMessage("375",       denora_event_375); addCoreMessage(IRCD,m);
    m = createMessage("376",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("401",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("402",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("451",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("461",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("481",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("422",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("482",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("503",       denora_event_null); addCoreMessage(IRCD,m);

    /* Ithildin Adds some replies */
    m = createMessage("771",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("773",       denora_event_null); addCoreMessage(IRCD,m);
    m = createMessage("774",       denora_event_null); addCoreMessage(IRCD,m);
}

/*************************************************************************/
