
/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: servers.c 431 2005-07-26 03:32:27Z trystan $
 *
 */

#include "denora.h"

Server *servlist = NULL;
Server *me_server = NULL;
char *uplink;
char *TS6SID;
char p10id[3];
char *TS6UPLINK;
Server *serverlist[1024];
ServStats *servstatlist[1024];
static Server *current;
static ServStats *currentss;
static int next_index;

/*************************************************************************/

void p10_gline(char *source, char **av)
{
    Server *s;
    char buf[BUFSIZE];
    char *user;
    char *host;
    char *address;

    *buf = '\0';
    s = findserver_uid(source);

    address = myStrGetToken(av[1], '!', 1);
    user = myStrGetToken(address, '@', 0);
    host = myStrGetToken(address, '@', 1);

    if (*av[1] == '+') {
      snprintf(buf, BUFSIZE - 1, "%ld", (long int) time(NULL));
      
      sql_do_server_bans_add(NULL, user, host,
                           (s ? s->name : source), buf, av[2],
                            av[3]);
    } else {
        sql_do_server_bans_remove(NULL, user, host);
    }
}

/*************************************************************************/

void update_sync_state(char *source, int ac)
{
    Server *s;
    if (!BadPtr(source)) {
        s = findserver(source);
    } else {
        s = findserver(denora->uplink);
    }
    if (!ac) {
        /* for future use  - start burst */
    } else {
        if (s) {
            s->sync = 1;
        }
    }    
}

/*************************************************************************/

void load_server_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    ServStats *ss = NULL;
    char *key, *value;
    int retval = 0;

    fill_db_ptr(dbptr, 0, SERVER_VERSION, s_StatServ, ServerDB);

    /* let's remove existing temp files here, because we only load dbs on startup */
    remove(dbptr->temp_name);

    /* Open the db, fill the rest of dbptr and allocate memory for key and value */
    if (new_open_db_read(dbptr, &key, &value)) {
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }

    while (1) {
        /* read a new entry and fill key and value with it -Certus */
        retval = new_read_db_entry(&key, &value, dbptr->fptr);

        if (retval == DB_READ_ERROR) {
            alog(LOG_NORMAL, langstr(ALOG_DB_ERROR), dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            free(dbptr);
            return;
        } else if (retval == DB_EOF_ERROR) {
            alog(LOG_EXTRADEBUG, langstr(ALOG_DEBUG_DB_OK), dbptr->filename);
            new_close_db(dbptr->fptr, &key, &value);
            free(dbptr);
            return;
        } else if (retval == DB_READ_BLOCKEND) {        /* DB_READ_BLOCKEND */
            /* a ctcp entry has completely been read. put any checks in here! */
        } else {
            /* DB_READ_SUCCESS */

            if (!*value || !*key)
                continue;

            if (!stricmp(key, "name")) {
                ss = make_servstats(value);
            } else if (!stricmp(key, "maxusers")) {
                ss->maxusers = atoi(value);
            } else if (!stricmp(key, "maxusertime")) {
                ss->maxusertime = atoi(value);
            } else if (!stricmp(key, "maxopers")) {
                ss->maxopers = atoi(value);
            } else if (!stricmp(key, "highestping")) {
                ss->highestping = atoi(value);
            } else if (!stricmp(key, "maxpingtime")) {
                ss->maxpingtime = atoi(value);
            } else if (!stricmp(key, "totalusersever")) {
                ss->totalusersever = atoi(value);
            } else if (!stricmp(key, "splits")) {
                ss->splits = atoi(value);
            } else if (!stricmp(key, "split_stats")) {
                ss->split_stats = 1;
            } else if (!stricmp(key, "lastseen")) {
                ss->lastseen = atoi(value);
            } else if (!stricmp(key, "ircopskills")) {
                ss->ircopskills = atoi(value);
            } else if (!stricmp(key, "serverkills")) {
                ss->serverkills = atoi(value);
            } else if (!stricmp(key, "maxopertime")) {
                ss->maxopertime = atoi(value);
            }


        }                       /* else */
    }                           /* while */
}

/*************************************************************************/

void save_server_db(void)
{
    DenoraDBFile *dbptr = scalloc(1, sizeof(DenoraDBFile));
    ServStats *ss;
    int i;

    fill_db_ptr(dbptr, 0, SERVER_VERSION, s_StatServ, ServerDB);

    /* time to backup the old db */
    rename(ServerDB, dbptr->temp_name);

    if (new_open_db_write(dbptr)) {
        rename(dbptr->temp_name, ServerDB);
        free(dbptr);
        return;                 /* Bang, an error occurred */
    }

    for (i = 0; i < 1024; i++) {
        for (ss = servstatlist[i]; ss; ss = ss->next) {
            new_write_db_entry("name", dbptr, "%s", ss->name);
            new_write_db_entry("maxusers", dbptr, "%d", ss->maxusers);
            new_write_db_entry("maxopers", dbptr, "%d", ss->maxopers);
            new_write_db_entry("highestping", dbptr, "%d",
                               ss->highestping);
            new_write_db_entry("maxpingtime", dbptr, "%d",
                               ss->maxpingtime);
            new_write_db_entry("totalusersever", dbptr, "%ld",
                               ss->totalusersever);
            new_write_db_entry("splits", dbptr, "%d", ss->splits);
            new_write_db_entry("split_stats", dbptr, "%ld",
                               ss->split_stats);
            new_write_db_entry("lastseen", dbptr, "%d", ss->lastseen);
            new_write_db_entry("maxusertime", dbptr, "%ld", ss->maxusertime);
            new_write_db_entry("ircopskills", dbptr, "%ld", ss->ircopskills);
            new_write_db_entry("serverkills", dbptr, "%ld", ss->serverkills);
            new_write_db_entry("maxopertime", dbptr, "%ld", ss->maxopertime);
            new_write_db_endofblock(dbptr);
        }
    }

    if (dbptr) {
        new_close_db(dbptr->fptr, NULL, NULL);  /* close file */
        remove(dbptr->temp_name);       /* saved successfully, no need to keep the old one */
        free(dbptr);            /* free the db struct */
    }
}

/*************************************************************************/

void sql_motd_store(char *source, char *string)
{
    char buf[NET_BUFSIZE];
    Server *s;
    char *string2;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif
    *buf = '\0';

    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }
    if (!denora->do_sql) {
        return;
    }

    if (ircd->p10) {
        s = findserver_uid(source);
    } else if (ircd->ts6 && UseTS6) {
        s = findserver_uid(source);
        if (!s) {
            s = findserver(source);
        }
    } else {
        s = findserver(source);
    }

    if (!s) {
        return;
    }

    string2 = sstrdup(string);
    source = rdb_escape(s->name);
    string = rdb_escape(string2);
    denora_free(string2);

    rdb_query("SELECT motd FROM %s WHERE server=\"%s\"", ServerTable, source);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
      if (mysql_num_rows(mysql_res)) {
        mysql_row = mysql_fetch_row(mysql_res);
        if (mysql_row[0]) {
          snprintf(buf, NET_BUFSIZE - 1, "%s\n\r%s", mysql_row[0], string);
          denora_free(string);
          string = rdb_escape(buf);
          rdb_query("UPDATE %s SET motd=\"%s\" WHERE server=\"%s\"", ServerTable, string, source);
        } else {
          rdb_query("UPDATE %s SET motd=\"%s\" WHERE server=\"%s\"", ServerTable, string, source);
        }
        denora_free(string);
      } else {
        rdb_query("UPDATE %s SET motd=\"%s\" WHERE server=\"%s\"", ServerTable, string, source);
      }
      mysql_free_result(mysql_res);
    } else {
        rdb_query("UPDATE %s SET motd=\"%s\" WHERE server=\"%s\"", ServerTable, string, source);
    }
#endif
    denora_free(source);
    denora_free(string);
}


/*************************************************************************/

/**
 * This function makes a new Server structure 
 * @param name Server Name
 * @return Server Struct
 */
ServStats *make_servstats(const char *server)
{
    ServStats *s, **list;

    SET_SEGV_LOCATION();
    s = scalloc(sizeof(ServStats), 1);
    if (!server) {
        server = "";
    }
    strscpy(s->name, server, SERVERMAX);
    list = &servstatlist[STATSSERVERHASH(s->name)];
    s->next = *list;
    if (*list)
        (*list)->prev = s;
    *list = s;
    s->currentusers = 0;
    return s;
}

/*************************************************************************/

/**
 * This function makes a new Server structure 
 * @param name Server Name
 * @return Server Struct
 */
static Server *make_server(const char *server)
{
    Server *s, **list;

    SET_SEGV_LOCATION();
    s = scalloc(sizeof(Server), 1);
    if (!server) {
        server = "";
    }
    strscpy(s->name, server, SERVERMAX);
    list = &serverlist[SERVERHASH(s->name)];
    s->next = *list;
    if (*list)
        (*list)->prev = s;
    *list = s;
    return s;
}

/*************************************************************************/

Server *do_server(const char *source, char *servername, char *hops,
                  char *descript, char *numeric)
{
    char *uplinkserver;
    char *sqluplinkserver;
    Server *serv;
    char buf[BUFSIZ];
    ServStats *ss;
    Server *servuplink;
    int servid;
    int add = 1;
    uint32 timenow, uptime;
    FILE *f;
    *buf = '\0';

    SET_SEGV_LOCATION();

    if (!source || !*source) {
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_SERVINTRO), servername);
        uplinkserver = sstrdup(ServerName);
        servuplink = findserver(ServerName);
    } else {
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_SERVINTRO_FORM), servername, source);
        uplinkserver = sstrdup(source);
        servuplink = findserver(source);
    }

    serv = findserver(servername);
    if (!serv) {
        SET_SEGV_LOCATION();

        /* internal struct started for the server */
        serv = make_server(servername);
        if (!BadPtr(descript)) {
            serv->desc = sstrdup(descript);
        }
        serv->uplink = sstrdup(uplinkserver);
        if (numeric) {
            serv->suid = sstrdup(numeric);
        } else {
            serv->suid = NULL;
        }
        serv->sync = -1;
        serv->links = NULL;
        ss = findserverstats(servername);
        SET_SEGV_LOCATION();
        if (!ss) {
            serv->ss = make_servstats(servername);
        } else {
            serv->ss = ss;
            serv->ss->split_stats = 0;
        }
        if (!serv->ss->maxusertime) {
                serv->ss->maxusertime = time(NULL);
        }
        SET_SEGV_LOCATION();

        /* get some stats */
        if (stricmp(servername, ServerName) && !LargeNet) {
              denora_cmd_version(servername);
              denora_cmd_stats(s_StatServ, "u", servername);
              if (ircd->uline) {
                  denora_cmd_stats(s_StatServ, "U", servername);
              }
              denora_motd(s_StatServ, servername);
        }
    }
    stats->servers++;
    SET_SEGV_LOCATION();

    send_event(EVENT_SERVER, 1, servername);

    SET_SEGV_LOCATION();
    if (denora->do_sql) {
        servername = rdb_escape(servername);
        descript = rdb_escape(descript);
        sqluplinkserver = rdb_escape(uplinkserver);
        if (ServerCacheTime
            && ((servid = db_checkserver(servername)) != -1)) {
            rdb_query
                ("UPDATE %s SET server=\"%s\", hops=\"%s\", comment=\"%s\", linkedto=%d, connecttime=NOW(), online=\"Y\", maxusers=%d, maxusertime=%d WHERE servid=%d",
                 ServerTable, servername, hops, descript,
                 db_getserver(sqluplinkserver), serv->ss->maxusers, serv->ss->maxusertime, servid);
            add = 0;
        }
        if (add) {
            rdb_query
                ("INSERT INTO %s (server, hops, comment, linkedto, connecttime, maxusers, maxusertime) VALUES(\'%s\',\'%s\',\'%s\',%d, NOW(), %d, %d)",
                 ServerTable, servername, hops, descript,
                 db_getserver(sqluplinkserver), serv->ss->maxusers, serv->ss->maxusertime);
        }
        denora_free(descript);
        denora_free(sqluplinkserver);
        servid = db_getserver(servername);
        serv->sqlid = servid;

        if (!stricmp(servername, ServerName)) {
            timenow = (long int) time(NULL);
            uptime  = timenow - denora->start_time;
            *buf = '\0';
            snprintf(buf, sizeof(buf), "Denora-%s", denora->version);
            serv->version = sstrdup(buf);

            rdb_query("UPDATE %s SET uptime=%ld, version=\"%s\" WHERE servid=%d",
              ServerTable, uptime, serv->version, servid);
            f = fopen(MOTDFilename, "r");
            if (f) {
             while (fgets(buf, BUFSIZE - 1, f)) {
                buf[strlen(buf) - 1] = 0;
                sql_motd_store(ServerName, buf);
             }
             fclose(f);
			}
        }
        denora_free(servername);
    }

    SET_SEGV_LOCATION();
    if (ServerCacheTime) {
        db_cleanserver();
    }
    denora_free(uplinkserver);
    SET_SEGV_LOCATION();
    do_checkservsmax();
    servers_hourly(NULL);
    return serv;
}

/*************************************************************************/

void sql_uline(char *server)
{
    int id;
    Server *serv;

    /* On Ratbox - ulines are also shared so you can
       get *.your.net at times so we ignore any server
       that starts with * */
    if (*server != '*') {
     id = db_getserver(server);
     if (id) {
        rdb_query("UPDATE %s SET uline=1 WHERE servid=%d", ServerTable, id);
     }
     serv = findserver(server);
     if (serv) {
        serv->uline = 1;
     }
    }
}

/*************************************************************************/

void server_store_pong(char *source, uint32 ts)
{
    Server *s;
    char *serv;

    if (LargeNet) {
        return;
    }

    s = findserver(source);
    if (s) {
        s->ping = ts - s->lastping;
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_PINGTIME), source, s->ping, ts, s->lastping);
        if (s->ping >= s->ss->highestping) {
            s->ss->highestping = s->ping;
            s->ss->maxpingtime = time(NULL);
        }
	    if (denora->do_sql) {
         serv = rdb_escape(source);
         rdb_query("UPDATE %s SET ping=%d, highestping=%d, maxpingtime=%ld WHERE servid=%d",
              ServerTable, s->ping, s->ss->highestping, (long int) s->ss->maxpingtime, db_getserver(serv));
         denora_free(serv);
	    }
    }
}

/*************************************************************************/

void ping_servers()
{
    Server *s, *next;
    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }

    s = first_server();
    while (s) {
        next = next_server();
        if (stricmp(ServerName, s->name)) {
            denora_cmd_ping(s->name);
            s->lastping = time(NULL);
        }
        s = next;
    }
}

/*************************************************************************/

void uptime_servers()
{
    Server *s, *next;
    uint32 timenow, uptime;
    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }

    s = first_server();
    while (s) {
        next = next_server();
        if (stricmp(ServerName, s->name)) {
            denora_cmd_stats(s_StatServ, "u", s->name);
        }
        s = next;
    }
    timenow = (long int) time(NULL);
    uptime  = timenow - denora->start_time;

    rdb_query("UPDATE %s SET uptime=%ld WHERE servid=%d",
              ServerTable, uptime, db_getserver(ServerName));
}

/*************************************************************************/

/**
 * Find a server by name, returns NULL if not found
 * @param s Server struct
 * @param name Server Name
 * @return Server struct
 */
Server *findserver(const char *name)
{
    Server *s;

    SET_SEGV_LOCATION();
    if (BadPtr(name)) {
        alog(LOG_DEBUG, langstr(ALOG_ERR_FINDSERVER));
        do_backtrace(0);
        return NULL;
    }

    alog(LOG_EXTRADEBUG, "debug: findserver(%s) -> 0x%p", name, name);
    s = serverlist[SERVERHASH(name)];
    while (s && stricmp(s->name, name) != 0) {
        s = s->next;
    }
    alog(LOG_EXTRADEBUG, "debug: findserver(%s) -> 0x%p", name,
         (void *) s);
    return s;
}

/*************************************************************************/

/**
 * Find a server by name, returns NULL if not found
 * @param s Server struct
 * @param name Server Name
 * @return Server struct
 */
ServStats *findserverstats(const char *name)
{
    ServStats *s;

    SET_SEGV_LOCATION();
    if (!name || !*name) {
        alog(LOG_DEBUG, langstr(ALOG_ERR_FINDSERVERSTATS));
        do_backtrace(1);
        return NULL;
    }

    alog(LOG_EXTRADEBUG, "debug: findserverstats(%s) -> 0x%p", name, name);
    s = servstatlist[STATSSERVERHASH(name)];
    SET_SEGV_LOCATION();

    while (s && stricmp(s->name, name) != 0) {
        s = s->next;
    }
    SET_SEGV_LOCATION();

    alog(LOG_EXTRADEBUG, "debug: findserverstats(%s) -> 0x%p", name,
         (void *) s);
    return s;
}


/*************************************************************************/

/**
 * Find a server by name, returns NULL if not found
 * @param s Server struct
 * @param name Server Name
 * @return Server struct
 */
Server *findserver_uid(const char *name)
{
    Server *s, *s2;

    SET_SEGV_LOCATION();
    if (!name || !*name) {
        alog(LOG_DEBUG, langstr(ALOG_ERR_FINDSERVERSUID));
        return NULL;
    }

    s = first_server();
    alog(LOG_EXTRADEBUG, "debug: findserver_suid(%s) -> 0x%p", name, name);
    while (s) {
        s2 = next_server();
        if (s->suid && !stricmp(s->suid, name)) {
            return s;
        }
        s = s2;
    }
    alog(LOG_EXTRADEBUG, "debug: findserver(%s) -> 0x%p", name,
         (void *) s);
    return NULL;
}

/*************************************************************************/

/**
 * Remove and denora_free a Server structure. This function is the most complete
 * remove treatment a server can get, as it first quits all clients which
 * still pretend to be on this server, then it walks through all connected
 * servers and disconnects them too. If all mess is cleared, the server
 * itself will be too.
 * @param Server struct
 * @param reason the server quit
 * @return void
 */

static void delete_serverstats(ServStats * serv)
{
    if (!serv) {
        alog(LOG_DEBUG, langstr(ALOG_ERR_DEL_SERVSTATS));
        return;
    }

    alog(LOG_EXTRADEBUG, langstr(ALOG_DEL_SERVSTATS_FOR), serv->name);

    if (serv->prev) {
        serv->prev->next = serv->next;
    } else {
        servstatlist[STATSSERVERHASH(serv->name)] = serv->next;
    }
    if (serv->next) {
        serv->next->prev = serv->prev;
    }

    alog(LOG_DEBUG, langstr(ALOG_DEL_SERVSTATS_DONE));
    denora_free(serv);
}

/*************************************************************************/

/**
 * Remove and free a Server structure. This function is the most complete
 * remove treatment a server can get, as it first quits all clients which
 * still pretend to be on this server, then it walks through all connected
 * servers and disconnects them too. If all mess is cleared, the server
 * itself will be too.
 * @param Server struct
 * @param reason the server quit
 * @return void
 */
void delete_server(Server * serv, const char *quitreason)
{
    User *u, *unext;
    Server *s, *snext;

    if (!serv) {
        alog(LOG_DEBUG, langstr(ALOG_ERR_DEL_SERVER));
        return;
    }

    alog(LOG_DEBUG, langstr(ALOG_DEL_SERVER_FOR), serv->name);

    if (ircdcap->noquit) {
        if (denora->capab & ircdcap->noquit) {
            alog(LOG_EXTRADEBUG, langstr(ALOG_DEL_SERVER_NOQUIT));
            u = firstuser();
            while (u) {
                unext = nextuser();
                if (u->server == serv) {
                    delete_user(u);
                }
                u = unext;
            }
        }
    }

    s = first_server();
    while (s) {
        snext = next_server();
        if (stricmp(s->name,ServerName)) {
         if (s->uplink) {
          if (!stricmp(s->uplink, serv->name)) {
            delete_server(s, quitreason);
          }
         }
        }
        s = snext;
    }

    denora_free(serv->desc);
    denora_free(serv->uplink);
    delete_serverstats(serv->ss);
    servers_hourly(NULL);
    if (serv->version) {
        denora_free(serv->version);
    }
    if (serv->suid) {
        denora_free(serv->suid);
    }
    if (serv->prev) {
        serv->prev->next = serv->next;
    } else {
        serverlist[SERVERHASH(serv->name)] = serv->next;
    }
    if (serv->next) {
        serv->next->prev = serv->prev;
    }
    if (serv->links) {
        serv->links = NULL;
    }
    alog(LOG_DEBUG, langstr(ALOG_DEL_SERVER_DONE));
}

/*************************************************************************/

Server *first_server(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && current == NULL)
        current = serverlist[next_index++];
    alog(LOG_EXTRADEBUG, "debug: first_server() returning %s",
         current ? current->name : "NULL (end of list)");
    return current;
}

/*************************************************************************/

Server *next_server(void)
{
    SET_SEGV_LOCATION();

    if (current)
        current = current->next;
    if (!current && next_index < 1024) {
        while (next_index < 1024 && current == NULL)
            current = serverlist[next_index++];
    }
    alog(LOG_EXTRADEBUG, "debug: next_server() returning %s",
         current ? current->name : "NULL (end of list)");
    return current;
}

/*************************************************************************/

ServStats *first_statsserver(void)
{
    next_index = 0;

    SET_SEGV_LOCATION();

    while (next_index < 1024 && currentss == NULL)
        currentss = servstatlist[next_index++];
    alog(LOG_EXTRADEBUG, "debug: first_statserver() returning %s",
         currentss ? currentss->name : "NULL (end of list)");
    return currentss;
}

/*************************************************************************/

ServStats *next_statserver(void)
{
    SET_SEGV_LOCATION();

    if (currentss)
        currentss = currentss->next;
    if (!currentss && next_index < 1024) {
        while (next_index < 1024 && currentss == NULL)
            currentss = servstatlist[next_index++];
    }
    alog(LOG_EXTRADEBUG, "debug: next_statserver() returning %s",
         currentss ? currentss->name : "NULL (end of list)");
    return currentss;
}

/*************************************************************************/

void sql_do_sqline(char *mask, char *reason)
{
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    if (!denora->do_sql) {
        return;
    }

    if (LargeNet) {
        return;
    }

    mask = rdb_escape(mask);
    reason = rdb_escape(reason);

    rdb_query
        ("SELECT mask FROM %s WHERE mask = \"%s\";", SqlineTable, mask);
#ifdef USE_MYSQL
    mysql_res = mysql_store_result(mysql);
    if (mysql_res) {
        if (mysql_num_rows(mysql_res) == 0) {   /* exists this nick already in the database? */
            mysql_free_result(mysql_res);
            rdb_query
                ("INSERT INTO %s (mask, reason) values('%s', '%s')",
                 SqlineTable, mask, reason);
        }
    }
#endif
    denora_free(mask);
    denora_free(reason);
}

/*************************************************************************/

void sql_do_sgline(char *length, char *mask)
{
    int len;
    char *reason;
#ifdef USE_MYSQL
    MYSQL_RES *mysql_res;
#endif

    if (!denora->do_sql) {
	    return;
    }
    if (LargeNet) {
        return;
    }

    len = strtol(length, NULL, 10);
    if (((int) strlen(mask) > len) && (mask[len]) == ':') {
        mask[len] = '\0';
        reason = mask + len + 1;
    } else {
        return;
    }
#ifdef USE_MYSQL
    rdb_query
        ("SELECT mask FROM %s WHERE mask = \"%s\";", SglineTable, mask);
    mysql_res = mysql_store_result(mysql);
    if (mysql_num_rows(mysql_res) == 0) {       /* exists this nick already in the database? */
        mysql_free_result(mysql_res);
        rdb_query
            ("INSERT INTO %s (mask, reason) values('%s', '%s')",
             SglineTable, mask, reason);
    }
#endif
    return;
}

/*************************************************************************/

void sql_do_uptime(char *source, char *uptime)
{
    char *tmp, *tmp4;
    char *days;
    int32 days_int;
    char *hours;
    int32 hours_int;
    char *mins;
    int32 mins_int;
    char *secs;
    int32 secs_int;
    int32 total = 0;
    Server *s;

    if (LargeNet) {
        return;
    }

    if (ircd->p10) {
        s = findserver_uid(source);
    } else if (ircd->ts6 && UseTS6) {
        s = findserver_uid(source);
        if (!s) {
            s = findserver(source);
        }
    } else {
        s = findserver(source);
    }

    if (!s) {
	    alog(LOG_DEBUG, "Nonexistant server (%s) uptime", source);
	    return;
    }

    tmp = myStrGetToken(uptime, ' ', 0);

    /* Thales and NeoStats send more then the rest */
    if (!stricmp(tmp, "GNU") || !stricmp(tmp, "Statistical")) {
        days = myStrGetToken(uptime, ' ', 3);
        tmp4 = myStrGetTokenRemainder(uptime, ' ', 5);

    } else {
        days = myStrGetToken(uptime, ' ', 2);
        tmp4 = myStrGetTokenRemainder(uptime, ' ', 4);
    }

    hours = myStrGetToken(tmp4, ':', 0);
    mins = myStrGetToken(tmp4, ':', 1);
    secs = myStrGetToken(tmp4, ':', 2);

    mins_int = atol(mins) * 60;
    secs_int = atol(secs);
    hours_int = atol(hours) * 3600;
    days_int = atol(days) * 86400;

    total = mins_int + secs_int + hours_int + days_int;
    rdb_query("UPDATE %s SET uptime=%ld WHERE servid=%d",
              ServerTable, total, db_getserver((s ? s->name : source)));

    if (s) {
        s->uptime = total;
    }

    denora_free(days);
    denora_free(hours);
    denora_free(mins);
    denora_free(secs);
    denora_free(tmp);
    denora_free(tmp4);
    return;
}

/*************************************************************************/

/* SDESC */
void sql_do_sdesc(char *user, char *msg)
{
    if (!denora->do_sql) {
	    return;
    }
    if (LargeNet) {
        return;
    }
    user = rdb_escape(user);
    msg = rdb_escape(msg);
    rdb_query("UPDATE %s SET comment=\'%s\' WHERE servid=%d",
              ServerTable, msg, db_getservfromnick(user));
    denora_free(user);
    denora_free(msg);
}

/*************************************************************************/

void server_set_desc(char *server, char *msg)
{
    Server *s;
    s = findserver(server);

    if (s) {
        if (s->desc) {
            denora_free(s->desc);
        }
        s->desc = sstrdup(msg);
    }
    sql_do_sdesc(server, msg);
}

/*************************************************************************/

void sql_do_server_bans_add(char *type, char *user, char *host,
                            char *setby, char *setat, char *expires,
                            char *reason)
{
    if (!denora->do_sql) {
	    return;
    }
    if (LargeNet) {
        return;
    }

    reason = rdb_escape(reason);

    if (type) {
        rdb_query
            ("INSERT INTO %s (type, user, host, setby, setat, expires, reason) VALUES(\'%s\',\'%s\',\'%s\',\'%s\',%ld, %ld,\'%s\')",
             GlineTable, type, user, host, setby, strtoul(setat, NULL, 10),
             strtoul(expires, NULL, 10), reason);
    } else {
        rdb_query
            ("INSERT INTO %s (user, host, setby, setat, expires, reason) VALUES(\'%s\',\'%s\',\'%s\',%ld, %ld,\'%s\')",
             GlineTable, user, host, setby, strtoul(setat, NULL, 10),
             strtoul(expires, NULL, 10), reason);
    }
    if (reason) {
        denora_free(reason);
    }
}

/*************************************************************************/

void sql_do_server_spam_add(char *target, char *action,
                            char *setby, char *expires, char *setat, char *duration,
                            char *reason, char *regex)
{
    if (!denora->do_sql) {
	    return;
    }
    if (LargeNet) {
        return;
    }

    reason = rdb_escape(reason);
    regex = rdb_escape(regex);

    rdb_query
            ("INSERT INTO %s (target, action, setby, expires, setat, duration, reason, regex) VALUES(\'%s\',\'%s\',\'%s\',%ld, %ld,%ld, \'%s\', \'%s\')",
             SpamTable, target, action, setby, strtoul(expires, NULL, 10),
             strtoul(setat, NULL, 10), strtoul(duration, NULL, 10), reason, regex);
    if (reason) {
        denora_free(reason);
    }
    if (regex) {
        denora_free(regex);
    }
}

/*************************************************************************/

void sql_do_server_bans_remove(char *type, char *user, char *host)
{
    if (!denora->do_sql) {
	    return;
    }

    if (LargeNet) {
        return;
    }

    if (type) {
        rdb_query
            ("DELETE FROM %s WHERE type=\'%s\' and user=\'%s\' and host=\'%s\'",
             GlineTable, type, user, host);
    } else {
        rdb_query
            ("DELETE FROM %s WHERE user=\'%s\' and host=\'%s\'",
             GlineTable, user, host);
    }
}

/*************************************************************************/

void sql_do_server_spam_remove(char *target, char *action, char *regex)
{
    if (!denora->do_sql) {
	    return;
    }

    if (LargeNet) {
        return;
    }

    regex = rdb_escape(regex);
    rdb_query
            ("DELETE FROM %s WHERE target=\'%s\' and action=\'%s\' and regex=\'%s'",
             SpamTable, target, action, regex);
    if (regex) {
        denora_free(regex);
    }
}


/*************************************************************************/

/* Asuka 
   AB 351 BjAAA u2.10.11.07+asuka(1.2.1). Auska.Nomadirc.net :B96AeEFfIKMpSU 

   Viagra
   viagra.nomadirc.net 351 Denora VIAGRA_IRCd 1.3.0 - BETA 7. viagra.nomadirc.net :Ci TS5ow-r[BETA]
                            0       1          2    3  4   5   6                    7

   RageIRCD
   rage2.nomadirc.net 351 Denora :rage(bluemoon)-2.0.0.beta-8(1.956) rage2.nomadirc.net EBTsIFcMO TS5ow-r[BETA]
                           0       1

    IRCNet
    :ircnet.nomadirc.net 351 StatServ 2.11.1p1. ircnet.nomadirc.net 000A :aEFHJKMRTu

    Atheme IRC Services
    :services.nomadirc.net 351 StatServ :atheme-0.2. services.nomadirc.net FljR TS5ow
    :services.nomadirc.net 351 StatServ :Compile time: Thu Jun 16 2005 at 00:21:15 EDT, build-id 525, build 4
    :services.nomadirc.net 351 StatServ :Compiled on: [Linux denora.elite-irc.net 2.6.9 #1 Fri Oct 22 21:07:15 EDT 2004 i686 unknown]
*/
void sql_do_server_version(char *server, int ac, char **av)
{
    char *version = NULL;
    char *sqlversion = NULL;
    char *temp = NULL;
    char buf[100];
    int servid;
    Server *s;
    *buf = '\0';

    SET_SEGV_LOCATION();

    if (LargeNet) {
        return;
    }

    if (ircd->p10) {
        s = findserver_uid(server);
        if (!s) {
            s = findserver(server);
        }
    } else if (ircd->ts6 && UseTS6) {
        s = findserver_uid(server);
        if (!s) {
            s = findserver(server);
        }
    } else {
        s = findserver(server);
    }

    SET_SEGV_LOCATION();

    if (!s) {
        alog(LOG_DEBUG,"Version for non-existatnt (%s)", server);
        return;
    }

    /* Viagra */
    if (ac == 8) {
        snprintf(buf, 100, "%s %s %s %s %s", av[1], av[2], av[3], av[4], av[5]);
        version = sstrdup(buf);        
    }

    /* Unreal / P10 */
    if (ac == 4) {
        version = sstrdup(av[1]);
    }

    /* Anope/Thales respond like so */
    /* QuakeIRCD uses 6 */
    if (ac == 5 || ac == 6) {
        snprintf(buf, 100, "%s %s", av[1], av[2]);
        version = sstrdup(buf);
    }

    /* SolarStats and NeoStats like this */
    if (ac == 2) {
        temp = sstrdup(myStrGetToken(av[1], ' ', 0));
        /* if you have NeoStats mods load they to will respond on VERSION saying they are a 
           module */
        if (!stricmp(temp, "MODULE")) {
            denora_free(temp);
            return;
        /* Atheme uses 351 to tell the user what os it was
           built on */
        } else if (!stricmp(temp, "Compile")) {
            denora_free(temp);
            return;
        } else if (!stricmp(temp, "Compiled")) {
            denora_free(temp);
            return;
        } else {
            version = sstrdup(temp);
            denora_free(temp);
        }
    }

    /* inspircd */
    if (ac == 1) {
        version = sstrdup(av[0]);
    }

    if (!version) {
        alog(LOG_DEBUG, langstr(ALOG_DEBUG_UNKNOWN_VERSION_REPLY));
        alog(LOG_DEBUG, "debug: ac count is %d", ac);
        return;
    }

    if (s) {
        s->version = sstrdup(version);
    }
    if (denora->do_sql) {
        sqlversion = rdb_escape(version);
        if (((servid = db_checkserver(s->name)) != -1)) {
            rdb_query("UPDATE %s SET version=\"%s\" WHERE servid=%d",
                  ServerTable, version, servid);
        }
        denora_free(sqlversion);
    }
    denora_free(version);
}

/*************************************************************************/

/* SQUIT */
void sql_do_squit(char *server)
{
    server = rdb_escape(server);
    if (ServerCacheTime) {
        rdb_query
            ("UPDATE %s SET online=\"N\", lastsplit=NOW(),linkedto=NULL WHERE servid=%d",
             ServerTable, db_getserver(server));
        db_cleanserver();
    } else {
        rdb_query("DELETE FROM %s WHERE server=\'%s\'", ServerTable,
                  server);
    }
    denora_free(server);
}

/*************************************************************************/

/**
 * Handle removing the server from the Server struct
 * @param source Name of the server leaving
 * @param ac Number of arguments in av
 * @param av Agruments as part of the SQUIT
 * @return void
 */
void do_squit(char *servername)
{
    char buf[BUFSIZE];
    Server *s;
    *buf = '\0';

    if (ircd->p10) {
        s = findserver_uid(servername);
        if (!s) {
            s = findserver(servername);
        }
    } else if (ircd->ts6 && UseTS6) {
        s = findserver_uid(servername);
        if (!s) {
            s = findserver(servername);
        }
    } else {
        s = findserver(servername);
    }
    if (!s) {
        do_backtrace(1);
        alog(LOG_NONEXISTANT, "SQUIT for nonexistent server (%s)!!",
             servername);
        return;
    }

    snprintf(buf, BUFSIZE - 1, "%s %s", s->name,
             (s->uplink ? s->uplink : ""));

    if (ircdcap->unconnect) {
        if (s->uplink && me_server->uplink) {
            if (!stricmp(s->uplink, me_server->uplink)
                && (denora->capab & ircdcap->unconnect)) {
                alog(LOG_DEBUG, "debug: Sending UNCONNECT SQUIT for %s", s->name);
                denora_cmd_squit(s->name, buf);
            }
        }
    }

    send_event(EVENT_SQUIT, 2, s->name, buf);

    if (denora->do_sql) {
        sql_do_squit(servername);
    }
    stats->servers--;
    s->ss->splits++;
    s->ss->split_stats = 1;
    s->ss->lastseen = time(NULL);
    delete_server(s, buf);
}

/*************************************************************************/

/**
 * Handle parsing the CAPAB/PROTOCTL messages
 * @param ac Number of arguments in av
 * @param av Agruments 
 * @return void
 */
void capab_parse(int ac, char **av)
{
    int i;
    char *s, *tmp;

    char *temp;

    for (i = 0; i < ac; i++) {
        temp = av[i];

        s = myStrGetToken(temp, '=', 0);
        tmp = myStrGetTokenRemainder(temp, '=', 1);

        if (!s) {
            continue;
        }

        if (!stricmp(s, "NOQUIT")) {
            denora->capab |= CAPAB_NOQUIT;
        }
        if (!stricmp(s, "TSMODE")) {
            denora->capab |= CAPAB_TSMODE;
        }
        if (!stricmp(av[i], "UNCONNECT")) {
            denora->capab |= CAPAB_UNCONNECT;
        }
        if (!stricmp(s, "NICKIP")) {
            denora->capab |= CAPAB_NICKIP;
            /* Update the struct so that we know we can get NICKIP */
            if (!ircd->nickip) {
                ircd->nickip = 1;
            }
        }
        if (!stricmp(s, "NICKCHARS")) { 
            denora->capab |= CAPAB_NICKCHARS; 
            if (tmp) { 
              ircd->nickchars = sstrdup(tmp); 
            } 
        } 
        if (!stricmp(s, "SSJOIN")) {
            denora->capab |= CAPAB_NSJOIN;
        }
        if (!stricmp(s, "ZIP")) {
            denora->capab |= CAPAB_ZIP;
        }
        if (!stricmp(s, "BURST")) {
            denora->capab |= CAPAB_BURST;
        }
        if (!stricmp(s, "TS5")) {
            denora->capab |= CAPAB_TS5;
        }
        if (!stricmp(s, "TS3")) {
            denora->capab |= CAPAB_TS3;
        }
        if (!stricmp(s, "DKEY")) {
            denora->capab |= CAPAB_DKEY;
        }
        if (!stricmp(s, "PT4")) {
            denora->capab |= CAPAB_PT4;
        }
        if (!stricmp(s, "SCS")) {
            denora->capab |= CAPAB_SCS;
        }
        if (!stricmp(s, "QS")) {
            denora->capab |= CAPAB_QS;
        }
        if (!stricmp(s, "UID")) {
            denora->capab |= CAPAB_UID;
        }
        if (!stricmp(s, "KNOCK")) {
            denora->capab |= CAPAB_KNOCK;
        }
        if (!stricmp(s, "CLIENT")) {
            denora->capab |= CAPAB_CLIENT;
        }
        if (!stricmp(s, "IPV6")) {
            denora->capab |= CAPAB_IPV6;
        }
        if (!stricmp(s, "SSJ5")) {
            denora->capab |= CAPAB_SSJ5;
        }
        if (!stricmp(s, "SN2")) {
            denora->capab |= CAPAB_SN2;
        }
        if (!stricmp(s, "TOK1")) {
            denora->capab |= CAPAB_TOKEN;
        }
        if (!stricmp(s, "TOKEN")) {
            denora->capab |= CAPAB_TOKEN;
        }
        if (!stricmp(s, "VHOST")) {
            denora->capab |= CAPAB_VHOST;
        }
        if (!stricmp(s, "SSJ3")) {
            denora->capab |= CAPAB_SSJ3;
        }
        if (!stricmp(s, "SJB64")) {
            denora->capab |= CAPAB_SJB64;
        }
        if (!stricmp(s, "CHANMODES")) {
            denora->capab |= CAPAB_CHANMODE;
            if (tmp) {
                ircd->chanmodes = sstrdup(tmp);
            }
        }
        if (s) {
            denora_free(s);
        }
        if (tmp) {
            denora_free(tmp);
        }

    }
}

/* EOF */
