
/*
 *
 * (C) 2004-2005 Denora Team
 * Contact us at info@nomadirc.net
 *
 * Please read COPYING and README for furhter details.
 *
 * Based on the original code of Anope by Anope Team.
 * Based on the original code of Thales by Lucas.
 * 
 * $Id: xml.c 427 2005-07-23 08:40:36Z trystan $
 *
 */

#include "denora.h"

char *xml_encode;
char *xml_doctype;
char *xml_header;

/*************************************************************************/

FILE *new_xml(char *filename)
{
    FILE *xmlfile;

    SET_SEGV_LOCATION();

    if ((xmlfile = fopen(filename, "w")) == NULL) {
        alog(LOG_ERROR, "Error accessing file: %s", filename);
        return NULL;
    }
    return xmlfile;
}

/*************************************************************************/

void xml_write_header(FILE * ptr)
{
    SET_SEGV_LOCATION();

    if (xml_encode) {
        fprintf(ptr, "<?xml version=\"1.0\" encoding=\"%s\" ?>\n\r", xml_encode);
    } else {
        fprintf(ptr, "<?xml version=\"1.0\" ?>\n\r");
    }
    if (xml_doctype) {
        fprintf(ptr, "%s\n\r", xml_doctype);
    }
    if (xml_header) {
        fprintf(ptr, "<%s>\n\r", xml_header);
    } else {
        fprintf(ptr, "<denora>\n\r");
    }
}

/*************************************************************************/

void xml_write_footer(FILE * ptr)
{
    SET_SEGV_LOCATION();
    if (xml_header) {
        fprintf(ptr, "</%s>\n\r", xml_header);
    } else {
        fprintf(ptr, "</denora>\n\r");
    }
    fclose(ptr);
}

/*************************************************************************/

void xml_write_tag(FILE * ptr, const char *tag, char *data)
{
    char *mydata;

    SET_SEGV_LOCATION();

    mydata = char_encode(data);

    fprintf(ptr, "<%s>%s</%s>\n\r", tag, mydata, tag);
    denora_free(mydata);
}

/*************************************************************************/

void xml_moduleData(FILE * ptr, ModuleData ** md)
{
    ModuleData *modcurrent = NULL;
    SET_SEGV_LOCATION();

    if (md) {
     fprintf(ptr, "<moduledata>");
     for (modcurrent = *md; modcurrent; modcurrent = modcurrent->next) {
        fprintf(ptr, "<%s>", modcurrent->moduleName);
        xml_write_tag(ptr, modcurrent->key, modcurrent->value);
        fprintf(ptr, "</%s>", modcurrent->moduleName);
     }
     fprintf(ptr, "</moduledata>");
    } else {
     fprintf(ptr, "<moduledata />");
    }
    return;
}

/*************************************************************************/

char *xml_prepare_tag(char *tag, char *data)
{
    char *mydata;
    char buf[BUFSIZE];
    *buf = '\0';

    SET_SEGV_LOCATION();

    mydata = char_encode(data);

    snprintf(buf, BUFSIZE, "<%s>%s</%s>\n\r", tag, mydata, tag);
    denora_free(mydata);
    return sstrdup(buf);
}

/*************************************************************************/

void xml_write_tag_int(FILE * ptr, const char *tag, long int data)
{
    SET_SEGV_LOCATION();
    fprintf(ptr, "<%s>%ld</%s>\n\r", tag, (long int) data, tag);
}

/*************************************************************************/

void xml_write_block(FILE * ptr, char *block, char *data)
{
    SET_SEGV_LOCATION();
    fprintf(ptr, "<%s>\n\r", block);
    fprintf(ptr, "%s", data);
    fprintf(ptr, "</%s>\n\r", block);
}

/*************************************************************************/

void xml_write_block_top(FILE * ptr, const char *block)
{
    SET_SEGV_LOCATION();
    fprintf(ptr, "<%s>\n\r", block);
}

/*************************************************************************/

void xml_write_block_bottom(FILE * ptr, const char *block)
{
    SET_SEGV_LOCATION();
    fprintf(ptr, "</%s>\n\r", block);
}

/*************************************************************************/

void xml_export_channels(char *file)
{
    FILE *ptr;
    Channel *c, *next;
    int i;
    char buf[BUFSIZE];
    CBMode *cbm;
    struct c_userlist *cu;
    int32 count;
    int ping = 0;
    int counttoping = 1;
    char **bans;

    *buf = '\0';

    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "channels");
        c = firstchan();
        while (c) {
            next = nextchan();
            if (ping) {
                denora_cmd_pong(ServerName, ServerName);
                ping = 0;
            }
            xml_write_block_top(ptr, "chans");
            xml_write_tag(ptr, "name", c->name);

            xml_write_tag(ptr, "topic", c->topic);
            xml_write_tag(ptr, "topicsetter", c->topic_setter);
            xml_write_tag_int(ptr, "creation_time",
                              (long int) c->creation_time);
            xml_write_tag_int(ptr, "topic_time", (long int) c->topic_time);
            xml_write_tag_int(ptr, "topic_count", c->stats->topic_count);
            xml_write_tag_int(ptr, "limit", c->limit);
            xml_write_tag_int(ptr, "usercount",
                              (long int) c->stats->usercount);
            xml_write_tag_int(ptr, "maxusercount",
                              (long int) c->stats->maxusercount);
            xml_write_tag_int(ptr, "kickcount",
                              (long int) c->stats->kickcount);
            xml_write_tag_int(ptr, "joincounter",
                              (long int) c->stats->joincounter);

            for (i = 0; i < 128; i++) {
                cbm = &cbmodes[i];
                if (c->mode & cbm->flag) {
                    snprintf(buf, BUFSIZE - 1, "mode_%c", (int) i);
                    xml_write_tag_int(ptr, buf, 1);
                    *buf = '\0';
                }
            }
            if (c->limit) {
                xml_write_tag_int(ptr, "limit", 1);
            }

            if (c->key) {
                xml_write_tag_int(ptr, "key", 1);
            }

            if (ircd->Lmode && c->redirect) {
                xml_write_tag(ptr, "redirect", c->redirect);
            }
            if (ircd->fmode && c->flood) {
                xml_write_tag(ptr, "flood", c->flood);
            }

            xml_write_block_top(ptr, "inchannel");
            for (cu = c->users; cu; cu = cu->next) {
                xml_write_tag(ptr, "user", cu->user->nick);
            }
            xml_write_block_bottom(ptr, "inchannel");

            if (c->bancount) {
                xml_write_tag_int(ptr, "bancount", c->bancount);
                xml_write_block_top(ptr, "bans");
                count = c->bancount;
                bans = scalloc(sizeof(char *) * count, 1);
                for (i = 0; i < count; i++) {
                    bans[i] = sstrdup(c->bans[i]);
                    for (i = 0; i < count; i++) {
                        xml_write_tag(ptr, "ban", bans[i]);
                    }
                }
                denora_free(bans);
                xml_write_block_bottom(ptr, "bans");
            }

            if (ircd->except && c->exceptcount) {
                xml_write_tag_int(ptr, "exceptcount", c->exceptcount);
                xml_write_block_top(ptr, "exceptions");
                count = c->exceptcount;
                bans = scalloc(sizeof(char *) * count, 1);
                for (i = 0; i < count; i++) {
                    bans[i] = sstrdup(c->excepts[i]);
                    for (i = 0; i < count; i++) {
                        xml_write_tag(ptr, "except", bans[i]);
                    }
                }
                denora_free(bans);
                xml_write_block_bottom(ptr, "exceptions");
            }

            if (ircd->invitemode && c->invitecount) {
                xml_write_tag_int(ptr, "invitecount", c->invitecount);
                xml_write_block_top(ptr, "invites");
                count = c->invitecount;
                bans = scalloc(sizeof(char *) * count, 1);
                for (i = 0; i < count; i++) {
                    bans[i] = sstrdup(c->invite[i]);
                    for (i = 0; i < count; i++) {
                        xml_write_tag(ptr, "invite", bans[i]);
                    }
                }
                denora_free(bans);
                xml_write_block_bottom(ptr, "invites");
            }


            xml_write_block_bottom(ptr, "chans");
            c = next;
            if (counttoping == 10) {
                ping = 1;
                counttoping = 1;
            } else {
                counttoping++;
            }
        }
        xml_write_block_bottom(ptr, "channels");
        xml_write_footer(ptr);
    }
}

/*************************************************************************/

void xml_export_tld(char *file)
{
    FILE *ptr;
    TLD *t, *next;

    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "tld");
        t = first_tld();
        while (t) {
            next = next_tld();
            denora_cmd_pong(ServerName, ServerName);
            xml_write_block_top(ptr, "domain");
            xml_write_tag(ptr, "countrycode", t->countrycode);
            xml_write_tag(ptr, "country", t->country);
            xml_write_tag_int(ptr, "current", t->count);
            xml_write_tag_int(ptr, "overall", t->overall);
            xml_write_block_bottom(ptr, "domain");
            t = next;
        }
        xml_write_block_bottom(ptr, "tld");
        xml_write_footer(ptr);
    }
}

/*************************************************************************/

void xml_export_ctcp(char *file)
{
    FILE *ptr;
    CTCPVerStats *t, *next;

    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "ctcp");
        t = first_ctcp();
        while (t) {
            next = next_ctcp();
            denora_cmd_pong(ServerName, ServerName);
            xml_write_block_top(ptr, "client");
            xml_write_tag(ptr, "version", t->version);
            xml_write_tag_int(ptr, "current", t->count);
            xml_write_tag_int(ptr, "overall", t->overall);
            xml_write_block_bottom(ptr, "client");
            t = next;
        }
        xml_write_block_bottom(ptr, "ctcp");
        xml_write_footer(ptr);
    }
}

/*************************************************************************/

void xml_export_servers(char *file)
{
    FILE *ptr;
    Server *s, *next;

    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "servers");
        s = first_server();
        while (s) {
            next = next_server();
            denora_cmd_pong(ServerName, ServerName);
            xml_write_block_top(ptr, "server");
            xml_write_tag(ptr, "name", s->name);
            xml_write_tag_int(ptr, "hops", s->hops);
            xml_write_tag(ptr, "desc", s->desc);
            xml_write_tag_int(ptr, "flags", s->flags);

            if (s->flags & SERVER_ISME) {
                xml_write_tag_int(ptr, "FLAG_ISME", 1);
            } else {
                xml_write_tag_int(ptr, "FLAG_ISME", 0);
            }
            if (s->flags & SERVER_JUPED) {
                xml_write_tag_int(ptr, "FLAG_JUPED", 1);
            } else {
                xml_write_tag_int(ptr, "FLAG_JUPED", 0);
            }
            xml_write_tag_int(ptr, "synced", s->sync);
            if (s->suid) {
                xml_write_tag(ptr, "suid", s->suid);
            }
            if (s->version) {
                xml_write_tag(ptr, "version", s->version);
            }
            if (s->uptime) {
                xml_write_tag_int(ptr, "uptime", s->uptime);
            }
            if (s->uplink) {
                xml_write_tag(ptr, "uplink", s->uplink);
            }
	    xml_write_tag_int(ptr, "uline", s->uline);
	    
            xml_write_block_bottom(ptr, "server");
            s = next;
        }
        xml_write_block_bottom(ptr, "servers");
        xml_write_footer(ptr);
    }
}

/*************************************************************************/

void xml_export_users(char *file)
{
    FILE *ptr;
    User *u, *next;
    int i;
    char buf[BUFSIZE];
    struct u_chanlist *uc;

    *buf = '\0';

    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "users");
        u = firstuser();
        SET_SEGV_LOCATION();
        while (u) {
            next = nextuser();
            denora_cmd_pong(ServerName, ServerName);
            SET_SEGV_LOCATION();
            xml_write_block_top(ptr, "user");
            xml_write_tag(ptr, "name", u->nick);
            xml_write_tag(ptr, "username", u->username);
            xml_write_tag(ptr, "host", u->host);
            if (ircd->vhost) {
                xml_write_tag(ptr, "vhost", u->vhost);
            }
            if (ircd->vident && u->vident) {
                xml_write_tag(ptr, "vident", u->vident);
            }
            xml_write_tag(ptr, "realname", u->realname);
            xml_write_tag(ptr, "server", u->server->name);
            if (u->ip) {
                xml_write_tag(ptr, "ip", u->ip);
            }
            xml_write_tag(ptr, "country_code", u->country_code);
            xml_write_tag(ptr, "country_name", u->country_name);
            xml_write_tag_int(ptr, "timestamp", (long int) u->timestamp);
            xml_write_tag_int(ptr, "my_signon", (long int) u->my_signon);
            xml_write_tag_int(ptr, "svid", u->svid);
            xml_write_tag_int(ptr, "mode", u->mode);
            xml_write_tag_int(ptr, "language", u->language);
            xml_write_tag_int(ptr, "admin", u->admin);
            for (i = 0; i < 128; i++) {
                if (u->mode & umodes[i]) {
                    snprintf(buf, BUFSIZE - 1, "mode_%c", (int) i);
                    xml_write_tag_int(ptr, buf, 1);
                    *buf = '\0';
                }
            }
            xml_write_block_top(ptr, "inchannels");
            for (uc = u->chans; uc; uc = uc->next) {
                xml_write_tag(ptr, "channel", uc->chan->name);
            }
            xml_write_block_bottom(ptr, "inchannels");
            xml_moduleData(ptr, &u->moduleData);
            xml_write_block_bottom(ptr, "user");
            u = next;
        }
        xml_write_block_bottom(ptr, "users");
        xml_write_footer(ptr);
    }
}

/*************************************************************************/

void xml_export_stats(char *file)
{
    FILE *ptr;
    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "stats");
        xml_write_tag_int(ptr, "users_max", stats->users_max);
        xml_write_tag_int(ptr, "users", stats->users);
        xml_write_tag_int(ptr, "chans_max", stats->chans_max);
        xml_write_tag_int(ptr, "chans", stats->chans);
        xml_write_tag_int(ptr, "servs_max", stats->servers_max);
        xml_write_tag_int(ptr, "servs", stats->servers);
        xml_write_block_bottom(ptr, "stats");
        xml_write_footer(ptr);
    }

}

/*************************************************************************/

void xml_export_all(char *file)
{
    FILE *ptr;
    User *u, *next;
    Server *s, *snext;
    TLD *t, *tnext;
    Channel *c, *cnext;
    int i, count;
    char buf[BUFSIZE];
    struct u_chanlist *uc;
    struct c_userlist *cu;
    CBMode *cbm;
    char **bans;
    int ping = 0;
    int counttoping = 1;

    ptr = new_xml(file);

    SET_SEGV_LOCATION();

    if (ptr) {
        xml_write_header(ptr);

        xml_write_block_top(ptr, "stats");
        xml_write_tag_int(ptr, "users_max", stats->users_max);
        xml_write_tag_int(ptr, "users", stats->users);
        xml_write_tag_int(ptr, "chans_max", stats->chans_max);
        xml_write_tag_int(ptr, "chans", stats->chans);
        xml_write_tag_int(ptr, "servs_max", stats->servers_max);
        xml_write_tag_int(ptr, "servs", stats->servers);
        xml_write_block_bottom(ptr, "stats");

        denora_cmd_pong(ServerName, ServerName);

        xml_write_block_top(ptr, "users");
        u = firstuser();
        SET_SEGV_LOCATION();
        while (u) {
            next = nextuser();
            if (ping) {
                denora_cmd_pong(ServerName, ServerName);
                ping = 0;
            }
            SET_SEGV_LOCATION();
            xml_write_block_top(ptr, "user");
            xml_write_tag(ptr, "name", u->nick);
            xml_write_tag(ptr, "username", u->username);
            xml_write_tag(ptr, "host", u->host);
            if (ircd->vhost) {
                xml_write_tag(ptr, "vhost", u->vhost);
            }
            if (ircd->vident && u->vident) {
                xml_write_tag(ptr, "vident", u->vident);
            }
            xml_write_tag(ptr, "realname", u->realname);
            xml_write_tag(ptr, "server", u->server->name);
            if (u->ip) {
                xml_write_tag(ptr, "ip", u->ip);
            }
            xml_write_tag(ptr, "country_code", u->country_code);
            xml_write_tag(ptr, "country_name", u->country_name);
            xml_write_tag_int(ptr, "timestamp", (long int) u->timestamp);
            xml_write_tag_int(ptr, "my_signon", (long int) u->my_signon);
            xml_write_tag_int(ptr, "svid", u->svid);
            xml_write_tag_int(ptr, "mode", u->mode);
            xml_write_tag_int(ptr, "language", u->language);
            xml_write_tag_int(ptr, "admin", u->admin);
            for (i = 0; i < 128; i++) {
                if (u->mode & umodes[i]) {
                    snprintf(buf, BUFSIZE - 1, "mode_%c", (int) i);
                    xml_write_tag_int(ptr, buf, 1);
                    *buf = '\0';
                }
            }
            xml_write_block_top(ptr, "inchannels");
            for (uc = u->chans; uc; uc = uc->next) {
                xml_write_tag(ptr, "channel", uc->chan->name);
            }
            xml_write_block_bottom(ptr, "inchannels");
            xml_moduleData(ptr, &u->moduleData);
            xml_write_block_bottom(ptr, "user");
            u = next;
            if (counttoping == 10) {
                ping = 1;
                counttoping = 1;
            } else {
                counttoping++;
            }
        }
        xml_write_block_bottom(ptr, "users");

        xml_write_block_top(ptr, "servers");
        s = first_server();
        while (s) {
            snext = next_server();
            if (ping) {
                denora_cmd_pong(ServerName, ServerName);
                ping = 0;
            }
            xml_write_block_top(ptr, "server");
            xml_write_tag(ptr, "name", s->name);
            xml_write_tag_int(ptr, "hops", s->hops);
            xml_write_tag(ptr, "desc", s->desc);
            xml_write_tag_int(ptr, "flags", s->flags);
            xml_write_tag_int(ptr, "synced", s->sync);
            if (s->suid) {
                xml_write_tag(ptr, "suid", s->suid);
            }
            if (s->version) {
                xml_write_tag(ptr, "version", s->version);
            }
            if (s->uptime) {
                xml_write_tag_int(ptr, "uptime", s->uptime);
            }
            if (s->uplink) {
                xml_write_tag(ptr, "uplink", s->uplink);
            }
            xml_write_block_top(ptr, "serverstats");
            if (s->ss->currentusers) {
                xml_write_tag_int(ptr, "currentusers",
                                  s->ss->currentusers);
            }
            xml_write_tag_int(ptr, "maxusers", s->ss->maxusers);
            xml_write_block_bottom(ptr, "serverstats");

            xml_write_block_bottom(ptr, "server");
            s = snext;
            if (counttoping == 10) {
                ping = 1;
                counttoping = 1;
            } else {
                counttoping++;
            }
        }
        xml_write_block_bottom(ptr, "servers");


        xml_write_block_top(ptr, "tld");
        t = first_tld();
        while (t) {
            tnext = next_tld();
            if (ping) {
                denora_cmd_pong(ServerName, ServerName);
                ping = 0;
            }
            xml_write_block_top(ptr, "domain");
            xml_write_tag(ptr, "countrycode", t->countrycode);
            xml_write_tag(ptr, "country", t->country);
            xml_write_tag_int(ptr, "current", t->count);
            xml_write_tag_int(ptr, "overall", t->overall);
            xml_write_block_bottom(ptr, "domain");
            t = tnext;
            if (counttoping == 10) {
                ping = 1;
                counttoping = 1;
            } else {
                counttoping++;
            }
        }
        xml_write_block_bottom(ptr, "tld");

        xml_write_block_top(ptr, "channels");
        c = firstchan();
        while (c) {
            cnext = nextchan();
            if (ping) {
                denora_cmd_pong(ServerName, ServerName);
                ping = 0;
            }
            xml_write_block_top(ptr, "chans");
            xml_write_tag(ptr, "name", c->name);

            xml_write_tag(ptr, "topic", c->topic);
            xml_write_tag(ptr, "topicsetter", c->topic_setter);
            xml_write_tag_int(ptr, "creation_time",
                              (long int) c->creation_time);
            xml_write_tag_int(ptr, "topic_time", (long int) c->topic_time);
            xml_write_tag_int(ptr, "topic_count", c->stats->topic_count);
            xml_write_tag_int(ptr, "limit", c->limit);
            xml_write_tag_int(ptr, "usercount",
                              (long int) c->stats->usercount);
            xml_write_tag_int(ptr, "maxusercount",
                              (long int) c->stats->maxusercount);
            xml_write_tag_int(ptr, "kickcount",
                              (long int) c->stats->kickcount);
            xml_write_tag_int(ptr, "joincounter",
                              (long int) c->stats->joincounter);

            for (i = 0; i < 128; i++) {
                cbm = &cbmodes[i];
                if (c->mode & cbm->flag) {
                    snprintf(buf, BUFSIZE - 1, "mode_%c", (int) i);
                    xml_write_tag_int(ptr, buf, 1);
                    *buf = '\0';
                }
            }
            if (c->limit) {
                xml_write_tag_int(ptr, "limit", 1);
            }

            if (c->key) {
                xml_write_tag_int(ptr, "key", 1);
            }

            if (ircd->Lmode && c->redirect) {
                xml_write_tag(ptr, "redirect", c->redirect);
            }
            if (ircd->fmode && c->flood) {
                xml_write_tag(ptr, "flood", c->flood);
            }

            xml_write_block_top(ptr, "inchannel");
            for (cu = c->users; cu; cu = cu->next) {
                xml_write_tag(ptr, "user", cu->user->nick);
            }
            xml_write_block_bottom(ptr, "inchannel");

            if (c->bancount) {
                xml_write_tag_int(ptr, "bancount", c->bancount);
                xml_write_block_top(ptr, "bans");
                count = c->bancount;
                bans = scalloc(sizeof(char *) * count, 1);
                for (i = 0; i < count; i++) {
                    bans[i] = sstrdup(c->bans[i]);
                    for (i = 0; i < count; i++) {
                        xml_write_tag(ptr, "ban", bans[i]);
                    }
                }
                denora_free(bans);
                xml_write_block_bottom(ptr, "bans");
            }

            if (ircd->except && c->exceptcount) {
                xml_write_tag_int(ptr, "exceptcount", c->exceptcount);
                xml_write_block_top(ptr, "exceptions");
                count = c->exceptcount;
                bans = scalloc(sizeof(char *) * count, 1);
                for (i = 0; i < count; i++) {
                    bans[i] = sstrdup(c->excepts[i]);
                    for (i = 0; i < count; i++) {
                        xml_write_tag(ptr, "except", bans[i]);
                    }
                }
                denora_free(bans);
                xml_write_block_bottom(ptr, "exceptions");
            }

            if (ircd->invitemode && c->invitecount) {
                xml_write_tag_int(ptr, "invitecount", c->invitecount);
                xml_write_block_top(ptr, "invites");
                count = c->invitecount;
                bans = scalloc(sizeof(char *) * count, 1);
                for (i = 0; i < count; i++) {
                    bans[i] = sstrdup(c->invite[i]);
                    for (i = 0; i < count; i++) {
                        xml_write_tag(ptr, "invite", bans[i]);
                    }
                }
                denora_free(bans);
                xml_write_block_bottom(ptr, "invites");
            }
            xml_write_block_bottom(ptr, "chans");
            c = cnext;
            if (counttoping == 10) {
                ping = 1;
                counttoping = 1;
            } else {
                counttoping++;
            }
        }
        xml_write_block_bottom(ptr, "channels");

        xml_write_footer(ptr);
    }
}

/*************************************************************************/


